#include <stdio.h>
#include <memory.h>
#include "lillibsdl.h"
#include <LibSdl2/include/SDL.h>

//-----------------------------------------------------------------------------------------------------------
// FILE IO
//-----------------------------------------------------------------------------------------------------------
struct LilIoMemHandle {
	lil_u8 *ptr;
	lil_u32 size;
	lil_u32 pos;
	lil_u32 maxPos;  // helpful to detect only used size...
};

//-----------------------------------------------------------------------------------------------------------
static LilErrorType LilIoFileRead(lil_io_handle handle, size_t byteCount, void *outMemPtr) {
	FILE *file = (FILE *)handle;
	if (fread(outMemPtr, 1, byteCount, file) != byteCount)
		return LE_IOOUTOFRANGE;
	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
static LilErrorType LilIoFileWrite(lil_io_handle handle, size_t byteCount, LIL_CONST void *inMemPtr) {
	FILE *file = (FILE *)handle;
	if (fwrite(inMemPtr, 1, byteCount, file) != byteCount)
		return LE_IOOUTOFRANGE;
	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
static LilErrorType LilIoFileSeek(lil_io_handle handle, lil_i64 offset, LilSeekType mode) {
	FILE *file = (FILE *)handle;
	int fsMode = SEEK_SET;
	if (mode == LS_RELATIVE) {
		fsMode = SEEK_CUR;
		if (offset == 0)
			return LE_NONE;  // This case crashes???
	}
	else if (mode == LS_END)
		fsMode = SEEK_END;
	fseek(file, offset, fsMode);
	//if (_fseeki64(file, offset, fsMode))
	//	return LE_IOOUTOFRANGE;
	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
static lil_u64 LilIoFileTell(lil_io_handle handle) {
	FILE *file = (FILE *)handle;
	return ftell(file);
}

//-----------------------------------------------------------------------------------------------------------
static lil_bool LilOpenFile(LilIoDesc *pIo, const char *filename, const char *type) {
	FILE *file = fopen(filename, type);
	if (!file)
		return lil_false;
	pIo->handle = (lil_io_handle)file;
	pIo->readFunc = LilIoFileRead;
	pIo->writeFunc = LilIoFileWrite;
	pIo->seekFunc = LilIoFileSeek;
	pIo->tellFunc = LilIoFileTell;
	return lil_true;
}

//-----------------------------------------------------------------------------------------------------------
static void LilCloseFile(LilIoDesc *pIo) {
	if (pIo->handle) {
		fclose((FILE *)pIo->handle);
		pIo->handle = 0;
		pIo->readFunc = 0;
		pIo->writeFunc = 0;
		pIo->seekFunc = 0;
		pIo->tellFunc = 0;
	}
}




//-----------------------------------------------------------------------------------------------------------
// IMAGE MEM IO
//-----------------------------------------------------------------------------------------------------------
static LilErrorType LilImageMemAlloc(lil_image_handle handle, LIL_CONST LilImageDesc *pImage) {
	SDL_Surface **ppSurface = (SDL_Surface **)handle;
	int bpp = (int)LilImageGetBpp(pImage);
	*ppSurface = SDL_CreateRGBSurface(0, pImage->data.width, pImage->data.height, bpp, 0xff, 0xff00, 0xff0000, 0xff000000);
	if (!*ppSurface)
		return LE_NOMEMBUFFER;
	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
static LilErrorType LilImageMemFree(lil_image_handle handle) {
	if (handle)
		SDL_FreeSurface(*(SDL_Surface **)handle);
	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
static LilErrorType LilImageMemWrite(lil_image_handle handle, LIL_CONST void *pMem, lil_u32 x, lil_u32 y, lil_u32 w, lil_u32 h, lil_u32 plane, lil_u32 bpl) {
	lil_u32 iY;
	SDL_Surface *pSurface = *(SDL_Surface **)handle;
	lil_u8 *pOut8 = ((lil_u8 *)pSurface->pixels) + y * pSurface->pitch + x * 4;
	LIL_CONST lil_u8 *pIn8 = (LIL_CONST lil_u8 *)pMem;

	for (iY = 0; iY < h; iY++) {
		memcpy(pOut8, pIn8, w * 4);
		pOut8 += pSurface->pitch;
		pIn8 += bpl;
	}
	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
static LilErrorType LilImageMemRead(lil_image_handle handle, void *pMem, lil_u32 x, lil_u32 y, lil_u32 w, lil_u32 h, lil_u32 plane, lil_u32 bpl) {
	lil_u32 iY;
	SDL_Surface *pSurface = *(SDL_Surface **)handle;
	lil_u8 *pIn8 = ((lil_u8 *)pSurface->pixels) + y * pSurface->pitch + x * 4;
	lil_u8 *pOut8 = (lil_u8 *)pMem;
	
	for (iY = 0; iY < h; iY++) {
		memcpy(pOut8, pIn8, w * 4);
		pIn8 += pSurface->pitch;
		pOut8 += bpl;
	}
	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
static LilErrorType LilMemRealloc(lil_mem_handle *pHandle, lil_u64 bytes) {
	if (!pHandle)
		return LE_NOMEMBUFFER;
	if (bytes == 0) {
		free(*pHandle);
		*pHandle = 0;
	}
	else
		*pHandle = realloc(*pHandle, (size_t)bytes);
	if (bytes && !*pHandle)
		return LE_NOTENOUGHMEMORY;
	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
static void SetLilMemFuncs(LilMemDesc *outLilMemDesc) {
	outLilMemDesc->memReallocFunc = LilMemRealloc;
	outLilMemDesc->imageAllocFunc = LilImageMemAlloc;
	outLilMemDesc->imageFreeFunc = LilImageMemFree;
	outLilMemDesc->imageWriteFunc = LilImageMemWrite;
	outLilMemDesc->imageReadFunc = LilImageMemRead;
	outLilMemDesc->mem = 0;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType LoadSdlTexFromFile(SDL_Texture **ppTex, SDL_Renderer *pRend, const char *filename, int page, int colorMask) {
    SDL_Surface *loadedImage = 0;
    SDL_Texture *texture = 0;

	LilErrorType err;
	LilIoDesc io;
	LilMemDesc mem;
	LilImageDesc img;
	LilFormatDesc format;
	LilFileFormatType fileFormat;
	void *pix;
	int span;

	if (!LilOpenFile(&io, filename, "rb"))
		return LE_IOERROR;
	fileFormat = LilUtilGetFileFormat(filename, &io);
	SetLilMemFuncs(&mem);
	LilImageInit(&img, &mem, (lil_image_handle)&loadedImage);
	LilUtilSetFormat(&format, LDF_SIMPLE, 8, 4, 4, lil_false, LCO_RGBA, lil_false);
	img.extended.pageIndex = page;

	err = LilImageLoad(fileFormat, &io, &img, LL_DATA_AND_INFO, &format);

	LilImageFree(&img);
	LilCloseFile(&io);
	if (err != LE_NONE)
		return err;
	if (!loadedImage)
		return LE_IOERROR;
	if (colorMask)
		SDL_SetColorKey(loadedImage, 1, colorMask);

	*ppTex = SDL_CreateTextureFromSurface(pRend, loadedImage);

	SDL_FreeSurface(loadedImage);
	
	return LE_NONE;
}
