#include "lillibwx.h"
#include <stdio.h>
#include <wx/image.h>

namespace {

	//-----------------------------------------------------------------------------------------------------------
	// FILE IO
	//-----------------------------------------------------------------------------------------------------------

	LilErrorType LilIoFileRead(lil_io_handle handle, size_t byteCount, void *outMemPtr) {
		FILE *file = reinterpret_cast<FILE *>(handle);
		if (fread(outMemPtr, 1, byteCount, file) != byteCount)
			return LE_IOOUTOFRANGE;
		return LE_NONE;
	}

	//-----------------------------------------------------------------------------------------------------------
	LilErrorType LilIoFileWrite(lil_io_handle handle, size_t byteCount, LIL_CONST void *inMemPtr) {
		FILE *file = reinterpret_cast<FILE *>(handle);
		if (fwrite(inMemPtr, 1, byteCount, file) != byteCount)
			return LE_IOOUTOFRANGE;
		return LE_NONE;
	}

	//-----------------------------------------------------------------------------------------------------------
	LilErrorType LilIoFileSeek(lil_io_handle handle, lil_i64 offset, LilSeekType mode) {
		FILE *file = reinterpret_cast<FILE *>(handle);
		int fsMode = SEEK_SET;
		if (mode == LS_RELATIVE) {
			fsMode = SEEK_CUR;
			if (offset == 0)
				return LE_NONE;  // This case crashes???
		}
		else if (mode == LS_END)
			fsMode = SEEK_END;
		if (fseek(file, offset, fsMode))
			return LE_IOOUTOFRANGE;
		return LE_NONE;
	}

	//-----------------------------------------------------------------------------------------------------------
	lil_u64 LilIoFileTell(lil_io_handle handle) {
		FILE *file = reinterpret_cast<FILE *>(handle);
		return ftell(file);
	}

	//-----------------------------------------------------------------------------------------------------------
	void GetLilIoFile(FILE *fileHandle, LilIoDesc *outLilIoDesc) {
		outLilIoDesc->handle = reinterpret_cast<lil_io_handle>(fileHandle);
		outLilIoDesc->readFunc = &LilIoFileRead;
		outLilIoDesc->writeFunc = &LilIoFileWrite;
		outLilIoDesc->seekFunc = &LilIoFileSeek;
		outLilIoDesc->tellFunc = &LilIoFileTell;
	}

	//-----------------------------------------------------------------------------------------------------------
	// IMAGE MEM IO
	//-----------------------------------------------------------------------------------------------------------

	LilErrorType LilImageAlloc(lil_image_handle handle, LIL_CONST LilImageDesc *pImage) {
		wxImage *pImg = reinterpret_cast<wxImage *>(handle);
		if (!pImg)
			return LE_NOIMAGE;
		if (!pImg->Create(pImage->data.width, pImage->data.height, false))
			return LE_NOTENOUGHMEMORY;
		return LE_NONE;
	}

	//-----------------------------------------------------------------------------------------------------------
	LilErrorType LilImageFree(lil_image_handle handle) {
		wxImage *pImg = reinterpret_cast<wxImage *>(handle);
		if (pImg)
			pImg->Destroy();
		return LE_NONE;
	}

	//-----------------------------------------------------------------------------------------------------------
	LilErrorType LilImageWrite(lil_image_handle handle, LIL_CONST void *pMem, lil_u32 x, lil_u32 y, lil_u32 w, lil_u32 h, lil_u32 plane, lil_u32 bpl) {
		wxImage *pImg = reinterpret_cast<wxImage *>(handle);
		lil_u32 outBpl = pImg->GetWidth() * 3;
		lil_u8 *pOut8 = pImg->GetData() + y * outBpl + x * 3;
		LIL_CONST lil_u8 *pIn8 = reinterpret_cast<LIL_CONST lil_u8 *>(pMem);
		for (lil_u32 iY = 0; iY < h; iY++) {
			memcpy(pOut8, pIn8, bpl);
			pOut8 += outBpl;
			pIn8 += bpl;
		}
		return LE_NONE;
	}

	//-----------------------------------------------------------------------------------------------------------
	LilErrorType LilImageRead(lil_image_handle handle, void *pMem, lil_u32 x, lil_u32 y, lil_u32 w, lil_u32 h, lil_u32 plane, lil_u32 bpl) {
		wxImage *pImg = reinterpret_cast<wxImage *>(handle);
		lil_u32 inBpl = pImg->GetWidth() * 3;
		LIL_CONST lil_u8 *pIn8 = pImg->GetData() + y * inBpl + x * 3;
		lil_u8 *pOut8 = reinterpret_cast<lil_u8 *>(pMem);
		for (lil_u32 iY = 0; iY < h; iY++) {
			memcpy(pOut8, pIn8, bpl);
			pOut8 += bpl;
			pIn8 += inBpl;
		}
		return LE_NONE;
	}

	//-----------------------------------------------------------------------------------------------------------
	LilErrorType LilMemRealloc(lil_mem_handle *pHandle, lil_u64 bytes) {
		if (!pHandle)
			return LE_NOMEMBUFFER;
		if (bytes == 0) {
			free(*pHandle);
			*pHandle = 0;
		}
		else
			*pHandle = realloc(*pHandle, bytes);
		if (bytes && !*pHandle)
			return LE_NOTENOUGHMEMORY;
		return LE_NONE;
	}

	//-----------------------------------------------------------------------------------------------------------
	void GetLilMemWx(LilMemReallocFunc realloc, LilMemViewFunc view, LilMemDesc *outLilMemDesc) {
		outLilMemDesc->memReallocFunc = realloc;
		outLilMemDesc->imageAllocFunc = LilImageAlloc;
		outLilMemDesc->imageFreeFunc = LilImageFree;
		outLilMemDesc->imageWriteFunc = LilImageWrite;
		outLilMemDesc->imageReadFunc = LilImageRead;
		outLilMemDesc->mem = 0;
	}

} // namespace



//-----------------------------------------------------------------------------------------------------------
// IMAGE IO
//-----------------------------------------------------------------------------------------------------------
LilErrorType LoadWxImage(const char *path, wxImage *output) {
    LilErrorType err = LE_NONE;

    FILE *file = fopen(path, "rb");
    if (!file)
        return LE_IOERROR;
    LilIoDesc io;
    LilMemDesc mem;
    LilFormatDesc format;
    GetLilIoFile(file, &io);
    GetLilMemWx(LilMemRealloc, 0, &mem);
    err = LilUtilSetFormat(&format, LDF_SIMPLE, 8, 3, 1, false, LCO_RGB, false);
    if (err != LE_NONE)
        return err;

    LilFileFormatType fileFormat = LilUtilGetFileFormat(path, &io);
    LilImageDesc image;
    LilImageInit(&image, &mem, output);
    err = LilImageLoad(fileFormat, &io, &image, LL_DATA_AND_INFO, &format);

    LilImageFree(&image);
    fclose(file);

    return err;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType SaveWxImage(const char *path, const wxImage *input, const LilFormatDesc *format) {
    LilErrorType err = LE_NONE;

    FILE *file = fopen(path, "wb");
    if (!file)
        return LE_IOERROR;
    LilIoDesc io;
    LilMemDesc mem;
    GetLilIoFile(file, &io);
    GetLilMemWx(LilMemRealloc, 0, &mem);

    LilFileFormatType fileFormat = LilUtilGetFileFormat(path, 0);
    LilImageDesc image;
    LilImageInit(&image, &mem, const_cast<wxImage *>(input));
    LilUtilSetFormat(&image.format, LDF_SIMPLE, 8, 3, 1, false, LCO_RGB, false);
    image.palette.numEntries = 256;
    LilUtilSetFormat(&image.palette.format, LDF_SIMPLE, 8, 3, 4, false, LCO_RGB, false);
    lil_u8 data[768];
    for (int i = 0; i < 768;) {
        int val = i / 3;
        data[i++] = 255;
        data[i++] = val;
        data[i++] = val;
    }
    data[3] = 0;  // Make entry 1 blue for mono images
    data[4] = 0;
    data[5] = 255;
    data[6] = 0;
    data[7] = 255;
    data[8] = 0;
    data[9] = 128;
    data[10] = 64;
    data[11] = 255;

    LilImageSetPalette(&image, data);
    image.data.width = input->GetWidth();
    image.data.height = input->GetHeight();
    err = LilImageSave(fileFormat, &io, &image, format);

    LilImageFree(&image);
    fclose(file);

    return err;
}

