#include "Lil.h"
#include "LilMem.h"
#include "LilConvert.h"
#include "LilBitManip.h"
#include "LilUtilFormat.h"
#include "LilPlatform.h"
#include "LilCodec.h"

//-----------------------------------------------------------------------------------------------------------
namespace {
	struct FileFormatFuncs {
		typedef bool (* IsExtensionFunc)(const char *fileName);
		typedef bool (* IsDetectableFunc)(const LilIoDesc *io);
		typedef LilErrorType (* LoadFunc)(const LilIoDesc *io, LilImageDesc *image, LilLoadType request, 
			const LilFormatDesc *imageFormat);
		typedef LilErrorType (* SaveFunc)(const LilIoDesc *io, const LilImageDesc *image, const LilFormatDesc *imageFormat);
		
		FileFormatFuncs() : IsExtension(0), IsDetectable(0), Load(0), Save(0) { }
		FileFormatFuncs(IsExtensionFunc isExtension, IsDetectableFunc isDetectable, LoadFunc load, SaveFunc save) : 
			IsExtension(isExtension), IsDetectable(isDetectable), Load(load), Save(save) { }
		IsExtensionFunc IsExtension;
		IsDetectableFunc IsDetectable;
		LoadFunc Load;
		SaveFunc Save;
	};

	bool initialized = false;

	FileFormatFuncs fileFormats[LFF_COUNT];

	FileFormatFuncs *GetFileFormatObject(LilFileFormatType fileFormat) {
		if (fileFormat == LFF_NONE || !fileFormats[fileFormat].Load)
			return 0;
		return &fileFormats[fileFormat];
	}
}

//-----------------------------------------------------------------------------------------------------------
namespace lil {
	#define DECLARE_FILE_FORMAT(name, capsName) \
		bool IsExtension##name(const char *fileName); \
		bool IsDetectable##name(const LilIoDesc *io); \
		LilErrorType Load##name(const LilIoDesc *io, LilImageDesc *image, LilLoadType request, \
			const LilFormatDesc *imageFormat); \
		LilErrorType Save##name(const LilIoDesc *io, const LilImageDesc *image, const LilFormatDesc *imageFormat);

	#ifdef LIL_SUPPORTS_BMP
	DECLARE_FILE_FORMAT(Bmp, BMP);
	#endif
	#ifdef LIL_SUPPORTS_TGA
	DECLARE_FILE_FORMAT(Tga, TGA);
	#endif
	#ifdef LIL_SUPPORTS_PSD
	DECLARE_FILE_FORMAT(Psd, PSD);
	#endif
	#ifdef LIL_SUPPORTS_TIF
	DECLARE_FILE_FORMAT(Tif, TIF);
	#endif
	#ifdef LIL_SUPPORTS_JPG
	DECLARE_FILE_FORMAT(Jpg, JPG);
	#endif
	#ifdef LIL_SUPPORTS_PCT
	DECLARE_FILE_FORMAT(Pct, PCT);
	#endif
	#ifdef LIL_SUPPORTS_GIF
	DECLARE_FILE_FORMAT(Gif, GIF);
	#endif
	#ifdef LIL_SUPPORTS_PNG
	DECLARE_FILE_FORMAT(Png, PNG);
	#endif
	#ifdef LIL_SUPPORTS_ICO
	DECLARE_FILE_FORMAT(Ico, ICO);
	#endif
	#ifdef LIL_SUPPORTS_CUR
	DECLARE_FILE_FORMAT(Cur, CUR);
	#endif
	#ifdef LIL_SUPPORTS_PNM
	DECLARE_FILE_FORMAT(Pnm, PNM);
	#endif
	#undef DECLARE_FILE_FORMAT
}

//-----------------------------------------------------------------------------------------------------------
LIL_API void LIL_CC LilLibInit() {
	if (initialized)
		return;

	initialized = true;

	#define SET_FILE_FORMAT(name, nameCaps) fileFormats[LFF_##nameCaps] = \
		FileFormatFuncs(lil::IsExtension##name, lil::IsDetectable##name, lil::Load##name, lil::Save##name)
	#ifdef LIL_SUPPORTS_BMP
	SET_FILE_FORMAT(Bmp, BMP);
	#endif
	#ifdef LIL_SUPPORTS_TGA
	SET_FILE_FORMAT(Tga, TGA);
	#endif
	#ifdef LIL_SUPPORTS_PSD
	SET_FILE_FORMAT(Psd, PSD);
	#endif
	#ifdef LIL_SUPPORTS_TIF
	SET_FILE_FORMAT(Tif, TIF);
	#endif
	#ifdef LIL_SUPPORTS_JPG
	SET_FILE_FORMAT(Jpg, JPG);
	#endif
	#ifdef LIL_SUPPORTS_PCT
	SET_FILE_FORMAT(Pct, PCT);
	#endif
	#ifdef LIL_SUPPORTS_GIF
	SET_FILE_FORMAT(Gif, GIF);
	#endif
	#ifdef LIL_SUPPORTS_PNG
	SET_FILE_FORMAT(Png, PNG);
	#endif
	#ifdef LIL_SUPPORTS_ICO
	SET_FILE_FORMAT(Ico, ICO);
	#endif
	#ifdef LIL_SUPPORTS_CUR
	SET_FILE_FORMAT(Cur, CUR);
	#endif
	#undef SET_FILE_FORMAT

	// Default configuration
	LilConfigDesc desc;
	memset(&desc, 0, sizeof(LilConfigDesc));
	desc.coeffRgbToLumin[0] = .2126f;
	desc.coeffRgbToLumin[1] = .7152f;
	desc.coeffRgbToLumin[2] = .0722f;
	desc.convertBits = LCB_LINEAR_SCALE;
	desc.convertPalette = LCP_USE_PALETTE;
	LilLibSetOptions(&desc);
}

//-----------------------------------------------------------------------------------------------------------
LIL_API void LIL_CC LilLibExit() {
	if (!initialized)
		return;

	initialized = false;
}

//-----------------------------------------------------------------------------------------------------------
LIL_API void LIL_CC LilLibSetOptions(const LilConfigDesc *config) {
	LilLibInit();  // Since LilInit calls this function again...
	memcpy(&lil::globalConfig, config, sizeof(LilConfigDesc));
}

//-----------------------------------------------------------------------------------------------------------
LIL_API LilFileFormatType LIL_CC LilUtilGetFileFormat(const char *fileName, const LilIoDesc *io) {
	LilLibInit();

	if (fileName) {
		for (lil_u32 i = 0; i < LIL_ARRAY_SIZE(fileFormats); i++) {
			if (fileFormats[i].IsExtension && fileFormats[i].IsExtension(fileName))
				return static_cast<LilFileFormatType>(i);
		}
	}
	if (io) {
		for (lil_u32 i = 0; i < LIL_ARRAY_SIZE(fileFormats); i++) {
			if (fileFormats[i].IsDetectable && fileFormats[i].IsDetectable(io))
				return static_cast<LilFileFormatType>(i);
		}
	}
	return LFF_NONE;
}

//-----------------------------------------------------------------------------------------------------------
LIL_API LilErrorType LIL_CC LilUtilSetFormat(LilFormatDesc *pFormat, LilDataFormatType dataFormat, lil_u8 bitsPerChannel,
	lil_u8 numChannels, lil_u8 blockLineByteAlign, lil_bool planar, LilChannelOrderType channelOrder, lil_bool flipped)
{
    LilErrorType err = LE_NONE;
	lil::SetFormat(pFormat, dataFormat, bitsPerChannel, numChannels, blockLineByteAlign, planar, channelOrder,
        flipped, &err);
	lil::SimplifyFormat(pFormat);
    return err;
}

//-----------------------------------------------------------------------------------------------------------
LIL_API void LIL_CC LilImageInit(LilImageDesc *image, LilMemDesc *memFuncs, lil_image_handle imageHandle) {
	memset(image, 0, sizeof(LilImageDesc));
	image->memFuncs = *memFuncs;
	image->data.data = imageHandle;
}

//-----------------------------------------------------------------------------------------------------------
LIL_API void LIL_CC LilImageFree(LilImageDesc *image) {
	// User's responsibility... We probably want to free the LilImageDesc image and be done with
	// the library, but keep the image that was actually loaded for further use.
	//image->memFuncs.imageFreeFunc(image->data.data);

	// Temporary until everything goes through mem pool
	image->memFuncs.memReallocFunc(&image->extended.data, 0);
	image->memFuncs.memReallocFunc(&image->palette.lut, 0);

//	lil::MemPoolFree(&image->memFuncs);
	memset(image, 0, sizeof(LilImageDesc));
}

//-----------------------------------------------------------------------------------------------------------
LIL_API LilErrorType LIL_CC LilImageLoad(LilFileFormatType fileFormat, const LilIoDesc *io, LilImageDesc *image,
	LilLoadType request, const LilFormatDesc *imageFormat)
{
	LilErrorType err = LE_NONE;

	LilLibInit();

	if (!image->memFuncs.memReallocFunc || !image->memFuncs.imageAllocFunc || !image->memFuncs.imageReadFunc)
		return LE_NEEDINITIMAGE;

	if (imageFormat && !lil::IsFormatValid(imageFormat, &err))
		return err;

	FileFormatFuncs *pLoader = GetFileFormatObject(fileFormat);
	if (!pLoader)
		return LE_BADFILEFORMAT;

	err = pLoader->Load(io, image, request, imageFormat);

	return err;
}

//-----------------------------------------------------------------------------------------------------------
LIL_API LilErrorType LIL_CC LilImageSave(LilFileFormatType fileFormat, const LilIoDesc *io,
	const LilImageDesc *image, const LilFormatDesc *imageFormat)
{
	LilErrorType err = LE_NONE;

	LilLibInit();

	if (!image->memFuncs.memReallocFunc || !image->memFuncs.imageAllocFunc || !image->memFuncs.imageWriteFunc)
		return LE_NEEDINITIMAGE;

	if (imageFormat && !lil::IsFormatValid(imageFormat, &err))
		return err;
	if (!lil::IsFormatValid(&image->format, &err))
		return err;

	FileFormatFuncs *pSaver = GetFileFormatObject(fileFormat);
	if (!pSaver)
		return LE_BADFILEFORMAT;

	err = pSaver->Save(io, image, imageFormat);

	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
LIL_API lil_u32 LIL_CC LilImageGetBpp(const LilImageDesc *image) {
	return lil::CalcBpp(&image->format);
}

//-----------------------------------------------------------------------------------------------------------
LIL_API lil_u32 LIL_CC LilImageGetBpl(const LilImageDesc *image) {
	return lil::CalcBpl(&image->format, image->data.width);
}

//-----------------------------------------------------------------------------------------------------------
LIL_API lil_u64 LIL_CC LilImageGetByteSize(const LilImageDesc *image) {
	return lil::CalcByteSize(&image->format, image->data.width, image->data.height);
}

//-----------------------------------------------------------------------------------------------------------
LIL_API LilErrorType LIL_CC LilImageConvertPalette(LilImageDesc *image, LIL_CONST LilFormatDesc *format) {
	LilErrorType err = LE_NONE;
    LilPaletteDesc newPal;

	newPal.format = *format;
    newPal.lut = 0;
    newPal.numEntries = 0;
	err = lil::ConvertPalette(image, &newPal, 0);
	if (err != LE_NONE)
		return err;

	image->memFuncs.memReallocFunc(&image->palette.lut, 0);
	image->palette.lut = newPal.lut;
	image->palette.format = *format;

	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
LIL_API LilErrorType LIL_CC LilImageSetPalette(LilImageDesc *image, const void *data) {
	size_t size = static_cast<size_t>(lil::CalcByteSize(&image->palette.format, image->palette.numEntries, 1));
    LilErrorType err = image->memFuncs.memReallocFunc(&image->palette.lut, size);
	if (err != LE_NONE)
		return err;
    memcpy(image->palette.lut, data, size);
	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
LIL_API void LIL_CC LilImageGetPalette(LIL_CONST LilImageDesc *image, void *data) {
	lil_u64 size = lil::CalcByteSize(&image->palette.format, image->palette.numEntries, 1);
    memcpy(data, image->palette.lut, static_cast<size_t>(size));
}

//-----------------------------------------------------------------------------------------------------------
LIL_API LilErrorType LIL_CC LilEncodeZlib(LilIoDesc *ioOut, LIL_CONST void *dataIn, lil_u32 numBytes,
	LilCompressionLevelType level, LilMemDesc *mem)
{
	return lil::EncodeZlib(dataIn, ioOut, numBytes, level, mem);
}

//-----------------------------------------------------------------------------------------------------------
LIL_API LilErrorType LIL_CC LilDecodeZlib(void *dataOut, LIL_CONST LilIoDesc *ioIn, LilMemDesc *mem) {
	return lil::DecodeZlib(dataOut, ioIn, 0, mem);
}

//-----------------------------------------------------------------------------------------------------------
// Support for windows dlls
#ifdef LIL_OS_WIN
#define WIN32_LEAN_AND_MEAN
#define VC_EXTRALEAN
#include <windows.h>

BOOL APIENTRY DllMain(HMODULE hModule, DWORD ul_reason_for_call, LPVOID lpReserved) {
	LIL_UNUSED_PARAM(lpReserved);
	LIL_UNUSED_PARAM(hModule);
	switch (ul_reason_for_call) {
	case DLL_PROCESS_ATTACH:
		LilLibInit();
		break;
	case DLL_THREAD_ATTACH:
		break;
	case DLL_THREAD_DETACH:
		break;
	case DLL_PROCESS_DETACH:
		LilLibExit();
		break;
	}
	return TRUE;
}
#endif
