#ifndef LIL_H_INCLUDED
#define LIL_H_INCLUDED

#include <stddef.h>

//-----------------------------------------------------------------------------------------------------------
// COMPILATION SETUP
//-----------------------------------------------------------------------------------------------------------
// OS
#if (defined(_WIN32) || defined(_WIN64) || defined(WIN32) || defined(WIN64) || defined(_WIN32_WCE))
	#define LIL_OS_WIN
#endif

// COMPILER
#if (defined(_MSC_VER))
	#define LIL_COMPILER_VS
#elif (defined(__GNUC__))
	#define LIL_COMPILER_GNU
#endif

// ENDIAN
#if 0
	#define LIL_ENDIAN_BIG
#else
	#define LIL_ENDIAN_LITTLE
#endif

// LANGUAGE
#ifdef __cplusplus
	#define LIL_LANG_CPP
#endif

// DLL CREATION
#ifdef LIL_OS_WIN
	#ifdef LIL_EXPORTS
		#ifdef LIL_COMPILER_VS
			#define LIL_API __declspec(dllexport)
		#elif defined(LIL_COMPILER_GNU)
            #define LIL_API extern
        #else
			#error Need support for dll export for other compilers.
		#endif
	#else
		#define	LIL_API extern
	#endif
#else
	#define	LIL_API extern
#endif

// CALLING CONVENTION
#ifdef LIL_COMPILER_VS
    #define LIL_CC __cdecl
#elif defined(LIL_COMPILER_GNU)
    #define LIL_CC
#endif

// C++ CONST
// (does c99 have const? - maybe we should detect this case if it's true)
#if defined(LIL_LANG_CPP) || !defined(__STDC_VERSION__) || __STDC_VERSION__ < 198900L
	#define LIL_CONST const
#else
	#define LIL_CONST
#endif


//-----------------------------------------------------------------------------------------------------------
// LIL CUSTOMIZATION
// Anything that customizes the library should go here.
//-----------------------------------------------------------------------------------------------------------
// Note these MUST match the way the library is compiled if it is being linked in binary form.
#define LIL_MAX_CHANNELS 8


//-----------------------------------------------------------------------------------------------------------
// BASIC DATA TYPES
//-----------------------------------------------------------------------------------------------------------
typedef unsigned char lil_u8;
typedef signed char lil_i8;
typedef unsigned short lil_u16;
typedef signed short lil_i16;
typedef unsigned int lil_u32;
typedef signed int lil_i32;
typedef unsigned long long lil_u64;
typedef signed long long lil_i64;
typedef int lil_bool;
static const lil_bool lil_true = 1;
static const lil_bool lil_false = 0;

typedef void * lil_mem_handle;
typedef void * lil_io_handle;
typedef void * lil_image_handle;


//-----------------------------------------------------------------------------------------------------------
// LIL DATA TYPES
//-----------------------------------------------------------------------------------------------------------
typedef enum LilErrorType {
	LE_NONE = 0,
	LE_CANCEL,                    // Cancel command came in from callback
	LE_NOTENOUGHMEMORY,
	LE_NOMEMBUFFER,
	LE_INTERNALERROR,
	LE_NOIMAGE,
	LE_NEEDINITIMAGE,             // Image needs initialized with LilInitImage or was corrupted some how

	// Errors in file formats rendering them unreadable.  These come from data input (not user).
	LE_BADFILEFORMAT,             // The file itself is just wrong
	LE_BADHEADER,
	LE_BADBPP,
	LE_BADORIENTATION,
	LE_BADCHANNELORDER,
	LE_BADCOLORSPACE,
	LE_BADCOMPRESSION,
	LE_BADIMAGEDATA,              // Could mean, for instance, not enough or too many bytes read.
	LE_BADMETADATA,
	LE_BADCRC,
	LE_BADPALETTE,
	LE_BADIMAGESIZE,

	// File format characteristics that LIL either doesn't handle yet (usually while loading),
	// or simply isn't possible in the file format itself (usually while saving).
	LE_UNSUPPORTEDFILEFEATURE,    // Must be something exotic that isn't in the baseline format!
	LE_UNSUPPORTEDFORMAT,         // More general... SOMETHING in the format isn't right.
	LE_UNSUPPORTEDPALETTEFORMAT,
	LE_UNSUPPORTEDORIENTATION,
	LE_UNSUPPORTEDFILEFORMAT,
	LE_UNSUPPORTEDCOLORSPACE,
	LE_UNSUPPORTEDCOMPRESSION,
	LE_UNSUPPORTEDVALUETYPE,      // Float or something... should be supported in future though.
	LE_UNSUPPORTEDVERSION,        // The file is likely a newer version than we support currently.
	LE_UNSUPPORTEDSIZE,
	LE_UNSUPPORTEDCSCONVERSION,   // The requested colorspace conversion isn't implemented

	// Format errors - impossible combinations in format desc.  These come from user input!
	LE_INVALIDFORMAT,             //
	LE_INVALIDPALETTEFORMAT,
	LE_INVALIDPAGE,               // As in, image index, page index, layer index, etc.

	// Format conversion errors - things LIL doesn't do.  These are left to an image
	// library (perhaps one that uses LIL for image i/o) because there are just too many ways
	// to do it, and different software and different libraries will want to do it their way.
	LE_NOPALETTESHRINKING,        // Many dithering options.  Let an image library do it.
	LE_UNSUPPORTEDCONVERSION,     // Generic error.  The cause of this will likely be fixed.

	// IO errors
	LE_IOERROR,                   // Generic error
	LE_IOSEEKUNSUPPORTED,         // Seek unsupported
	LE_IOOUTOFRANGE,              // Operation seek/read/etc out of range
} LilErrorType;

//-----------------------------------------------------------------------------------------------------------
typedef enum LilFileFormatType {
	LFF_NONE = -1,
	LFF_BMP = 0,
	LFF_ICO,
	LFF_CUR,
	LFF_TIF,
	LFF_JPG,
	LFF_PNG,
	LFF_PSD,
	LFF_TGA,
	LFF_GIF,
	LFF_PCT,
	LFF_COUNT
} LilFileFormatType;

//-----------------------------------------------------------------------------------------------------------
typedef enum LilLoadType {
	LL_DATA_AND_INFO = 0,
	LL_INFO
} LilLoadType;

//-----------------------------------------------------------------------------------------------------------
// A note about data format.  Whatever order the data is in will be by channel.  This goes for
// simple formats, float formats, 16 bit formats, etc.  So BGR on an 8 bit 4 channel
// simple formatted image indicates data in the form BB GG RR AA in memory.  The endianess
// of the machine is irrelevant.  On little endian machines, a dword of a 4 channel 8 bit
// image (32 bpp) in BGR order would be 0xAARRGGBB.  On big endian machines it would be 0xBBGGRRAA.
// So for windows, for instance, to be GDI compatible, you would actually want RGB order.
// GDI expects the 32 bit dwords to be in the form 0xAABBGGRR.  In memory, since
// windows machines are little endian machines, this is RR GG BB AA RR GG BB AA - so Lil considers
// this to be RGB order.
// (This is why people get confused and think windows colors are bgr ordered).
typedef enum LilDataFormatType {
	LDF_SIMPLE = 0,  // bitsPerChannel == X, numChannels == Y, integer or packed integer type
//	LDF_FLOAT,       // bitsPerChannel == X, numChannels == Y, float type
	LDF_SIMPLEBITS,  // bitsPerChannel == 0, numChannels == Y, packed integer type with varying bit counts
} LilDataFormatType;

//-----------------------------------------------------------------------------------------------------------
typedef enum LilChannelOrderType {
	LCO_UNKNOWN = 0, // Anything.  User can only map to another User in conversion.
	LCO_SINGLE,   // Either grayscale or palette look up, depending if a palette is defined
	LCO_YA,       // Luminosity-Alpha
	LCO_RGB,      // R-G-B order like Mac
	LCO_BGR,      // B-G-R order like Windows
	LCO_ARGB,     // A-R-G-B order like Mac
	LCO_ABGR,     // A-B-G-R
	LCO_RGBA,     // R-G-B-A
	LCO_BGRA,     // B-G-R-A order like Windows
	LCO_CMY,      // C-M-Y
	LCO_CMYA,     // C-M-Y-Alpha
	LCO_CMYK,     // C-M-Y-K
	LCO_CMYKA,    // C-M-Y-K-Alpha
	LCO_YCBCR,    // Y-CB-CR
	LCO_YCBCRA,   // Y-CB-CR-Alpha
} LilChannelOrderType;

//-----------------------------------------------------------------------------------------------------------
typedef enum LilConvertPaletteType {
	LCP_USE_PALETTE = 0,
	LCP_NO_PALETTE
} LilConvertPaletteType;

//-----------------------------------------------------------------------------------------------------------
typedef enum LilConvertBitsType {
	LCB_LINEAR_SCALE = 0, // Maps maximum to maximum, and minimum to minimum.  Ie 255 to 65535, if 8 to 16 bit.  etc.
	LCB_NO_SCALE          // Directly copies values.  255 in 8 bit is converted to 255 in 16 or 12 bit.  etc.
} LilConvertBitsType;

//-----------------------------------------------------------------------------------------------------------
// For when you want this kind of control.
// NOTE that if you choose a compression type that doesn't support the image format (say, LC_BMP_RLE4
// for an 8 bit image), it will fail... EVEN if the file format (.bmp) supports the image format (8 bpp)
// with another compression type.
// NOTE that these numbers do not necessarily indicate the same value that is in the file format - but
// only the options available in LIL in an (somewhat) organized way.
typedef enum LilCompressionType {
	LC_ANY = 0,      // Let LIL choose
	LC_BMP_NONE = 1, LC_BMP_RLE8, LC_BMP_RLE4, LC_BMP_BITFIELDS, LC_BMP_JPEG, LC_BMP_PNG,
	LC_ICO_NONE = 1, LC_ICO_PNG,
	LC_CUR_NONE = 1,
	LC_TGA_NONE = 1, LC_TGA_RLE, LC_TGA_ZIP, LC_TGA_ZIPPRED,
	LC_PSD_NONE = 1, LC_PSD_PACKBITS, LC_PSD_ZIPWO, LC_PSD_ZIPWITH,
	LC_JPG_BASELINE = 1,
//	LC_EPS_NONE = 1, LC_EPS_TIF, LC_EPS_WMF, LC_EPS_PICT, LC_EPS_EPSI,
	LC_PCT_NONE = 1, LC_PCT_PACKBITS, LC_PCT_JPEG,
	LC_GIF_LZW = 1,
	LC_PNG_ZIP = 1,
	LC_PNM_NONE = 1, LC_PNM_BINARY,
} LilCompressionType;

//-----------------------------------------------------------------------------------------------------------
typedef enum LilCompressionLevelType {
	LCL_NONE = 0,
	LCL_FAST = 1,
	LCL_NORMAL = 2,
	LCL_MAX = 3,
} LilCompressionLevelType;

//-----------------------------------------------------------------------------------------------------------
typedef enum LilPngFilterType {
	LPF_NONE = 0,
	LPF_SUB,
	LPF_UP,
	LPF_AVG,
	LPF_PAETH,
	LPF_CHOOSE,  // uses sum of abs heuristic
} LilPngFilterType;

//-----------------------------------------------------------------------------------------------------------
typedef enum LilSeekType {
	LS_SET = 0,
	LS_RELATIVE,
	LS_END
} LilSeekType;

//-----------------------------------------------------------------------------------------------------------
typedef enum LilDpuType {
	LD_UNKNOWN,
	LD_METERS,
	LD_CENTIMETERS,
	LD_INCHES
} LilDpuType;

//-----------------------------------------------------------------------------------------------------------
typedef struct LilDataDesc {
	lil_image_handle data;
	lil_u32 width;
	lil_u32 height;
} LilDataDesc;

//-----------------------------------------------------------------------------------------------------------
typedef struct LilFormatDesc {
	LilDataFormatType dataFormat;      // How to interpret each pixel or data
	lil_u8 bitsPerChannel;                 // The number of bits per channel of pixel or data - if consistent
	lil_u8 bitCounts[LIL_MAX_CHANNELS];    // The number of bits per channel of pixel or data - if variant
	lil_u8 numChannels;                    // The number of channels per pixel or data
	lil_u8 blockLineByteAlign;             // The byte alignment for a single horizontal line of data, per block
	lil_bool planar;                       // true if the data is planar vs. chunky
	LilChannelOrderType channelOrder;  // Order of channels
	lil_bool flipped;                      // true if it is upside down.
} LilFormatDesc;

//-----------------------------------------------------------------------------------------------------------
typedef struct LilPaletteDesc {
	lil_mem_handle lut;
	lil_u32 numEntries;
	LilFormatDesc format;
} LilPaletteDesc;

//-----------------------------------------------------------------------------------------------------------
typedef struct LilTextDesc {
	// This structure is in relation to an arbitrary lil_mem_handle
	lil_u64 nameLength;  // In bytes
	lil_u64 nameOffset;  // In bytes
	lil_u64 textLength;  // In bytes
	lil_u64 textOffset;  // In bytes
} LilTextDesc;

//-----------------------------------------------------------------------------------------------------------
typedef struct LilExtendedDesc {
	lil_u64 textFieldsOffset;  // In bytes
	lil_u32 numTextFields;
	lil_u32 pageIndex;  // All image formats have pages, even if it's implicitly just one.
	lil_u32 numPages;
	LilCompressionType compression;
	lil_u32 dpuX, dpuY;
	LilDpuType dotsPerWhat;

	union {
		//struct {
		//} bmp;
		struct {
			lil_bool interlaced;
			lil_u8 bgColorIndex;
		} gif;
		struct {
			lil_bool png;
		} ico;
		struct {
			lil_u8 hotPointX;
			lil_u8 hotPointY;
		} cur;
		struct {
			lil_u8 subsampling[3];  // 4:4:4, 4:2:2, 4:2:1, 4:1:1, 4:0:0:, 4:2:0, etc.
		} jpg;
		struct {
			LilPngFilterType filter;
			lil_bool interlaced;
		} png;
		struct {
			lil_u64 duotoneDataSize;  // In bytes
			lil_u64 duotoneDataOffset;  // In bytes
		} psd;
		//struct {
		//} tga;
		//struct {
		//} tif;
	} fileFormat;

	lil_u64 dataSize;  // In bytes
	lil_mem_handle data;  // Generic.  Used to hold all variable-length metadata.
} LilExtendedDesc;

//-----------------------------------------------------------------------------------------------------------
typedef LilErrorType (* LilIoReadFunc)(lil_io_handle, size_t, void *);
typedef LilErrorType (* LilIoWriteFunc)(lil_io_handle, size_t, LIL_CONST void *);
typedef LilErrorType (* LilIoSeekFunc)(lil_io_handle, lil_i64, LilSeekType);
typedef lil_u64 (* LilIoTellFunc)(lil_io_handle);

struct LilImageDesc;
typedef struct LilImageDesc LilImageDesc;
typedef LilErrorType (* LilMemReallocFunc)(lil_mem_handle *, lil_u64);
typedef void * (* LilMemViewFunc)(lil_mem_handle, lil_u64, lil_u64);
typedef LilErrorType (* LilImageAllocFunc)(lil_image_handle, LIL_CONST LilImageDesc *);
typedef LilErrorType (* LilImageFreeFunc)(lil_image_handle);
typedef LilErrorType (* LilImageWriteFunc)(lil_image_handle, LIL_CONST void *, lil_u32, lil_u32, lil_u32, lil_u32, lil_u32, lil_u32);
typedef LilErrorType (* LilImageReadFunc)(lil_image_handle, void *, lil_u32, lil_u32, lil_u32, lil_u32, lil_u32, lil_u32);

//-----------------------------------------------------------------------------------------------------------
typedef struct LilIoDesc {
	LilIoReadFunc readFunc;
	LilIoWriteFunc writeFunc;
	LilIoSeekFunc seekFunc;
	LilIoTellFunc tellFunc;
	lil_io_handle handle;
} LilIoDesc;

//-----------------------------------------------------------------------------------------------------------
typedef struct LilMemDesc {
	lil_mem_handle mem;
	LilMemReallocFunc memReallocFunc;
	LilImageAllocFunc imageAllocFunc;
	LilImageFreeFunc imageFreeFunc;
	LilImageWriteFunc imageWriteFunc;
	LilImageReadFunc imageReadFunc;
} LilMemDesc;

//-----------------------------------------------------------------------------------------------------------
struct LilImageDesc {
	LilDataDesc data;
	LilPaletteDesc palette;
	LilFormatDesc format;
	LilExtendedDesc extended;
	LilMemDesc memFuncs;
};

//-----------------------------------------------------------------------------------------------------------
typedef struct LilConfigDesc {
	float coeffRgbToLumin[3];  // Default: .2126, .7152, .0722
//	float coeffRgbToCmy[3];  // Default:
//	float coeffRgbToCmyk[3];  // Default:
//	float coeffRgbToYcbcr[9];  // Default:

	LilConvertPaletteType convertPalette;  // Default: LCP_USE_PALETTE
	LilConvertBitsType convertBits;  // Default: LCB_LINEAR_SCALE
} LilConfigDesc;


//-----------------------------------------------------------------------------------------------------------
// FUNCTIONS
//-----------------------------------------------------------------------------------------------------------
// Let's be as compatible as possible.
#ifdef LIL_LANG_CPP
extern "C" {
#endif

// *** LIB FUNCS ***
// These are unnecessary if this is a dll being automatically loaded
LIL_API void LIL_CC                LilLibInit();
LIL_API void LIL_CC                LilLibExit();

// Set global library options - see defaults in LilConfigDesc structure.
LIL_API void LIL_CC                LilLibSetOptions(LIL_CONST LilConfigDesc *config);


// *** UTILITY ***
// Shortcut to set a LilFormatDesc structure
LIL_API LilErrorType LIL_CC        LilUtilSetFormat(LilFormatDesc *pFormat, LilDataFormatType dataFormat, lil_u8 bitsPerChannel,
	                                                lil_u8 numChannels, lil_u8 blockLineByteAlign, lil_bool planar, 
												    LilChannelOrderType channelOrder, lil_bool flipped);
// Get the file format of an image if possible
LIL_API LilFileFormatType LIL_CC   LilUtilGetFileFormat(LIL_CONST char *fileName, LIL_CONST LilIoDesc *io);


// *** IMAGE FUNCS ***
// Prepare an image description/free internal buffers for that image description
LIL_API void LIL_CC                LilImageInit(LilImageDesc *image, LilMemDesc *memFuncs, lil_image_handle imageHandle);
LIL_API void LIL_CC                LilImageFree(LilImageDesc *image);

// Load/save an image
LIL_API LilErrorType LIL_CC        LilImageLoad(LilFileFormatType fileFormat, LIL_CONST LilIoDesc *io, LilImageDesc *image,
												LilLoadType request, LIL_CONST LilFormatDesc *imageFormat);
LIL_API LilErrorType LIL_CC        LilImageSave(LilFileFormatType fileFormat, LIL_CONST LilIoDesc *io, 
												LIL_CONST LilImageDesc *image, LIL_CONST LilFormatDesc *imageFormat);
// Get/Set properties of a LilImageDesc
LIL_API lil_u32 LIL_CC             LilImageGetBpp(LIL_CONST LilImageDesc *image);
LIL_API lil_u32 LIL_CC             LilImageGetBpl(LIL_CONST LilImageDesc *image);
LIL_API lil_u64 LIL_CC             LilImageGetByteSize(LIL_CONST LilImageDesc *image);
LIL_API void LIL_CC                LilImageGetPalette(LIL_CONST LilImageDesc *image, void *data);
LIL_API LilErrorType LIL_CC        LilImageSetPalette(LilImageDesc *image, LIL_CONST void *data);
LIL_API LilErrorType LIL_CC        LilImageConvertPalette(LilImageDesc *image, LIL_CONST LilFormatDesc *format);


// *** CODEC ***
LIL_API LilErrorType LIL_CC        LilEncodeZlib(LilIoDesc *ioOut, LIL_CONST void *dataIn, lil_u32 numBytes,
                                                 LilCompressionLevelType level, LilMemDesc *mem);
LIL_API LilErrorType LIL_CC        LilDecodeZlib(void *dataOut, LIL_CONST LilIoDesc *ioIn, LilMemDesc *mem);


#ifdef LIL_LANG_CPP
}
#endif

#endif /* LIL_H_INCLUDED */
