#ifndef LILBITMANIP_H_INCLUDED
#define LILBITMANIP_H_INCLUDED

#include <algorithm>
#include "Lil.h"
#include "LilPlatform.h"

namespace lil {

// Bit stream (use the concrete types)
template <class T> struct BitStreamT {
	T *pBase;
	T *pPos;
	lil_u32 hasBits;
	lil_u32 currBits;
	lil_u32 bytesLeft;
};
typedef BitStreamT<lil_u8> BitStreamOut;
typedef BitStreamT<const lil_u8> BitStreamIn;

//-------------------------------------------------------------------------------------------------------
// DECLARATIONS
//-------------------------------------------------------------------------------------------------------
lil_u32 UpdateCrc32(lil_u32 crc, const lil_u8 *buf, size_t len);
lil_u32 CalcCrc32(const lil_u8 *buf, size_t len);
lil_u32 UpdateAdler32(lil_u32 adler, const lil_u8 *buf, size_t len);
lil_u32 CalcAdler32(const lil_u8 *buf, size_t len);

// Bit and byte manipulation
inline lil_u32 BitSet(lil_u32 bits, lil_u32 mask);
inline lil_u32 BitReset(lil_u32 bits, lil_u32 mask);
inline lil_u32 BitFlip(lil_u32 bits, lil_u32 mask);
inline lil_u32 BitToggle(lil_u32 bits, lil_u32 mask, bool toggle);
inline lil_u8 BitReverse8(lil_u8 value);
inline lil_u16 BitReverse16(lil_u16 value);
inline lil_u32 BitReverse32(lil_u32 value);
inline int BitGetCount(lil_u32 value);
inline int BitGetMsb(lil_u32 value);
inline int BitGetLsb(lil_u32 value);
inline void BitInvertBuf(void *pData, lil_u32 numBytes);
inline void BitReverseBuf8(void *pData, lil_u32 numBytes);
inline void ByteSwap16(lil_u16 &value);
inline void ByteSwap32(lil_u32 &value);
inline void ByteSwap64(lil_u64 &value);
inline void ByteSwapBuf16(void *pData, lil_u32 count); // count in words, not bytes
inline void ByteSwapBuf32(void *pData, lil_u32 count); // count in double words, not bytes

// Bit streams
inline void BsInit(BitStreamIn *pBits, const void *pBase, lil_u32 byteSize);
inline void BsInit(BitStreamOut *pBits, void *pBase, lil_u32 byteSize);
inline bool BsReadLsb(BitStreamIn *pBits, lil_u32 *pOut, lil_u8 numBits);
inline bool BsReadMsb(BitStreamIn *pBits, lil_u32 *pOut, lil_u8 numBits);
inline void BsReadFlush(BitStreamIn *pBits);  // Flush partial byte if there is one
inline bool BsWriteLsb(BitStreamOut *pBits, lil_u32 in, lil_u8 numBits);
inline bool BsWriteMsb(BitStreamOut *pBits, lil_u32 in, lil_u8 numBits);
inline void BsWriteFlushLsb(BitStreamOut *pBits);  // Flush partial byte if there is one
inline void BsWriteFlushMsb(BitStreamOut *pBits);  // Flush partial byte if there is one
template <class BitStream> inline void BsSetPosLsb(BitStream *pBits, lil_u32 bitPos);
template <class BitStream> inline void BsSetPosMsb(BitStream *pBits, lil_u32 bitPos);
template <class BitStream> inline void BsSkipLsb(BitStream *pBits, lil_i32 numBits);
template <class BitStream> inline void BsSkipMsb(BitStream *pBits, lil_i32 numBits);
template <class BitStream> inline void BsUndoLsb(BitStream *pBits, lil_u32 in, lil_u8 numBits);
template <class BitStream> inline void BsUndoMsb(BitStream *pBits, lil_u32 in, lil_u8 numBits);

//-------------------------------------------------------------------------------------------------------
// DEFINITIONS
//-------------------------------------------------------------------------------------------------------
inline lil_u32 BitSet(lil_u32 bits, lil_u32 mask) {	return bits | mask; }

//-------------------------------------------------------------------------------------------------------
inline lil_u32 BitReset(lil_u32 bits, lil_u32 mask) { return bits & ~mask; }

//-------------------------------------------------------------------------------------------------------
inline lil_u32 BitFlip(lil_u32 bits, lil_u32 mask) { return bits ^ mask; }

//-------------------------------------------------------------------------------------------------------
inline lil_u32 BitToggle(lil_u32 bits, lil_u32 mask, bool toggle) {
	return toggle ? BitSet(bits, mask) : BitReset(bits, mask);
}

//-------------------------------------------------------------------------------------------------------
inline lil_u8 BitReverse8(lil_u8 value) {
	return static_cast<lil_u8>(((value * 0x0802ul & 0x22110ul) | (value * 0x8020ul & 0x88440ul)) * 0x10101ul >> 16);
}

//-------------------------------------------------------------------------------------------------------
inline lil_u16 BitReverse16(lil_u16 value) {
	value = ((value & 0xaaaa) >> 1) | ((value & 0x5555) << 1);
	value = ((value & 0xcccc) >> 2) | ((value & 0x3333) << 2);
	value = ((value & 0xf0f0) >> 4) | ((value & 0x0f0f) << 4);
	return (value >> 8) | (value << 8);
}

//-------------------------------------------------------------------------------------------------------
inline lil_u32 BitReverse32(lil_u32 value) {
	value = ((value & 0xaaaaaaaa) >> 1) | ((value & 0x55555555) << 1);
	value = ((value & 0xcccccccc) >> 2) | ((value & 0x33333333) << 2);
	value = ((value & 0xf0f0f0f0) >> 4) | ((value & 0x0f0f0f0f) << 4);
	value = ((value & 0xff00ff00) >> 8) | ((value & 0x00ff00ff) << 8);
	return (value >> 16) | (value << 16);
}

//-------------------------------------------------------------------------------------------------------
inline int BitGetCount(lil_u32 value) {
   value = value - ((value >> 1) & 0x55555555);
   value = (value & 0x33333333) + ((value >>  2) & 0x33333333);
   return ((value + (value >> 4) & 0xf0f0f0f0) * 0x01010101) >> 24;
}

//-------------------------------------------------------------------------------------------------------
inline int BitGetMsb(lil_u32 value) {
	static const int MultiplyDeBruijnBitPosition[32] = { 0, 9, 1, 10, 13, 21, 2, 29, 
		11, 14, 16, 18, 22, 25, 3, 30, 8, 12, 20, 28, 15, 17, 24, 7, 19, 27, 23, 6, 26, 5, 4, 31 };
	value |= value >> 1;
	value |= value >> 2;
	value |= value >> 4;
	value |= value >> 8;
	value |= value >> 16;
	return MultiplyDeBruijnBitPosition[(value * 0x07c4acddu) >> 27];
}

//-------------------------------------------------------------------------------------------------------
inline int BitGetLsb(lil_u32 value) {
	static const int MultiplyDeBruijnBitPosition[32] = { 0, 1, 28, 2, 29, 14, 24, 3, 
		30, 22, 20, 15, 25, 17, 4, 8, 31, 27, 13, 23, 21, 19, 16, 7, 26, 12, 18, 6, 11, 5, 10, 9 };
	return MultiplyDeBruijnBitPosition[((value & -static_cast<lil_i32>(value)) * 0x077cb531u) >> 27];
}

//-----------------------------------------------------------------------------------------------------------
inline void BitInvertBuf(void *pData, lil_u32 numBytes) {
	lil_u32 *pData32 = reinterpret_cast<lil_u32 *>(pData);
	for (lil_u32 i = numBytes >> 2; i--; pData32++)
		*pData32 = ~(*pData32);
	lil_u8 *pData8 = reinterpret_cast<lil_u8 *>(pData32);
	for (lil_u32 i = numBytes & 3; i--; pData8++)
		*pData8 = ~(*pData8);
}

//-----------------------------------------------------------------------------------------------------------
inline void BitReverseBuf8(void *pData, lil_u32 numBytes) {
	lil_u32 *pData32 = reinterpret_cast<lil_u32 *>(pData);
	for (lil_u32 i = numBytes >> 2; i--; pData32++) {  // Reverse 4 bytes at a time in parallel
		*pData32 = ((*pData32 & 0xaaaaaaaa) >> 1) | ((*pData32 & 0x55555555) << 1);
		*pData32 = ((*pData32 & 0xcccccccc) >> 2) | ((*pData32 & 0x33333333) << 2);
		*pData32 = ((*pData32 & 0xf0f0f0f0) >> 4) | ((*pData32 & 0x0f0f0f0f) << 4);
	}
	lil_u8 *pData8 = reinterpret_cast<lil_u8 *>(pData32);
	for (lil_u32 i = numBytes & 3; i--; pData8++)
		*pData8 = static_cast<lil_u8>(((*pData8 * 0x0802ul & 0x22110ul) | (*pData8 * 0x8020ul & 0x88440ul)) * 0x10101ul >> 16);
}

//-------------------------------------------------------------------------------------------------------
inline void ByteSwap16(lil_u16 &value) { value = (value >> 8) | (value << 8); }

//-------------------------------------------------------------------------------------------------------
inline void ByteSwap32(lil_u32 &value) {
	value = ((value & 0xff00ff00) >> 8) | ((value & 0x00ff00ff) << 8);
	value = (value >> 16) | (value << 16);
}

//-------------------------------------------------------------------------------------------------------
inline void ByteSwap64(lil_u64 &value) {
	value = ((value & 0xff00ff00ff00ff00ull) >> 8) | ((value & 0x00ff00ff00ff00ffull) << 8);
	value = ((value & 0xffff0000ffff0000ull) >> 16) | ((value & 0x0000ffff0000ffffull) << 16);
	value = (value >> 32) | (value << 32);
}

//-----------------------------------------------------------------------------------------------------------
inline void ByteSwapBuf16(void *pData, lil_u32 count) {
	lil_u16 *pData16 = reinterpret_cast<lil_u16 *>(pData);
	for (lil_u32 i = 0; i < count; i++)
		ByteSwap16(pData16[i]);
}

//-----------------------------------------------------------------------------------------------------------
inline void ByteSwapBuf32(void *pData, lil_u32 count) {
	lil_u32 *pData32 = reinterpret_cast<lil_u32 *>(pData);
	for (lil_u32 i = 0; i < count; i++)
		ByteSwap32(pData32[i]);
}

//-------------------------------------------------------------------------------------------------------
inline void BsInit(BitStreamIn *pBits, const void *pBase, lil_u32 byteSize) {
	pBits->pPos = pBits->pBase = reinterpret_cast<const lil_u8 *>(pBase);
	pBits->bytesLeft = byteSize;
	pBits->hasBits = pBits->currBits = 0;
}

//-----------------------------------------------------------------------------------------------------------
inline void BsInit(BitStreamOut *pBits, void *pBase, lil_u32 byteSize) {
	pBits->pPos = pBits->pBase = reinterpret_cast<lil_u8 *>(pBase);
	pBits->bytesLeft = byteSize;
	pBits->hasBits = pBits->currBits = 0;
}

//-----------------------------------------------------------------------------------------------------------
inline bool BsReadLsb(BitStreamIn *pBits, lil_u32 *pOut, lil_u8 numBits) {
	while (pBits->hasBits < numBits) {
		if (pBits->bytesLeft == 0)
			return false;
		pBits->currBits |= *pBits->pPos++ << pBits->hasBits;
		pBits->bytesLeft--;
		pBits->hasBits += 8;
	}
	*pOut = pBits->currBits & ((1 << numBits) - 1);
	pBits->currBits >>= numBits;
	pBits->hasBits -= numBits;
	return true;
}

//-----------------------------------------------------------------------------------------------------------
inline bool BsReadMsb(BitStreamIn *pBits, lil_u32 *pOut, lil_u8 numBits) {
	while (pBits->hasBits < numBits) {
		if (pBits->bytesLeft == 0)
			return false;
		pBits->currBits |= *pBits->pPos++ << (24 - pBits->hasBits);
		pBits->bytesLeft--;
		pBits->hasBits += 8;
	}
	*pOut = pBits->currBits >> (32 - numBits);
	pBits->currBits <<= numBits;
	pBits->hasBits -= numBits;
	return true;
}

//-----------------------------------------------------------------------------------------------------------
inline void BsReadFlush(BitStreamIn *pBits) {
	if (pBits->hasBits) {
		pBits->pPos++;
		pBits->bytesLeft--;
		pBits->currBits = 0;
		pBits->hasBits = 0;
	}
}

//-----------------------------------------------------------------------------------------------------------
inline bool BsWriteLsb(BitStreamOut *pBits, lil_u32 in, lil_u8 numBits) {
	pBits->currBits |= in << pBits->hasBits;
	pBits->hasBits += numBits;
	while (pBits->hasBits >= 8) {
		*pBits->pPos++ = pBits->currBits & 0xff;
		pBits->bytesLeft--;
        pBits->currBits >>= 8;
        pBits->hasBits -= 8;
	}
	return true;
}

//-----------------------------------------------------------------------------------------------------------
inline bool BsWriteMsb(BitStreamOut *pBits, lil_u32 in, lil_u8 numBits) {
	pBits->currBits |= in << (32 - numBits - pBits->hasBits);
	pBits->hasBits += numBits;
	while (pBits->hasBits >= 8) {
		*pBits->pPos++ = pBits->currBits >> 24;
		pBits->bytesLeft--;
        pBits->currBits <<= 8;
        pBits->hasBits -= 8;
	}
	return true;
}

//-----------------------------------------------------------------------------------------------------------
inline void BsWriteFlushLsb(BitStreamOut *pBits) {
	BsWriteLsb(pBits, 0, 0);
    if (pBits->hasBits) {
		lil_u8 mask = (1 << pBits->hasBits) - 1;
		*pBits->pPos = *pBits->pPos & ~mask;
		*pBits->pPos++ |= pBits->currBits & mask;
		pBits->bytesLeft--;
        pBits->hasBits = 0;
        pBits->currBits = 0;
    }
}

//-----------------------------------------------------------------------------------------------------------
inline void BsWriteFlushMsb(BitStreamOut *pBits) {
	BsWriteMsb(pBits, 0, 0);
    if (pBits->hasBits) {
		lil_u8 invmask = (1 << (8 - pBits->hasBits)) - 1;
		*pBits->pPos = *pBits->pPos & invmask;
		*pBits->pPos++ |= (pBits->currBits >> 24) & ~invmask;
		pBits->bytesLeft--;
        pBits->hasBits = 0;
        pBits->currBits = 0;
    }
}

//-----------------------------------------------------------------------------------------------------------
template <class BitStream> inline void BsSetPosLsb(BitStream *pBits, lil_u32 bitPos) {
	pBits->bytesLeft = pBits->bytesLeft + (pBits->pPos - pBits->pBase) - ((bitPos + 7) >> 3);
	pBits->pPos = pBits->pBase + (bitPos >> 3);
	pBits->hasBits = (8 - (bitPos & 7)) & 7;
	pBits->currBits = 0;
	if (pBits->hasBits)
		pBits->currBits = *pBits->pPos++ >> (bitPos & 7);
}

//-----------------------------------------------------------------------------------------------------------
template <class BitStream> inline void BsSetPosMsb(BitStream *pBits, lil_u32 bitPos) {
	pBits->bytesLeft = pBits->bytesLeft + (pBits->pPos - pBits->pBase) - ((bitPos + 7) >> 3);
	pBits->pPos = pBits->pBase + (bitPos >> 3);
	pBits->hasBits = (8 - (bitPos & 7)) & 7;
	pBits->currBits = 0;
	if (pBits->hasBits)
		pBits->currBits = *pBits->pPos++ << (24 + (bitPos & 7));
}

//-----------------------------------------------------------------------------------------------------------
template <class BitStream> inline void BsSkipLsb(BitStream *pBits, lil_i32 numBits) {
	lil_i32 curr = static_cast<lil_i32>(((pBits->pPos - pBits->pBase) << 3) + pBits->hasBits);
	BsSetPosLsb(pBits, static_cast<lil_u32>(curr + numBits));
}

//-----------------------------------------------------------------------------------------------------------
template <class BitStream> inline void BsSkipMsb(BitStream *pBits, lil_i32 numBits) {
	lil_i32 curr = static_cast<lil_i32>(((pBits->pPos - pBits->pBase) << 3) + pBits->hasBits);
	BsSetPosMsb(pBits, static_cast<lil_u32>(curr + numBits));
}

//-----------------------------------------------------------------------------------------------------------
template <class BitStream> inline void BsUndoLsb(BitStream *pBits, lil_u32 in, lil_u8 numBits) {
	pBits->currBits <<= numBits;
	pBits->currBits |= in & ((1 << numBits) - 1);
	pBits->hasBits += numBits;
}

//-----------------------------------------------------------------------------------------------------------
template <class BitStream> inline void BsUndoMsb(BitStream *pBits, lil_u32 in, lil_u8 numBits) {
	pBits->currBits >>= numBits;
	pBits->currBits |= in << (32 - numBits);
	pBits->hasBits += numBits;
}






// For example of a 5 bit number.  Notice the numbers themselves are still MSB-LSB,
// regardless of the packing method in which they are stored.
//
// MSB (to LSB) order of bit packing
// Most significant bit of the stored number goes to the most significant available
// bit in storage.
// 54321543 21543215 43215432 15432154 32154321
// -------- -------- -------- -------- --------
// 11111222 22333334 44445555 56666677 77788888
//
// LSB (to MSB) order of bit packing
// Least significant bit of the stored number goes to the least significant available
// bit in storage.
// 32154321 15432154 43215432 21543215 54321543
// -------- -------- -------- -------- --------
// 22211111 43333322 55554444 77666665 88888777

//-------------------------------------------------------------------------------------------------------
// DECLARATIONS
//-------------------------------------------------------------------------------------------------------

class BitIoMsb;
class BitInLsb;
class BitInMsb;

// BitIoMsb, [BitIoLsb], BitInLsb, [BitOutLsb], BitInMsb, [BitOutMsb]
// MSB aligned bit input/output
class BitIoMsb {
public:
	LIL_INLINE BitIoMsb();
	LIL_INLINE BitIoMsb(void *pBase);
	LIL_INLINE BitIoMsb(const void *pBase);

	LIL_INLINE void Reset(void *pBase);
	LIL_INLINE void Reset(const void *pBase);
	LIL_INLINE lil_u32 ReadBits(lil_u8 numBits);
	LIL_INLINE void WriteBits(lil_u8 numBits, lil_u32 value);
	LIL_INLINE void SkipBits(lil_i64 numBits);
	// Sets position from base.
	LIL_INLINE void SetBitPos(lil_u64 numBits);
	LIL_INLINE void AlignToByte();

private:
	lil_u8 *pBase;
	lil_u8 *pIoByte;
	lil_u8 bitPos;
	bool readOnly;
};

// LSB aligned bit input streaming
class BitInLsb {
public:
	LIL_INLINE BitInLsb(const void *pBase, size_t byteSize);

	LIL_INLINE void ResetStreamPtr(const void *pBase, size_t byteSize);
	LIL_INLINE bool ReadBits(lil_u32 &output, lil_u8 numBits);

private:
	const lil_u8 *pByte;
	lil_u8 hasBits;
	lil_u32 currBits;
	size_t bytesLeft;
};

// MSB aligned bit input streaming
class BitInMsb {
public:
	LIL_INLINE BitInMsb(const void *pBase, size_t byteSize);

	LIL_INLINE void ResetStreamPtr(const void *pBase, size_t byteSize);
	LIL_INLINE bool ReadBits(lil_u32 &output, lil_u8 numBits);

private:
	const lil_u8 *pByte;
	lil_u8 hasBits;
	lil_u32 currBits;
	size_t bytesLeft;
};


//-------------------------------------------------------------------------------------------------------
// BITIOMSB IMPLEMENTATION
//-------------------------------------------------------------------------------------------------------
LIL_INLINE BitIoMsb::BitIoMsb() : pBase(0), pIoByte(0), bitPos(7), readOnly(false) { }

//-------------------------------------------------------------------------------------------------------
LIL_INLINE BitIoMsb::BitIoMsb(void *pBase) : pBase(0), pIoByte(0), bitPos(7), readOnly(false) {
	this->pBase = reinterpret_cast<lil_u8 *>(pBase);
	pIoByte = this->pBase;
}

//-------------------------------------------------------------------------------------------------------
LIL_INLINE BitIoMsb::BitIoMsb(const void *pBase) : pBase(0), pIoByte(0), bitPos(7), readOnly(true) {
	this->pBase = reinterpret_cast<lil_u8 *>(const_cast<void *>(pBase));
	pIoByte = this->pBase;
}

//-------------------------------------------------------------------------------------------------------
LIL_INLINE void BitIoMsb::Reset(void *pBase) {
	readOnly = false;
	this->pBase = reinterpret_cast<lil_u8 *>(pBase);
	pIoByte = this->pBase;
	bitPos = 7;
}

//-------------------------------------------------------------------------------------------------------
LIL_INLINE void BitIoMsb::Reset(const void *pBase) {
	readOnly = true;
	this->pBase = reinterpret_cast<lil_u8 *>(const_cast<void *>(pBase));
	pIoByte = this->pBase;
	bitPos = 7;
}

//-------------------------------------------------------------------------------------------------------
LIL_INLINE lil_u32 BitIoMsb::ReadBits(lil_u8 numBits)
{
	const lil_u8 lowBitMask[8] = { 1, 3, 7, 15, 31, 63, 127, 255 };
	const lil_u8 highBitMask[7] = { 0x80, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc, 0xfe };

	// Partial first byte
	lil_u32 retVal = 0;
	retVal = *pIoByte & lowBitMask[bitPos];

	// For numBits of size < 8, all the bits could be in the same byte, and we won't want
	// to go to the next byte.
	if (bitPos >= numBits) {
		bitPos -= numBits;
		return retVal >> (bitPos + 1);
	}
	numBits = numBits - (bitPos + 1);
	// Whole bytes
	while (numBits & 0xf8) {
		numBits -= 8;
		pIoByte++;
		retVal <<= 8;
		retVal |= *pIoByte;
	}
	bitPos = 7;
	pIoByte++;
	// Partial last byte
	if (numBits) {
		bitPos -= numBits;
		retVal <<= numBits;
		retVal |= ((*pIoByte & highBitMask[numBits - 1]) >> (8 - numBits));
	}

	return retVal;
}

//-------------------------------------------------------------------------------------------------------
LIL_INLINE void BitIoMsb::WriteBits(lil_u8 numBits, lil_u32 value)
{
	if (readOnly) {
		LIL_BREAK();
		SkipBits(numBits);
	}

	const lil_u32 highBitMask[8] = { 0x80000000, 0xc0000000, 0xe0000000, 0xf0000000,
		0xf8000000, 0xfc000000, 0xfe000000, 0xff000000 };

	// Now the value is msb aligned, instead of lsb aligned.  This makes thinking about it
	// easier...
	value <<= 32 - numBits;

	// For when numBits still fits in the current byte
	if (bitPos >= numBits) {
		bitPos -= numBits;
		lil_u32 rightShift = ((32 - numBits) - (bitPos + 1));
		lil_u8 clearNotMask = static_cast<lil_u8>(highBitMask[numBits - 1] >> rightShift);
		*pIoByte &= ~clearNotMask;
		*pIoByte |= (value >> rightShift);
		return;
	}

	// Partial first byte
	if (bitPos < 7) {
		// The clear not mask is important to avoid clearing bits that aren't even involved!
		lil_u8 clearNotMask = static_cast<lil_u8>(highBitMask[bitPos] >> (31 - bitPos));
		*pIoByte &= ~clearNotMask;
		*pIoByte |= (value & highBitMask[bitPos]) >> (31 - bitPos);
		value <<= (bitPos + 1);
		numBits = numBits - (bitPos + 1);
		bitPos = 7;
		pIoByte++;
	}

	// Whole bytes
	while (numBits & 0xf8) {
		numBits -= 8;
		*pIoByte = value >> 24;
		value <<= 8;
		pIoByte++;
	}

	// Partial final byte
	if (numBits) {
		// The clear not mask is important to avoid clearing bits that aren't even involved!
		// This would only matter if actually decrementing the bit pointer each time, otherwise
		// you'd never notice the need.
		lil_u8 clearNotMask = static_cast<lil_u8>(highBitMask[numBits - 1] >> 24);
		*pIoByte &= ~clearNotMask;

		bitPos -= numBits;
		// Actually, everything to the right of the lowest bits would already be 0.. but just in case
		// I am not thinking straight we'll keep the relatively expensive extra AND for now.
		*pIoByte |= (value & highBitMask[numBits - 1]) >> 24;
	}
}

//-------------------------------------------------------------------------------------------------------
LIL_INLINE void BitIoMsb::SkipBits(lil_i64 numBits)
{
	// Can go backwards if need be.
	if (numBits < 0) {
		// "- numBits" adds the absolute value of numBits
		lil_u64 bits = (bitPos - numBits);
		pIoByte -= bits >> 3;
		bitPos = bits & 7;
	}
	// For numBits of size < 8, all the bits could be in the same byte, and we won't want
	// to go to the next byte.
	else if (bitPos >= numBits)
		bitPos -= static_cast<lil_u8>(numBits);
	else {
		numBits -= (bitPos + 1);  // Partial first byte
		pIoByte += 1 + (numBits >> 3);
		bitPos = 7 - static_cast<lil_u8>(numBits & 7);
	}
}

//-------------------------------------------------------------------------------------------------------
LIL_INLINE void BitIoMsb::SetBitPos(lil_u64 numBits)
{
	pIoByte = pBase + (numBits >> 3);
	bitPos = 7 - static_cast<lil_u8>(numBits & 7);
}

//-------------------------------------------------------------------------------------------------------
LIL_INLINE void BitIoMsb::AlignToByte()
{
	if (bitPos < 7)
		SkipBits(bitPos + 1);
}



//-------------------------------------------------------------------------------------------------------
// BITINLSB IMPLEMENTATION
//-------------------------------------------------------------------------------------------------------
LIL_INLINE BitInLsb::BitInLsb(const void *pBase, size_t byteSize) : pByte(0), hasBits(0), currBits(0), bytesLeft(0)
{
	ResetStreamPtr(pBase, byteSize);
}

//-------------------------------------------------------------------------------------------------------
LIL_INLINE void BitInLsb::ResetStreamPtr(const void *pBase, size_t byteSize)
{
	pByte = reinterpret_cast<const lil_u8 *>(pBase);
	bytesLeft = byteSize;
}

//-------------------------------------------------------------------------------------------------------
LIL_INLINE bool BitInLsb::ReadBits(lil_u32 &output, lil_u8 numBits)
{
	while (hasBits < numBits) {
		if (bytesLeft == 0)
			return false;
		currBits |= *pByte << hasBits;
		pByte ++;
		bytesLeft --;
		hasBits += 8;
	}
	output = currBits & ((1 << numBits) - 1);
	currBits >>= numBits;
	hasBits -= numBits;
	return true;
}



//-------------------------------------------------------------------------------------------------------
// BITINMSB IMPLEMENTATION
//-------------------------------------------------------------------------------------------------------
LIL_INLINE BitInMsb::BitInMsb(const void *pBase, size_t byteSize) : pByte(0), hasBits(0), currBits(0), bytesLeft(0)
{
	ResetStreamPtr(pBase, byteSize);
}

//-------------------------------------------------------------------------------------------------------
LIL_INLINE void BitInMsb::ResetStreamPtr(const void *pBase, size_t byteSize)
{
	pByte = reinterpret_cast<const lil_u8 *>(pBase);
	bytesLeft = byteSize;
}

//-------------------------------------------------------------------------------------------------------
LIL_INLINE bool BitInMsb::ReadBits(lil_u32 &output, lil_u8 numBits)
{
	while (hasBits < numBits) {
		if (bytesLeft == 0)
			return false;
		currBits = (currBits << 8) | *pByte;
		pByte++;
		bytesLeft--;
		hasBits += 8;
	}
	hasBits -= numBits;
	output = (currBits >> hasBits);
	currBits &= ((1 << hasBits) - 1);
	return true;
}

}  // namespace lil

#endif // LILBITMANIP_H_INCLUDED
