#include <memory.h>
#include "Lil.h"
#include "LilIo.h"
#include "LilBitManip.h"
#include "LilUtilFormat.h"
#include "LilMem.h"

#define MINIZ_HEADER_FILE_ONLY
#include "miniz.c"

// LZ77 is n-size window with Length, Distance, [Literal] information
// LZSS uses a 1 bit flag to indicate whether a Length-Distance pair or Literal is stored
//   It also allows a predefined length-offset so that is added to all length values
//   if, for instance, you only use literals if the string is less than 4 symbols wide
// Deflate combines the flag as well as info on how to read the proceeding length
//   info into a single huffman code.  Another huffman code is used to indicate how to read
//   the distance.

// Huffman codes are packed LSB aligned but in MSB-LSB stored order
// All other data is packed LSB aligned and in LSB-MSB stored order

namespace lil {

namespace {

const lil_u32 MAX_HUFFMAN_BITS = 15;
const lil_u32 DEFLATE_STREAM_END = 256;
const lil_u32 WINDOW_SIZE = 32768;
const lil_u32 NUM_BUCKETS = 512;
#define HASH(x1, x2, x3) (((x1 | (x2 << 8) | (x3 << 16)) * 2654435761UL) & (NUM_BUCKETS - 1))
#define HASHBUF_PREV(n) pState->hashBuf[(n) & 0xffff]
#define HASHBUF_NEXT(n) pState->hashBuf[(n) >> 16]

// These are from deflate spec
const lil_u8 BITS_READ[3] = { 2, 3, 7 };
const lil_u32 BITS_ADD[19] = { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 3, 11 };
const lil_u8 LEN_EXTRA[29] = { 0,0,0,0,0,0,0,0,1,1,1,1,2,2,2,2,3,3,3,3,4,4,4,4,5,5,5,5,0 };
const lil_u32 LEN_BASE[29] = { 3,4,5,6,7,8,9,10,11,13,15,17,19,23,27,31,35,43,51,59,67,83,99,115,131,163,195,227,258 };
const lil_u8 DIST_EXTRA[30] = { 0,0,0,0,1,1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9,10,10,11,11,12,12,13,13 };
const lil_u32 DIST_BASE[31] = { 1,2,3,4,5,7,9,13,17,25,33,49,65,97,129,193,257,385,513,769,1025,1537,2049,3073,4097,6145,8193,12289,16385,24577,32768 };

// huffman code size for reading the huffman code size for both dist and lit/len alphabets
static const lil_u32 CODE_LEN_ORDER[19] = { 16,17,18,0,8,7,9,6,10,5,11,4,12,3,13,2,14,1,15 };

//-----------------------------------------------------------------------------------------------------------
struct HuffmanDeflate {
	// Spec for rfc 1950 canonical huffman tree building
	// 'values' will be the value for each code (it's a dictionary)
	// 'maxCode' will be the max code available for each number of bits
	// 'bitLens' is the number of bits for each value
	//           Each code is determined from bit counts for values.
	lil_u32 values[32768];
	lil_u32 valuesCount;
	lil_u32 codes[288 + 32];
	lil_u32 codesCount;
	lil_u8 bitLens[288 + 32];
	lil_u32 bitLensCount;

	lil_u32 maxCode[MAX_HUFFMAN_BITS + 1];
	lil_u8 minBitLen;
	lil_u8 maxBitLen;

	HuffmanDeflate() : minBitLen(0), maxBitLen(0), valuesCount(0), codesCount(0), bitLensCount(0) { }

	void SetNumCodes(lil_u32 count) {
		LIL_ASSERT(count <= 288 + 32);
		bitLensCount = count;
		memset(bitLens, 0, count);
	}

	void LoadFixed() {
		bitLensCount = 288;
		for (int i = 0; i <= 143; i++) bitLens[i] = 8;
		for (int i = 144; i <= 255; i++) bitLens[i] = 9;
		for (int i = 256; i <= 279; i++) bitLens[i] = 7;
		for (int i = 280; i <= 287; i++) bitLens[i] = 8;
		BuildHuffmanTable(0);
	}

	void LoadFixedDist() {
		bitLensCount = 32;
		for (int i = 0; i <= 31; i++) bitLens[i] = 5;
		BuildHuffmanTable(0);
	}

	bool LoadDynamic(lil_u32 numCodes, HuffmanDeflate &source, BitReadLsb &reader, LilErrorType *pErr) {
		SetNumCodes(numCodes);

		lil_u32 extra = 0;
		lil_u32 bits = 0;
		for (lil_u32 i = 0; i < numCodes;) {
			lil_u32 prev = i ? i - 1 : 0;
			if (!source.Decode(bits, reader, 18, pErr))
				return false;;
			if (bits <= 15) {
				extra = BITS_ADD[bits];
				bitLens[i] = static_cast<lil_u8>(bits);
			}
			else {
				reader.ReadBits(extra, BITS_READ[bits - 16]);
				extra += BITS_ADD[bits];
				for (lil_u32 j = 0; j < extra; j++)
					bitLens[i + j] = (bits == 16 ? bitLens[prev] : 0);
			}
			i += extra;
			if (i > numCodes)
				return SetError(LE_BADCOMPRESSION, pErr);
		}
		return SetError(LE_NONE, pErr);
	}

	bool BuildHuffmanTable(LilErrorType *pErr) {
		maxBitLen = 0;
		minBitLen = MAX_HUFFMAN_BITS;
		codesCount = bitLensCount;
		for (lil_u32 i = 0; i < bitLensCount; i++) {
			if (bitLens[i] > maxBitLen)
				maxBitLen = bitLens[i];
			if (bitLens[i] && bitLens[i] < minBitLen)
				minBitLen = bitLens[i];
		}
		valuesCount = 1 << maxBitLen;
		LIL_ASSERT(valuesCount <= 32768);

		// Count codes with x bits
		lil_u32 bitLengthCount[MAX_HUFFMAN_BITS + 1];
		memset(bitLengthCount, 0, sizeof(bitLengthCount));
		for (lil_u32 i = 0; i < bitLensCount; i++) {
			if (bitLens[i] > MAX_HUFFMAN_BITS)
				return SetError(LE_BADCOMPRESSION, pErr);
			bitLengthCount[bitLens[i]] ++;
		}
		bitLengthCount[0] = 0;  // Bit length of 0 means unused... so reset it here!

		// Get smallest value for each code length
		maxCode[0] = 0;
		for (lil_u32 i = 1; i <= MAX_HUFFMAN_BITS; i++)
			maxCode[i] = (maxCode[i - 1] + bitLengthCount[i - 1]) << 1;

		// Assign codes to value-to-code array, and values to code-to-value array
		for (lil_u32 i = 0; i < bitLensCount; i++)
			if (bitLens[i]) {
				values[maxCode[bitLens[i]]] = i; // for decode
				codes[i] = maxCode[bitLens[i]] << (32 - bitLens[i]); // for encode
				codes[i] = BitReverse32(codes[i]);
				maxCode[bitLens[i]]++;
			}
		return SetError(LE_NONE, pErr);
	}

	LIL_INLINE bool Decode(lil_u32 &output, BitReadLsb &reader, lil_u32 maxOutput, LilErrorType *pErr) {
		// minBits actually changes and is the current bits
		reader.ReadBits(output, minBitLen);

		lil_u8 bitCount = minBitLen;
		output = BitReverse32(output);
		output >>= (32 - bitCount);
		while (output >= maxCode[bitCount]) {
			lil_u32 moreBits = 0;
			reader.ReadBits(moreBits, 1);
			output = (output << 1) | moreBits;
			bitCount++;
			if (bitCount > maxBitLen)
				return SetError(LE_BADCOMPRESSION, pErr);
		}
		output = values[output];
		LIL_ASSERT(bitLens[output] == bitCount);
		if (output > maxOutput)
			return SetError(LE_BADCOMPRESSION, pErr);

		return SetError(LE_NONE, pErr);
	}

	LIL_INLINE bool Encode(lil_u32 input, BitWriteLsb &writer, LilErrorType *pErr) {
		writer.WriteBits(codes[input], bitLens[input]);
		return SetError(LE_NONE, pErr);
	}
};

//-----------------------------------------------------------------------------------------------------------
struct HashEntry {
	lil_u32 nextPrevI;  // hi 16 bits are next index, lo 16 bits are prev index
	const lil_u8 *pPos;
};

//-----------------------------------------------------------------------------------------------------------
struct EncodeState {
	HuffmanDeflate huffFixed;
	HuffmanDeflate huffFixedDist;

	HuffmanDeflate huffDyn;
	HuffmanDeflate huffDynDist;
	HashEntry hashBuf[WINDOW_SIZE + 1];  // index 0 is a dummy to make indexing prev/next entry in list simpler
	HashEntry *buckets[NUM_BUCKETS];
	int hashBufI;
};

//-----------------------------------------------------------------------------------------------------------
struct DecodeState {
	HuffmanDeflate huffFixed;
	HuffmanDeflate huffFixedDist;
	HuffmanDeflate huffDyn;
	HuffmanDeflate huffDynDist;
	HuffmanDeflate huffCodeLen;
	HuffmanDeflate huffTemp;
};

//-----------------------------------------------------------------------------------------------------------
LilErrorType DecodeDeflate(void *pOut, const LilIoDesc *pIo, lil_u32 *pNumBytes, const LilMemDesc *mem) {
	DecodeState *pState = (DecodeState *)malloc(sizeof(DecodeState));//MemNew<DecodeState>(mem);
	memset(pState, 0, sizeof(DecodeState));
	//new pState(
	//RaiiMemDelete<DecodeState> del(pState);
    lil_u32 numBytes = 0;
	HuffmanDeflate *pHuff = 0, *pHuffDist = 0;

	//pState->huffFixed.LoadFixed();
	//pState->huffFixedDist.LoadFixedDist();

	// Remember bits are reversed for huffman codes!!
	lil_u8 *pOut8 = reinterpret_cast<lil_u8 *>(pOut);

	BitReadLsb reader(pIo);

	lil_u32 finalBlock = 0;
	do {
		LilErrorType err = LE_NONE;
		lil_u32 bits = 0, compressType = 0;
		reader.ReadBits(finalBlock, 1);
		reader.ReadBits(compressType, 2);
		switch (compressType) {
		case 0: // no compression
			reader.AlignToByte();
			bits = Read32le(pIo);
			if ((~bits >> 16) != (bits & 0xffff))  // should be 1s compliment
				return LE_BADCOMPRESSION;
			bits &= 0xffff;
			ReadN(pIo, bits, pOut8);
			pOut8 += bits;
			numBytes += bits;
			break;
		case 2: // dynamic huffman codes
			{
				reader.ReadBits(bits, 5);
				lil_u32 litLenCodes = 257 + bits;
				reader.ReadBits(bits, 5);
				lil_u32 distCodes = 1 + bits;
				reader.ReadBits(bits, 4);
				lil_u32 codeLenCodes = 4 + bits;

				pState->huffCodeLen.SetNumCodes(19);
				for (lil_u32 i = 0; i < codeLenCodes; i++) {
					reader.ReadBits(bits, 3);
					pState->huffCodeLen.bitLens[CODE_LEN_ORDER[i]] = static_cast<lil_u8>(bits);
				}
				if (!pState->huffCodeLen.BuildHuffmanTable(&err))
					return err;

				if (!pState->huffTemp.LoadDynamic(litLenCodes + distCodes, pState->huffCodeLen, reader, &err))
					return err;

				// 288??
				pState->huffDyn.SetNumCodes(litLenCodes);
				memcpy(pState->huffDyn.bitLens, pState->huffTemp.bitLens, litLenCodes);
				if (!pState->huffDyn.BuildHuffmanTable(&err))
					return err;

				// 32??
				pState->huffDynDist.SetNumCodes(distCodes);
				memcpy(pState->huffDynDist.bitLens, pState->huffTemp.bitLens + litLenCodes, distCodes);
				if (!pState->huffDynDist.BuildHuffmanTable(&err))
					return err;

				pHuff = &pState->huffDyn;
				pHuffDist = &pState->huffDynDist;
			}

		// *** FALL THROUGH ***

		case 1: // fixed huffman codes
			{
				if (compressType == 1) {
					pHuff = &pState->huffFixed;
					pHuffDist = &pState->huffFixedDist;
				}
				lil_u32 value = 0;
				while (value != DEFLATE_STREAM_END) {
					if (!pHuff->Decode(value, reader, 32768 + 257, &err))
						return err;

					if (value < DEFLATE_STREAM_END) {
						*pOut8++ = static_cast<lil_u8>(value);
                        numBytes++;
					}
					else if (value > DEFLATE_STREAM_END) {
						lil_u32 offset = value - 257;
						reader.ReadBits(bits, LEN_EXTRA[offset]);
						lil_u32 len = LEN_BASE[offset] + bits;
						if (!pHuffDist->Decode(offset, reader, 65535, &err))
							return err;
						reader.ReadBits(bits, DIST_EXTRA[offset]);
						lil_u32 dist = DIST_BASE[offset] + bits;
						lil_u8 *pIn = pOut8 - dist;
						for (lil_u32 i = 0; i < len; i++)
							*pOut8++ = *pIn++;
                        numBytes += len;
					}
				}
			}
			break;
		case 3: // reserved
			return LE_BADCOMPRESSION;
		}
	} while (!finalBlock);
	*pNumBytes = numBytes;
	free(pState);
	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE lil_u32 Compare(const lil_u8 *pIn8, const lil_u8 *pTest, lil_u32 maxLen) {
	if (static_cast<size_t>(pIn8 - pTest) >= WINDOW_SIZE)
		return 0;
	lil_u32 currLen = 0;
	const lil_u32 *pIn32 = reinterpret_cast<const lil_u32 *>(pIn8);
	const lil_u32 *pTest32 = reinterpret_cast<const lil_u32 *>(pTest);
	lil_u32 maxLen4 = maxLen >> 2;
	while (pTest32[currLen] == pIn32[currLen] && currLen < maxLen4)
		currLen++;
	currLen <<= 2;
	while (pTest[currLen] == pIn8[currLen] && currLen < maxLen)
		currLen++;
	return currLen;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType EncodeDeflate(const void *pIn, const LilIoDesc *pIo, lil_u32 numBytes, const LilMemDesc *mem,
	lil_u32 compLevel)
{
	EncodeState *pState = MemNew<EncodeState>(mem);
	memset(pState, 0, sizeof(EncodeState));
	pState->hashBufI = 1;  // entry 0 is a dummy entry
	pState->huffFixed.LoadFixed();
	pState->huffFixedDist.LoadFixedDist();

	const lil_u8 *pIn8 = reinterpret_cast<const lil_u8 *>(pIn);

	LilErrorType err = LE_NONE;
	BitWriteLsb writer(pIo);

	lil_u32 finalBlock = 0;
	lil_u32 compressType = GetMin(compLevel, 1u);

	do {
		lil_u32 bits = numBytes > 65535 ? 65535 : numBytes;
		numBytes -= bits;
		if (numBytes == 0)
            finalBlock = 1;
		writer.WriteBits(finalBlock, 1);
		writer.WriteBits(compressType, 2);

		switch (compressType) {
		case 0: // no compression
			writer.FlushByte();
			Write32le(pIo, bits | ((~bits) << 16));  // hiword needs to be 1s compliment of loword

			err = WriteN(pIo, bits, pIn8);
			if (err != LE_NONE)
                return err;
			pIn8 += bits;
			break;
		case 2: // dynamic huffman codes


        case 1: // fixed huffman codes
			// We're going to implement algorithm given in spec. Coming up with something
			// faster wouldn't be too hard. Unfortunately it's very difficult to compare
			// an alternative algorithm to established patents :/ thus the documenting
			// of this specific approach in deflate specs.
			while (bits >= 3) {
				lil_u32 iB = HASH(pIn8[0], pIn8[1], pIn8[2]);
				lil_u32 bestLen = 2; // Must match 3+ chars to use dist-len codes
				const lil_u8 *bestMatch = 0;
				lil_u32 maxLen = GetMin(bits, 257u);
				HashEntry *pBucket = pState->buckets[iB];
				while (pBucket && pBucket->pPos) {
					lil_u32 currLen = Compare(pIn8, pBucket->pPos, maxLen);
					if (currLen > bestLen) {
						bestLen = currLen;
						bestMatch = pBucket->pPos;
						if (currLen == maxLen)
							break;
					}
					pBucket = &pState->hashBuf[pBucket->nextPrevI >> 16];
				}
				// lazy-check - affects speed but gives small gains, but uneven tradeoff though...
				if (compLevel == LCL_MAX && bits >= (bestLen + 4) && bestMatch && bestLen < maxLen) {
					pIn8++;
					lil_u32 iB = HASH(pIn8[0], pIn8[1], pIn8[2]);
					HashEntry *pBucket = pState->buckets[iB];
					while (pBucket && pBucket->pPos) {
						lil_u32 currLen = Compare(pIn8, pBucket->pPos, maxLen);
						if (currLen > bestLen) {
							bestMatch = 0;
							break;
						}
						pBucket = &pState->hashBuf[pBucket->nextPrevI >> 16];
					}
					pIn8--;
				}
				if (bestMatch) {
					// encode len
					lil_u32 dist = static_cast<lil_u32>(pIn8 - bestMatch);
					lil_u32 i = 0;
					while (LEN_BASE[i] <= bestLen)
						i++;
					i--;
					pState->huffFixed.Encode(i + 257, writer, &err);
					if (LEN_EXTRA[i])
						writer.WriteBits(bestLen - LEN_BASE[i], LEN_EXTRA[i]);
					// encode dist
					i = 0;
					while (DIST_BASE[i] <= dist)
						i++;
					i--;
					pState->huffFixedDist.Encode(i, writer, &err);
					if (DIST_EXTRA[i])
						writer.WriteBits(dist - DIST_BASE[i], DIST_EXTRA[i]);
				}
				else {
					if (!pState->huffFixed.Encode(pIn8[0], writer, &err))
						return err;
					bestLen = 1;
				}
				// Storing hashes for all parts of recent string is a pretty big
				// win in most cases - but of course it's slightly slower
				size_t lesser = (compLevel > LCL_FAST ? GetMin(bestLen, bits - 3) : 1);
				for (const lil_u8 *pIn82 = pIn8; pIn82 < pIn8 + lesser; pIn82++) {
					iB = HASH(pIn82[0], pIn82[1], pIn82[2]);

					// Add to bucket
					HashEntry *pHashOld = pState->buckets[iB];
					HashEntry *pHashNew = &pState->hashBuf[pState->hashBufI];
					pState->hashBufI++;
					if (pState->hashBufI >= LIL_ARRAY_SIZE(pState->hashBuf))
						pState->hashBufI = 1;  // index 0 is a special dummy index for nextPrevI ease of use

					pState->buckets[iB] = pHashNew;

					// Update hash circular-table
					//LIL_ASSERT((pHashNew->nextPrevI >> 16) == 0);  // No nexts = nothing older than this guy should exist
					HASHBUF_PREV(pHashNew->nextPrevI).nextPrevI &= 0xffff;  // Tell the newer entry (prev) that there are no older entries
					if (pHashOld && pHashOld != pHashNew) {
						pHashNew->nextPrevI = static_cast<lil_u32>((pHashOld - &pState->hashBuf[1]) << 16);
						pHashOld->nextPrevI |= (pHashNew - &pState->hashBuf[1]);  // should be & 0xffff already
					}
					else
						pHashNew->nextPrevI = 0;
					pHashNew->pPos = pIn82;
				}
				pIn8 += bestLen;
				bits -= bestLen;
			}
			// Store last couple bytes as literals
			while (bits) {
				if (!pState->huffFixed.Encode(*pIn8, writer, &err))
					return err;
				pIn8++;
				bits--;
			}
			pState->huffFixed.Encode(DEFLATE_STREAM_END, writer, &err);

            break;
		}
	} while (!finalBlock);

	return LE_NONE;
}

}  // namespace

//-----------------------------------------------------------------------------------------------------------
lil_u32 DecodeZlibGetMem() {
	return sizeof(DecodeState);
}

//-----------------------------------------------------------------------------------------------------------
lil_u32 EncodeZlibGetMem() {
	return sizeof(EncodeState);
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType DecodeZlib(void *pOut, const LilIoDesc *io, lil_u32 numBytesIn, const LilMemDesc *mem) {
/*	lil_u16 header = Read16be(io);

	// The only supported type in zlib currently is deflate, or method "8" of low 4 bits of 1st byte
	// High 4 bits of first byte is log2 of window size.  Max 7.
	// Lower 5 bits of 2nd byte ensure that the total header word is a multiple of 31.
	if ((header & 0x0f00) != 0x0800 || (header & 0x8000) || (header % 31))
		return LE_BADCOMPRESSION;
	//lil_u32 dictSize = 1 << (8 + ((header >> 12) & 7));

	lil_u32 dictId = 0;  // Adler32 of initial dictionary to id it for decompression
	if (header & 0x0020)
		dictId = Read32le(io);

    lil_u32 numBytes = 0;
	LilErrorType err = DecodeDeflate(pOut, io, &numBytes, mem);
	if (err != LE_NONE)
		return err;

	lil_u32 adler32 = Read32be(io);
    lil_u32 checkAdler = CalcAdler32(reinterpret_cast<lil_u8 *>(pOut), numBytes);
    if (adler32 != checkAdler)
        return LE_BADCRC;
	return LE_NONE;
*/

	const lil_u32 BUF_SIZE = 8192;
	lil_u8 in[BUF_SIZE];
	z_stream stream;
	memset(&stream, 0, sizeof(stream));
	stream.next_in = in;
	stream.avail_in = 0;
	stream.next_out = reinterpret_cast<lil_u8 *>(pOut);
	stream.avail_out = BUF_SIZE;

    if (inflateInit(&stream))
	  return LE_INTERNALERROR;

	LilErrorType err = LE_NONE;

    while (1) {
      int status;
      if (!stream.avail_in) {
        // Input buffer is empty, so read more bytes from input file.
		  lil_u32 n = GetMin(BUF_SIZE, numBytesIn);

		  if (!SetError(io->readFunc(io->handle, n, in), &err)) {
			  return err;
		  }

        stream.next_in = in;
        stream.avail_in = n;

        numBytesIn -= n;
      }

      status = inflate(&stream, Z_SYNC_FLUSH);

      if ((status == Z_STREAM_END) || (!stream.avail_out)) {
        //lil_u32 n = BUF_SIZE - stream.avail_out;
        //stream.next_out += n;
        stream.avail_out = BUF_SIZE;
      }

      if (status == Z_STREAM_END)
        break;
      else if (status != Z_OK) {
		  return LE_BADCOMPRESSION;
      }
    }

    if (inflateEnd(&stream) != Z_OK)
		return LE_BADCOMPRESSION;

	//lil_u32 adler32 = Read32be(io);
    //lil_u32 checkAdler = CalcAdler32(reinterpret_cast<lil_u8 *>(pOut), numBytes);
//if (adler32 != stream.adler)
  //      return LE_BADCRC;
	return LE_NONE;
}

static mz_bool write(const void* pBuf, int len, void *pUser) {
	return ((LilIoDesc *)pUser)->writeFunc(((LilIoDesc *)pUser)->handle, len, pBuf) == LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType EncodeZlib(const void *pIn, const LilIoDesc *io, lil_u32 numBytesIn, LilCompressionLevelType level, const LilMemDesc *mem) {
/*    // png specs specifically state that there be NO dictionary id in header
    lil_u32 dictId = 0;

    lil_u16 header = 0x7800;
    if (dictId)
        header |= 0x0020;
    header += 31 - (header % 31);
    Write16be(io, header);

	//lil_u32 dictSize = 1 << (8 + ((header >> 12) & 7));

	LilErrorType err = EncodeDeflate(pIn, io, numBytes, mem, level);
	if (err != LE_NONE)
		return err;

*/	tdefl_compress_mem_to_output(pIn, numBytesIn, 
		//[](const void* pBuf, int len, void *pUser) -> mz_bool { ((LilIoDesc *)pUser)->writeFunc(((LilIoDesc *)pUser)->handle, len, pBuf); return true; },
		write,
		(void *)io, TDEFL_WRITE_ZLIB_HEADER | TDEFL_MAX_PROBES_MASK);
	
    //lil_u32 adler32 = CalcAdler32(reinterpret_cast<const lil_u8 *>(pIn), numBytesIn);
    //Write32be(io, adler32);

	return LE_NONE;
}

}  // namespace lil
