#ifndef LILCODECJPEG_H_INCLUDED
#define LILCODECJPEG_H_INCLUDED

#include "Lil.h"
#include "LilMem.h"

namespace lil {

// This is the version which we are compatible up to.
const lil_u16 SUPPORTED_VERSION = 0x0102;

const lil_u32 ID_JFIF = 'JFIF';            // JFIF main APP0 header
const lil_u32 ID_JFXX = 'JFXX';            // JFXX extension APP0 header

const lil_u8 UNITS_ASPECT_RATIO = 0;
const lil_u8 UNITS_DPI = 1;
const lil_u8 UNITS_DPC = 2;

const lil_u8 THUMB_JPEG = 0x10;
const lil_u8 THUMB_1 = 0x11;
const lil_u8 THUMB_3 = 0x13;

const lil_u8 PRECISION_8_BIT = 0;
const lil_u8 PRECISION_16_BIT = 1;

// [Spec 4.4 Fig 5 - Zig Zag]
// [Spec A.4 Fig 6 - Zig Zag]
const lil_u8 FROM_ZIG_ZAG[64] = { 
	0,  1,  8,  16, 9,  2,  3,  10,
	17, 24, 32, 25, 18, 11, 4,  5,
	12, 19, 26, 33, 40, 48, 41, 34,
	27, 20, 13, 6,  7,  14, 21, 28,
	35, 42, 49, 56,	57, 50, 43, 36,
	29, 22, 15, 23, 30, 37, 44, 51,
	58, 59, 52, 45, 38, 31, 39, 46, 
	53, 60, 61, 54, 47, 55, 62, 63 };
const lil_u8 TO_ZIG_ZAG[64] = { 
	0,  1,  5,  6,  14, 15, 27, 28,
	2,  4,  7,  13, 16, 26, 29, 42,
	3,  8,  12, 17, 25, 30, 41, 43,
	9,  11, 18, 24, 31, 40, 44, 53,
	10, 19, 23, 32, 39, 45, 52, 54,
	20, 22, 33, 38, 46, 51, 55, 60,
	21, 34, 37, 47, 50, 56, 59, 61,
	35, 36, 48, 49, 57, 58, 62, 63 };


// [Spec B.1.1.3 - Marker types]
// All markers start with 0xff
const lil_u16 MARKER_START_OF_IMAGE = 0xffd8;
const lil_u16 MARKER_END_OF_IMAGE = 0xffd9;
const lil_u16 MARKER_START_OF_SCAN = 0xffda;
const lil_u16 MARKER_QUANTIZATION = 0xffdb;
const lil_u16 MARKER_NUM_LINES = 0xffdc;
const lil_u16 MARKER_RESTART_INTERVAL = 0xffdd;
const lil_u16 MARKER_COMMENT = 0xfffe;
const lil_u16 MARKER_APP0 = 0xffe0;             // Application segment 0
const lil_u16 MARKER_APPF = 0xffef;
const lil_u16 MARKER_TEMP_PRIVATE = 0xff01;     // For temporary private use in arithmetic coding
const lil_u16 MARKER_NONE_PADDING = 0xffff;     // Keep reading until no more 0xff
const lil_u16 MARKER_RESTART_0 = 0xffd0;
const lil_u16 MARKER_RESTART_7 = 0xffd7;
// Misc
const lil_u16 MARKER_HUFFMAN_SPEC = 0xffc4;
const lil_u16 MARKER_ARITH_SPEC = 0xffcc;
// Start of frame markers, non differential, huffman coding
const lil_u16 MARKER_BASELINE_DCT = 0xffc0;     // SOF0
const lil_u16 MARKER_EXT_SEQ_DCT_H = 0xffc1;    // SOF1
const lil_u16 MARKER_PROG_DCT_H = 0xffc2;       // SOF2
const lil_u16 MARKER_LOSSLESS_H = 0xffc3;       // SOF3
// Start of frame markers, differential, huffman coding
const lil_u16 MARKER_DIFF_SEQ_DCT_H = 0xffc5;   // SOF5;
const lil_u16 MARKER_DIFF_PROG_DCT_H = 0xffc6;  // SOF6;
const lil_u16 MARKER_DIFF_LOSSLESS_H = 0xffc7;  // SOF7
// Start of frame markers, non differnetial, arithmetic coding
const lil_u16 MARKER_JPG = 0xffc8;              // Reserved for jpeg extensions
const lil_u16 MARKER_EXT_SEQ_DCT_A = 0xffc9;    // SOF9
const lil_u16 MARKER_PROG_DCT_A = 0xffca;       // SOF10
const lil_u16 MARKER_LOSSLESS_A = 0xffcb;       // SOF11
// Start of frame markers, differential, arithmetic coding
const lil_u16 MARKER_DIFF_SEQ_DCT_A = 0xffcd;   // SOF13;
const lil_u16 MARKER_DIFF_PROG_DCT_A = 0xffce;  // SOF14;
const lil_u16 MARKER_DIFF_LOSSLESS_A = 0xffcf;  // SOF15


class JpegBitReadMsb;
struct HuffmanMsb;

// Decoding is done according to needs of Lil's design, but with full compatibility
// with Decoder Control Procedures in Jpeg spec.
struct LilCodecJpeg {
	LilCodecJpeg(const LilIoDesc *pIoInput, LilImageDesc *pImage, const LilFormatDesc *pFormat);
	~LilCodecJpeg();

	// LilIoDesc who's read/write routines use this codec
	const LilIoDesc *GetIoOutputPtr() const { return &ioOutput; }
	const LilIoDesc *GetIoInputPtr() const { return mIo; }

	LilErrorType DecodeStart();
	LilErrorType DecodeImageInfo();
	// Prepares a single decode strip
	LilErrorType DecodeImageData();

	lil_u16 ReadMarker();
	LilErrorType InterpretMarker(lil_u16 marker);

	void DecodeReset();  // Literally resets our decoder.  There is a marker to call this...
	LilErrorType DecodeMCU();  // MCU is minimal coded unit.  Usually is a single 8x8 block in the file.

	bool IsReady() const { return mReady; }

	
	// --- VARIABLES ---

	const LilIoDesc *mIo;
	LilImageDesc *mImage;
	const LilFormatDesc *pFinalFormat;
	JpegBitReadMsb *mReader;

	// This has the result of the Codec.  Use the read function to get the result of encoding, and use
	// the write function to get the result of decoding.
	LilIoDesc ioOutput;

	//-------
	// Constant
	lil_u16 mHeight, mWidth;  // In pixels.
	lil_u8 mBits, mCh;
	lil_u8 mThumbW, mThumbH;  // In pixels.
	lil_u16 mFrameMarker;   // Which start-of-frame marker was found.  Defines way to decode!
	lil_u32 mMcuW;  // In pixels.  Subsampling will make MCUs have different counts of 8x8 data units.
	lil_u32 mMcuH;  // In pixels.
	lil_u32 mMcuCountX;  // In MCUs.
	lil_u32 mMcuCountY;
	lil_u16 mJfifVersion;

	//-------
	// Variable (possibly)
	lil_u16 mQt[4][64];  // Quantization tables
	lil_u16 mRestartInterval;
	lil_u16 mRestartCurr;  // When this reaches restart interval, we need to call DecodeReset
	HuffmanMsb *mHuffDc[4], *mHuffAc[4];  // There can only possibly be 4 of each at any one time
	
	//-------
	// Constant - but only per-scan
	lil_u8 mScanNumComp;  // Number of componenets to be processed during scan
	lil_u8 mScanCompOrder[4];

	//-------
	// Variable - changes during scan
	
	// These are used for moving decoded data to the output buffer.  We HAVE to decode at least
	// an MCU at a time, and I suspect memory access patterns will be more efficient
	// if we do an entire scan line X mcu height at a time.  THEN we'll move it to the output
	// buffer.
	bool mReady, mDecoding;
	LilBuffer mStrip;
	lil_u32 mStripBpl;
	lil_u32 mTransferOffset;  // into buffer
	lil_u32 mTransferX;  // so we know when each scanline end is hit
	lil_i32 mIter;
	lil_u32 mMcuX, mMcuY;

	//-------
	// We find out about our components when we find out about our global image data
	// Some parts simple get set, and some parts will be tracked
	struct Component {
		Component() : id(0), sampFactH(0), sampFactV(0), quantI(0), w(0), h(0), pred(0), 
			huffDcI(0), huffAcI(0) { }
		//-------
		// Constant
		lil_u8 id;
		lil_u8 sampFactH;  // horizontal sampling factor
		lil_u8 sampFactV;  // vertical sampling factor
		lil_u8 quantI;  // qantization table index
		lil_u32 w, h;  // down-sampled width and height

		//-------
		// Variable - changes during scan
		lil_i32 pred;  // dc predictor
		lil_u8 huffDcI, huffAcI;

		lil_u8 data[1024];  // 64 x 4 x 4 = 64 x max sampFactH x max sampFactV
	};
	Component mComp[4];  // Technically the spec allows for 256.  But what would you do with those components?
};

} // lil

#endif // LILCODECJPEG_H_INCLUDED
