#include <memory.h>
#include "LilCodecLzw.h"
#include "Lil.h"
#include "LilIo.h"
#include "LilMem.h"
#include "LilBitManip.h"
#include "LilUtilFormat.h"

namespace lil {

namespace {

struct Codec {
    LilCodecLzw *pHandle;
	lil_u32 offset;
	lil_u8 *pOut8;
	const lil_u8 *pIn8;
	lil_u32 bpl;
	lil_u32 rowByte;
	size_t size;
	lil_u32 byteCount;
	bool filled;

	Codec(lil_io_handle handle, lil_u32 byteCount, const void *pIn, void *pOut) :
	    pHandle(0), offset(0), pOut8(0), pIn8(0), bpl(0), rowByte(0),
        size(0), byteCount(byteCount), filled(false)
    {
        pHandle = reinterpret_cast<LilCodecLzw *>(handle);
        offset = pHandle->transferOffset;
        pOut8 = reinterpret_cast<lil_u8 *>(pOut);
        pIn8 = reinterpret_cast<const lil_u8 *>(pIn);
        bpl = CalcBpl(pHandle->pFormat, pHandle->pImage->data.width);
        rowByte = pHandle->transferRowByte;
        size = pHandle->fullImage.GetSize();
	}

	~Codec() {
	    pHandle->transferOffset = offset;
        pHandle->transferRowByte = rowByte;
	}

	bool TransferData(LilErrorType *pErr) {
		LIL_UNUSED_PARAM(pErr);
		lil_u32 count = byteCount < bpl - rowByte ? byteCount : bpl - rowByte;
		memcpy(pOut8, &pIn8[offset], count);
		pOut8 += count;
		rowByte += count;
		offset += count;
		byteCount -= count;
		if (rowByte == bpl) {
			rowByte = 0;
			if (offset == size)
				filled = !filled;
		}
		return true;
	}

	bool IsFilled() const { return filled; }
	bool IsDone() const { return byteCount == 0; }
};

//-----------------------------------------------------------------------------------------------------------
LilErrorType LilIoLzwRead(lil_io_handle handle, size_t byteCount, void *outMemPtr) {
	LilErrorType err = LE_NONE;
	// I don't believe gifs can be more than 4 gigs
    Codec codec(handle, static_cast<lil_u32>(byteCount), reinterpret_cast<LilCodecLzw *>(handle)->fullImage.GetPtr<void>(), outMemPtr);

	while (!codec.IsDone()) {
		// Get next set of data
		if (!codec.IsFilled()) {
			if (!codec.pHandle->DecodeLzw(&err))
                return err;
			codec.offset = 0;
			codec.filled = true;
		}

        if (!codec.TransferData(&err))
            return err;
	}

	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType LilIoLzwWrite(lil_io_handle handle, size_t byteCount, const void *inMemPtr) {
	LilErrorType err = LE_NONE;
    // I don't believe gifs can be more than 4 gigs
	Codec codec(handle, static_cast<lil_u32>(byteCount), inMemPtr, reinterpret_cast<LilCodecLzw *>(handle)->fullImage.GetPtr<void>());

	while (!codec.IsDone()) {
	    if (!codec.TransferData(&err))
            return err;

		// Write next set of data
		if (codec.IsFilled()) {
			if (!codec.pHandle->EncodeLzw(&err))
				return err;
			codec.offset = 0;
			codec.filled = false;
		}
	}

	return LE_NONE;
}

} // namespace



//-----------------------------------------------------------------------------------------------------------
// CODEC LZW
//-----------------------------------------------------------------------------------------------------------
LilCodecLzw::LilCodecLzw(const LilIoDesc *pIoInput, const LilImageDesc *pImage, const LilFormatDesc *pFormat,
	bool interlaced, lil_u8 minCodeSize)
	: pIoInput(pIoInput), pImage(pImage), pFormat(pFormat),
	clearCode(1 << minCodeSize), endCode((1 << minCodeSize) + 1),
	charBits(pFormat->bitsPerChannel), minCodeSize(minCodeSize),
	tree(&pImage->memFuncs, 1 << MAX_DICT_BITS),
	nextDictPos(0), prevCode(0),
	codeReader(pIoInput), codeWriter(pIoInput), needBitCount(0),
	transferOffset(0), transferRowByte(0), pInterlacing(0)
{
	// Set up IO
	ioOutput.handle = reinterpret_cast<lil_io_handle>(this);
	ioOutput.readFunc = LilIoLzwRead;
	ioOutput.writeFunc = LilIoLzwWrite;
	ioOutput.seekFunc = 0;
	ioOutput.tellFunc = 0;

	// Set up dictionary
	for (lil_u32 i = 0; i <= endCode; i++) {
		pDict[i].prefix = i;
		pDict[i].newChar = i;
		pDict[i].numPrefix = 0;
	}

	ResetDict();

	// Set up other stuff
	if (interlaced)
		pInterlacing = new GifInterlacing(pImage->data.width, pImage->data.height);

	// Need enough room to swizzle if need-be.
	fullImage.Realloc(&pImage->memFuncs, CalcByteSize(pFormat, pImage->data.width, pImage->data.height));
	decodedIo.Reset(fullImage.GetPtr<lil_u8>());
}

//-----------------------------------------------------------------------------------------------------------
LilCodecLzw::~LilCodecLzw() {
	delete pInterlacing;
}

//-----------------------------------------------------------------------------------------------------------
bool LilCodecLzw::DecodeLzw(LilErrorType *pErr) {
	lil_u32 currCode = 0;

	// END CODE
	codeReader.ReadBits(currCode, needBitCount);
	while (currCode != endCode) {
		// CLEAR CODE - SHOULD BE FIRST CODE READ
		if (currCode == clearCode) {
			ResetDict();
			while (currCode == clearCode) {
				codeReader.ReadBits(currCode, needBitCount);
				if (currCode == endCode)
                    return SetError(LE_NONE, pErr);
			}
			// Finally have a good code
			DecodeWrite(static_cast<lil_u8>(pDict[currCode].newChar));
		}
		else {
			if (currCode > nextDictPos)
				return SetError(LE_BADIMAGEDATA, pErr);

			// Output the string for the current input code
			DictEntry *pDictCurr = 0;
			if (currCode >= nextDictPos)
				pDictCurr = &pDict[prevCode];  // DOESN'T EXIST
			else
				pDictCurr = &pDict[currCode];  // EXISTS

			// Copy string in reverse order from current code to base code.
			int len = pDictCurr->numPrefix + 1;
			// len + 1 because we skip back charBits * 2 in the loop.
			DecodeSkip(len + 1);
			for (int i = 0; i < len; i++) {
				DecodeSkip(-2);
				DecodeWrite(static_cast<lil_u8>(pDictCurr->newChar));
				// We initialize the prefix for base-level entries to themselves, so now
				// on the last iteration pDictCurr just equals itself again.  It will always
				// be the FIRST character of the stream output.
				pDictCurr = &pDict[pDictCurr->prefix];
			}
			DecodeSkip(len - 1);

			// DOESN'T EXIST CASE
			if (currCode >= nextDictPos)
				DecodeWrite(static_cast<lil_u8>(pDictCurr->newChar));

			// Update the dictionary
			if (nextDictPos < (1 << MAX_DICT_BITS)) {
				pDict[nextDictPos].prefix = prevCode;
				pDict[nextDictPos].newChar = pDictCurr->newChar;
				pDict[nextDictPos].numPrefix = pDict[prevCode].numPrefix + 1;
				nextDictPos++;

				if (nextDictPos == (1U << needBitCount) && needBitCount < MAX_DICT_BITS)
					needBitCount++;
			}
		}
		prevCode = currCode;
		codeReader.ReadBits(currCode, needBitCount);
	}

	return SetError(LE_NONE, pErr);
}

//-----------------------------------------------------------------------------------------------------------
void LilCodecLzw::EncodeWrite(lil_u32 code) {
    codeWriter.WriteBits(code, static_cast<lil_u8>(needBitCount));

    nextDictPos++;

    // Use "- 1" because the next code will be read to indicate a clear code, and we want it to still
    // be needBitCount size.
    if (nextDictPos == (1u << needBitCount) + 1)
        needBitCount++;
    if (nextDictPos == (1u << MAX_DICT_BITS)) {
        codeWriter.WriteBits(clearCode, needBitCount);
        ResetDict();
    }
}

//-----------------------------------------------------------------------------------------------------------
bool LilCodecLzw::EncodeLzw(LilErrorType *pErr) {
	codeWriter.WriteBits(clearCode, needBitCount);

    TreeNode *pCurr = tree.GetRoot();
	for (lil_u32 y = 0; y < pImage->data.height; y++) {
		for (lil_u32 x = 0; x < pImage->data.width; x++) {
			lil_u8 val = static_cast<lil_u8>(decodedIo.ReadBits(static_cast<lil_u8>(charBits)));
			TreeNode *pLast = pCurr;
            pCurr = tree.FindChild(pLast, val);
            if (pCurr)
                continue;

            tree.AddChild(pLast, val, nextDictPos);
            EncodeWrite(pLast->value);
            pCurr = tree.FindChild(tree.GetRoot(), val);
		}
		decodedIo.AlignToByte();
	}

    if (pCurr)
        EncodeWrite(pCurr->value);
	codeWriter.WriteBits(endCode, needBitCount);

	return SetError(LE_NONE, pErr);
}

//-----------------------------------------------------------------------------------------------------------
void LilCodecLzw::ResetDict() {
	// Give room for clear-code and end-code
	nextDictPos = endCode + 1;
	needBitCount = minCodeSize + 1;

	tree.Reset();
	for (lil_u8 i = 0; i < clearCode; i++)
        tree.AddChild(tree.GetRoot(), i, i);
}

}  // namespace lil
