#ifndef LILCODECLZW_H_INCLUDED
#define LILCODECLZW_H_INCLUDED

#include "Lil.h"
#include "LilMem.h"
#include "LilBitManip.h"
#include <string>
#include <map>

namespace lil {

// Both tif and gif max out at 12 bits
const lil_u32 MAX_DICT_BITS = 12;

//-----------------------------------------------------------------------------------------------------------
// Can do some ifdef'ing if need to have some sort of trade off for
// speed vs memory consumption... but even at 12 bits max dictionary size
// the dictionary is < 64k
struct DictEntry {
    DictEntry() : newChar(0xffffffff), prefix(0xffffffff), numPrefix(0xffffffff) { }
	lil_u32 newChar;
	lil_u32 prefix;
	lil_u32 numPrefix;
};

//-----------------------------------------------------------------------------------------------------------
struct TreeNode {
    TreeNode() : id(0), value(0), pNext(0), pChild(0), pLastChild(0) { }
    TreeNode(lil_u8 id, lil_u32 value) : id(id), value(value), pNext(0), pChild(0), pLastChild(0) { }
    lil_u8 id;
    lil_u32 value;
    TreeNode *pNext, *pChild, *pLastChild;
};
struct Tree {
    Tree(const LilMemDesc *memFuncs, lil_u32 maxSize) : pMemFuncs(memFuncs) {
        buf.Realloc(pMemFuncs, maxSize * sizeof(TreeNode));
        pos = 0;
    }

    TreeNode *AddChild(TreeNode *parent, lil_u8 id, lil_u32 value) {
        LIL_ASSERT(parent);
        LIL_ASSERT(pos * sizeof(TreeNode) < buf.GetSize());
        TreeNode *retVal = new (&buf.GetPtr<TreeNode>()[pos]) TreeNode(id, value);
        pos++;
        if (parent->pLastChild)
            parent->pLastChild->pNext = retVal;
        else
            parent->pChild = retVal;
        parent->pLastChild = retVal;
        return retVal;
    }

    TreeNode *FindChild(TreeNode *parent, lil_u8 id) {
        LIL_ASSERT(parent);
        TreeNode *retVal = parent->pChild;
        while (retVal && retVal->id != id)
            retVal = retVal->pNext;
        return retVal;
    }

    void Reset() {
        pos = 0;
        root.pChild = 0;
        root.pNext = 0;
        root.pLastChild = 0;
        root.id = 0;
        root.value = 0;
    }

    LIL_INLINE TreeNode *GetRoot() { return &root; }

    const LilMemDesc *pMemFuncs;
    lil_u32 pos;
    LilBuffer buf;
    TreeNode root;
};

//-----------------------------------------------------------------------------------------------------------
struct GifInterlacing {
	GifInterlacing(lil_u32 w, lil_u32 h) : width(w), height(h), x(0), y(0), passNum(0) { }

	// Returns true if the scanline position was changed.
	// Otherwise returns false.
	bool Offset(lil_u32 charOffset) {
		// We need the extra one on the end in case we skip too far ahead as will happen
		// during interlaced decoding.
		x += charOffset;
		// Checks for clipping in one compare
		if (static_cast<lil_u32>(x) < width)
			return false;

		static lil_u32 intInfo[5][2] = { { 0, 8 }, { 4, 8 }, { 2, 4 }, { 1, 2 }, { 1, 2 } };
		if (x < 0) {
			while (x < 0)
				x += width;
			y -= intInfo[passNum][1];
			if (y < 0) {
				passNum --;
				lil_u32 base = intInfo[passNum][0];
				y = base + RoundDownToNearest(height - 1 - base, intInfo[passNum][1]);
			}
			return true;
		}

		// This case is the only possibility left
//		if (x >= width) {
			x = x % width;
			y += intInfo[passNum][1];
			if (y >= static_cast<lil_i32>(height)) {
				passNum ++;
				y = intInfo[passNum][0];
			}
			return true;
//		}
	}

	lil_u32 width, height;
	lil_i32 x, y;
	lil_u8 passNum;
};

// This is made specifically for variable-length bit codes.
// Assumes clear code at 1 << StartBits, and End-Of-Info at (1 << StartBits) + 1
// This is consistent with gif and tif.
// So table is always 1 << (StartBits + 1) in size at initialization.
// ENCODE should always output a clear code when (1 << maxBits) - 2 is encountered
// or 4094 for 12 bit maxBits.  Etc.
struct LilCodecLzw {
	LilCodecLzw(const LilIoDesc *pIoInput, const LilImageDesc *pImage, const LilFormatDesc *pFormat,
		bool interlaced, lil_u8 minCodeSize);
	~LilCodecLzw();

	// LilIoDesc who's read/write routines use this codec
	virtual const LilIoDesc *GetIoPtr() const { return &ioOutput; }

	// Called after filling the buffer in some manner.
	bool DecodeLzw(LilErrorType *pErr);
	bool EncodeLzw(LilErrorType *pErr);

private:
	// Clears the dictionary to initial state.
	void ResetDict();

	// Pixel buffer I/O - writing
	LIL_INLINE void DecodeWrite(lil_u8 newChar) {
		decodedIo.WriteBits(static_cast<lil_u8>(charBits), newChar);
		UpdateInterlacing(1);
	}

	// Pixel buffer I/O - skipping position
	// Notice the ability to skip backwards...
	LIL_INLINE void DecodeSkip(lil_i32 chars) {
		decodedIo.SkipBits(chars * charBits);
		UpdateInterlacing(chars);
	}

	// Pixel buffer I/O - helper
	LIL_INLINE void UpdateInterlacing(lil_u32 charOffset) {
		if (pInterlacing && pInterlacing->Offset(charOffset)) {
			lil_u32 offsetBits = charBits * (pInterlacing->y * pInterlacing->width + pInterlacing->x);
			decodedIo.SetBitPos(offsetBits);
		}
	}

    void EncodeWrite(lil_u32 code);

public:
	// ENCODE AND DECODE
	// This will be the what goes into the codec (whether it's data to decode, or data to encode)
	const LilIoDesc *pIoInput;
	// Holds information like width and height... DONT use for image data - that will come
	// to/from memory pointers.
	const LilImageDesc *pImage;
	// This may be different from the format the image is in.
	// When decoding, it is the format we want to convert to ultimately.
	// When encoding, it is the format we want to be in the file.
	const LilFormatDesc *pFormat;

	// Special codes
	lil_u8 clearCode, endCode;

	// signed int so we can traverse forward and backwards when bit-packed
	lil_i32 charBits;
	lil_u8 minCodeSize;

	//----------- TO REFACTOR
	// Buffer for dictionary and lzw data
	DictEntry pDict[1 << MAX_DICT_BITS];
	Tree tree;

	// Dictionary usage
	lil_u32 nextDictPos, prevCode;

	// For retrieving codes
	BitReadLsb codeReader;
	BitWriteLsb codeWriter;
	lil_u8 needBitCount;

	// Output tracking
	BitIoMsb decodedIo;
	lil_u32 transferOffset;
	lil_u32 transferRowByte;

	GifInterlacing *pInterlacing;

	LilBuffer fullImage;

private:
	// This has the result of the Codec.  Use the read function to get the result of encoding, and use
	// the write function to get the result of decoding.
	LilIoDesc ioOutput;
};

}  // namespace lil

#endif // LILCODECLZW_H_INCLUDED
