// http://en.wikipedia.org/wiki/PackBits

#include <memory.h>
#include "LilCodecRle.h"
#include "Lil.h"
#include "LilIo.h"
#include "LilMem.h"

namespace lil {

class RleEncoder {
public:
	RleEncoder(lil_u32 maxLen, lil_u32 width, lil_u32 pixBytes) : mMaxLen(maxLen), mWriteSize(0), mReadPos(0), mRleMode(false),
		mW(width), mWriteX(0), mPixBytes(pixBytes), mPixByte(0), mReady(false), mReadSize(0), mReadX(0) { }
	~RleEncoder() { }

	LIL_INLINE void AddData(lil_u8 data) {
		mBuf[mWriteSize++] = data;
		if (mPixBytes == 1) {
			mWriteX++;
			if (mWriteSize > 3 && (mRleMode != (mBuf[mWriteSize - 1] == mBuf[mWriteSize - 2]))) {
				mReadSize = mWriteSize - 1;
				mReady = true;
			}
			else {
				if (mWriteSize == 3)
					mRleMode = (mBuf[0] == mBuf[1] && mBuf[0] == mBuf[2]);
				if ((mWriteSize == mMaxLen) || (mWriteX == mW))
					HardStop();
			}
		}
		else {
			mPixByte++;
			if (mWriteSize > 3 * mPixBytes) {
				bool checkRle = true;
				for (lil_u32 i = 0; i < mPixBytes; i++)
					if (mBuf[mWriteSize - 1 - i] != mBuf[mWriteSize - 1 - i - mPixBytes])
						checkRle = false;
				if (checkRle != mRleMode) {
					mReadSize = mWriteSize - mPixByte;
					mReady = true;
				}
			}
			if (mPixByte == mPixBytes) {
				mPixByte = 0;
				mWriteX++;  // increase when we're done with this pixel
			}
			if (!mReady) {
				if (mWriteSize == 3 * mPixBytes) {
					mRleMode = true;
					for (lil_u32 i = 0; i < mPixBytes; i++)
						if (mBuf[i] != mBuf[i + mPixBytes] || mBuf[i] != mBuf[i + mPixBytes + mPixBytes])
							mRleMode = false;
				}
				if ((mWriteSize == mMaxLen * mPixBytes) || (mWriteX == mW))
					HardStop();
			}
		}
	}

	LIL_INLINE void HardStop() {
		mReadSize = mWriteSize;
		mReady = true;
	}
	LIL_INLINE bool IsReady() const { return mReady; }
	LIL_INLINE bool IsRleMode() const { return mRleMode; }
	LIL_INLINE bool IsEOL() const { return IsFoundEOL() && mReadSize == mWriteSize; }
	LIL_INLINE bool IsFoundEOL() const { return mWriteX == mW; }

	LIL_INLINE lil_u8 GetData() const {
		lil_u8 ret = mBuf[mReadPos++];
		LIL_ASSERT(mReadPos <= mReadSize);
		return ret;
	}

	LIL_INLINE lil_u32 GetCount() const { return mReadSize / mPixBytes; }

	// Call after every IsReady == true followed by reading all of mReadSize bytes.
	LIL_INLINE void Reset() {
		for (lil_u32 i = mReadSize; i < mWriteSize; i++)
			mBuf[i - mReadSize] = mBuf[i];
		mWriteSize -= mReadSize;
		mReadPos = 0;
		mRleMode = false;

		if (mWriteX == mW) {
			if (mWriteSize)
				mReadSize = mWriteSize;
			else {
				mReady = false;
				mWriteX = 0;
				mPixByte = 0;
			}
		}
		else
			mReady = false;
	}

private:
	lil_u32 mMaxLen;
	lil_u8 mBuf[257 * 8];  // Allow for up to 16 bit X 4 channels
	lil_u32 mWriteSize, mReadSize;
	mutable lil_u32 mReadPos;
	bool mRleMode, mReady;
	lil_u32 mWriteX, mReadX;
	lil_u32 mW;
	lil_u32 mPixBytes, mPixByte;
};



namespace {

//-----------------------------------------------------------------------------------------------------------
// CODEC RLE BMP
//-----------------------------------------------------------------------------------------------------------
LilErrorType LilIoRleBmpRead(lil_io_handle handle, size_t byteCount, void *outMemPtr) {
	LilCodecRleBmp *pHandle = reinterpret_cast<LilCodecRleBmp *>(handle);
	const LilIoDesc *pIo = pHandle->pIoInput;
	lil_u8 *out8 = reinterpret_cast<lil_u8 *>(outMemPtr);

	// Image will contain the format IN THE FILE... we don't want to use the format
	// requested by the user... that is for post-conversion.
	lil_u32 lineBytes = CalcBpl(pHandle->pFormat, pHandle->pImage->data.width);
	lil_u32 bytePix = 1 << pHandle->shift;

	while (byteCount) {
		// Read next command
		if (!pHandle->ready) {
			pHandle->ready = true;
			pHandle->buf[0] = Read8(pIo);
			pHandle->buf[1] = Read8(pIo);
			pHandle->bytesPending = (pHandle->buf[0] + pHandle->shift) >> pHandle->shift;
			if (pHandle->bytesPending)
				pHandle->x += bytePix * pHandle->bytesPending;
			else {
				if (pHandle->buf[1] == 0) {  // end of line
					pHandle->bytesPending = CalcRemainder(pHandle->totalDecoded, lineBytes);
					pHandle->y++;
					pHandle->x = 0;
					if (!pHandle->bytesPending)
						pHandle->ready = false;
				}
				else if (pHandle->buf[1] == 1) // end of bitmap
					pHandle->bytesPending = 
						CalcByteSize(pHandle->pFormat, pHandle->pImage->data.width, pHandle->pImage->data.height) - 
						pHandle->totalDecoded;
				else if (pHandle->buf[1] == 2) {  // delta x and y... skip 'dont care' pixels
					pHandle->buf[2] = Read8(pIo);
					pHandle->buf[3] = Read8(pIo);
					pHandle->bytesPending = (pHandle->buf[3] - pHandle->y) * lineBytes +
						((pHandle->buf[2] - pHandle->x + pHandle->shift) >> pHandle->shift);
					if (!pHandle->bytesPending)
						pHandle->ready = false;
				}
				else // (pHandle->buf[1] >= 3)
					pHandle->bytesPending = pHandle->buf[1] >> pHandle->shift;  // If shift, should always be even number!
			}
		}
		// Write decoded data
		else {
			if (pHandle->buf[0])  // RLE
				*out8++ = pHandle->buf[1];
			else if (pHandle->buf[1] >= 3)  // Absolute
				*out8++ = Read8(pIo);
			else
				*out8++ = 0;  // It's something else but the pixels don't matter... end of line or delta or something

			pHandle->bytesPending--;
			if (pHandle->bytesPending == 0) {
				pHandle->ready = false;
				// If reading absolute pixels, file position MUST be word aligned...
				if (pHandle->buf[0] == 0 && pHandle->buf[1] >= 3 && ((pHandle->buf[1] + pHandle->shift) >> pHandle->shift) & 1)
					Read8(pIo);
			}
			byteCount--;
			pHandle->totalDecoded++;
		}
	}

	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType LilIoRleBmpWrite(lil_io_handle handle, size_t byteCount, LIL_CONST void *inMemPtr) {
	LilErrorType err = LE_NONE;
	LilCodecRleBmp *pHandle = reinterpret_cast<LilCodecRleBmp *>(handle);
	const LilIoDesc *pIo = pHandle->pIoInput;
	const lil_u8 *in8 = reinterpret_cast<const lil_u8 *>(inMemPtr);

	// pFormat will contain the format IN MEMORY... which is what was in the original image.
	// we don't want to use the format originally in the image - we want the dest. image's format.
	lil_u32 lineBytes = CalcBpl(pHandle->pFormat, pHandle->pImage->data.width);
	lil_u32 bytePix = 1 << pHandle->shift;

	while (byteCount) {
		pHandle->encoder->AddData(*in8);
		// @@ Test this case below when we find EOL... maybe have separate IsFoundEOL function or something
		lil_u32 addBytes = pHandle->encoder->IsFoundEOL() ? CalcRemainder(pHandle->totalDecoded, lineBytes) : 1;
		in8 += addBytes;
		byteCount -= addBytes;
		pHandle->totalDecoded += addBytes;
		while (pHandle->encoder->IsReady()) {
			// Write data
			lil_u32 count = pHandle->encoder->GetCount();
			if (count < 3)
				// If this happens.. it HAS to be single rle runs since absolute runs MUST be at least 3 length
				for (lil_u32 i = 0; i < count; i++)
					WriteFle(pIo, "11", static_cast<lil_u8>(bytePix), pHandle->encoder->GetData());
			else {
				if (pHandle->encoder->IsRleMode())  // RLE run
					WriteFle(pIo, "11", static_cast<lil_u8>(count << pHandle->shift), pHandle->encoder->GetData());
				else {  // Absolute run
					LIL_ASSERT(count >= 3);
					WriteFle(pIo, "11", (lil_u8)0, static_cast<lil_u8>(count << pHandle->shift));
					for (lil_u32 i = 0; i < count; i++)
						Write8(pIo, pHandle->encoder->GetData());
					if (count & 1)  // Must be word padded
						Write8(pIo, 0);
				}
			}
			if (pHandle->encoder->IsEOL()) {
				pHandle->y++;
				if (pHandle->y == static_cast<lil_i32>(pHandle->pImage->data.height)) {
					// End of image
					err = Write16be(pIo, 0x0001);
					LIL_ASSERT(byteCount == 0);
				}
				else // End of line
					err = Write16le(pIo, 0x0000);
				if (err != LE_NONE)
					return err;
			}
			pHandle->encoder->Reset();
		}
	}
	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
// CODEC RLE TGA
//-----------------------------------------------------------------------------------------------------------
LilErrorType LilIoRleTgaRead(lil_io_handle handle, size_t byteCount, void *outMemPtr) {
	LilCodecRleTga *pHandle = reinterpret_cast<LilCodecRleTga *>(handle);
	const LilIoDesc *pIo = pHandle->pIoInput;
	lil_u8 *pOut8 = reinterpret_cast<lil_u8 *>(outMemPtr);

	while (byteCount) {
		// Get next set of data
		if (!pHandle->ready) {
			pHandle->ready = true;
			pHandle->buf[0] = Read8(pIo);
			if (pHandle->buf[0] < 128)  // Absolute
				pHandle->bytesPending = (1 + pHandle->buf[0]) * pHandle->pixBytes;
			else {  // RLE
				pHandle->bytesPending = (pHandle->buf[0] - 127) * pHandle->pixBytes;
				ReadN(pIo, pHandle->pixBytes, &pHandle->buf[1]);
			}
		}
		// Write next decoded data
		if (pHandle->buf[0] < 128)
			*pOut8++ = Read8(pIo);
		else
			*pOut8++ = pHandle->buf[pHandle->pixByte];
		// Stride is always aligned to a single byte, so no need to compensate

		pHandle->pixByte++;
		if (pHandle->pixByte > pHandle->pixBytes)
			pHandle->pixByte = 1;
		pHandle->bytesPending--;
		if (pHandle->bytesPending == 0)
			pHandle->ready = false;
		byteCount--;
		pHandle->totalDecoded++;
	}
	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType LilIoRleTgaWrite(lil_io_handle handle, size_t byteCount, LIL_CONST void *inMemPtr) {
	LilCodecRleTga *pHandle = reinterpret_cast<LilCodecRleTga *>(handle);
	const LilIoDesc *pIo = pHandle->pIoInput;
	const lil_u8 *in8 = reinterpret_cast<const lil_u8 *>(inMemPtr);

	// pFormat will contain the format IN MEMORY... which is what was in the original image.
	// we don't want to use the format originally in the image - we want the dest. image's format.
	lil_u32 lineBytes = CalcBpl(pHandle->pFormat, pHandle->pImage->data.width);

	while (byteCount) {
		pHandle->encoder->AddData(*in8);
		// @@ Likewise to bmp writer
		lil_u32 addBytes = pHandle->encoder->IsFoundEOL() ? CalcRemainder(pHandle->totalDecoded, lineBytes) : 1;
		in8 += addBytes;
		byteCount -= addBytes;
		pHandle->totalDecoded += addBytes;
		while (pHandle->encoder->IsReady()) {
			lil_u32 count = pHandle->encoder->GetCount();
			if (pHandle->encoder->IsRleMode()) {  // RLE run
				Write8(pIo, static_cast<lil_u8>(count + 127));
				for (lil_u32 i = 0; i < pHandle->pixBytes; i++)
					Write8(pIo, pHandle->encoder->GetData());
			}
			else {  // Absolute run
				Write8(pIo, static_cast<lil_u8>(count - 1));
				for (lil_u32 i = 0; i < count * pHandle->pixBytes; i++)
					Write8(pIo, pHandle->encoder->GetData());
			}
			pHandle->encoder->Reset();
		}
	}
	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
// CODEC PACKBITS
//-----------------------------------------------------------------------------------------------------------
LilErrorType LilIoPackbitsRead(lil_io_handle handle, size_t byteCount, void *outMemPtr) {
	LilCodecPackbits *pHandle = reinterpret_cast<LilCodecPackbits *>(handle);
	const LilIoDesc *pIo = pHandle->pIoInput;
	lil_i8 *pOut8 = reinterpret_cast<lil_i8 *>(outMemPtr);
	const lil_i8 *pIn8 = &pHandle->rawBuf.GetPtr<const lil_i8>()[pHandle->rawBufPos];

	if (pHandle->totalEncoded == 0 && pHandle->preImage) {
		lil_u32 count = pHandle->pImage->data.height * pHandle->pFormat->numChannels;
		for (lil_u32 i = 0; i < count; i++)
			pHandle->pRowBytes[i] = Read16be(pIo);
	}

	while (byteCount) {
		// Get next set of data
		if (!pHandle->ready) {
			// Read data
			pHandle->ready = true;

			if (!pHandle->preImage)
				*pHandle->pRowBytes = Read16be(pIo);
			pHandle->currRowEnd += *pHandle->pRowBytes++;
			lil_u8 *pBuf8 = pHandle->rawBuf.GetPtr<lil_u8>();
			while (pHandle->totalEncoded < pHandle->currRowEnd) {
				lil_i8 cByte = -128;
				while (cByte == -128) {
					cByte = static_cast<lil_i8>(Read8(pIo));
					pHandle->totalEncoded++;
				}
				if (cByte >= 0) {  // Absolute
					lil_u32 count = 1 + cByte;
					ReadN(pIo, count, pBuf8);
					pBuf8 += count;
					pHandle->bytesPending += count;
					pHandle->totalEncoded += count;
				}
				else {  // RLE
					lil_u32 count = 1 - cByte;
					lil_u8 val = Read8(pIo);
					for (lil_u32 i = 0; i < count; i++)
						*pBuf8++ = val;
					pHandle->bytesPending += count;
					pHandle->totalEncoded++;
				}
			}
			if (pHandle->totalEncoded != pHandle->currRowEnd)
				return LE_BADIMAGEDATA;
			pHandle->rawBufPos = 0;
			if (!pHandle->pFormat->planar && pHandle->pFormat->numChannels > 1 &&
				pHandle->pFormat->dataFormat == LDF_SIMPLE)
			{
				// Use second half of raw buffer to store reorderning of the 1st half of raw buffer
				lil_u32 ch = pHandle->pFormat->numChannels;
				lil_u32 len = pHandle->rawBuf.GetSize() / 2;
				const lil_u8 *pSrcBuf8 = pHandle->rawBuf.GetPtr<lil_u8>();
				for (lil_u32 iP = 0; iP < ch; iP++)
					for (lil_u32 i = iP; i < len; i += ch)
						pBuf8[i] = *pSrcBuf8++;
				pHandle->rawBufPos = len;
			}

			pIn8 = &pHandle->rawBuf.GetPtr<const lil_i8>()[pHandle->rawBufPos];
		}

		lil_u32 count = byteCount < pHandle->bytesPending ? byteCount : pHandle->bytesPending;
		memcpy(pOut8, pIn8, count);
		pOut8 += count;
		pIn8 += count;
		byteCount -= count;
		pHandle->bytesPending -= count;
		// Stride is always aligned to a single byte, so no need to compensate
		if (pHandle->bytesPending == 0)
			pHandle->ready = false;

		pHandle->totalDecoded += count;
	}

	pHandle->rawBufPos = pIn8 - (&pHandle->rawBuf.GetPtr<const lil_i8>()[pHandle->rawBufPos]);
	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType LilIoPackbitsWrite(lil_io_handle handle, size_t byteCount, LIL_CONST void *inMemPtr) {
	LilCodecPackbits *pHandle = reinterpret_cast<LilCodecPackbits *>(handle);
	const LilIoDesc *pIo = pHandle->pIoInput;
	const lil_i8 *in8 = reinterpret_cast<const lil_i8 *>(inMemPtr);
	lil_u8 *out8 = pHandle->rawBuf.GetPtr<lil_u8>();

	// pFormat will contain the format IN MEMORY... which is what was in the original image.
	// we don't want to use the format originally in the image - we want the dest. image's format.
	lil_u32 lineBytes = CalcBpl(pHandle->pFormat, pHandle->pImage->data.width);

	if (pHandle->totalEncoded == 0 && pHandle->preImage) {
		pHandle->rowByteStart = pIo->tellFunc(pIo->handle);
		lil_u32 rowCounts = pHandle->pImage->data.height *
			(pHandle->pFormat->planar ? pHandle->pFormat->numChannels : 1);
		pIo->seekFunc(pIo->handle, rowCounts * 2, LS_RELATIVE);
	}

	while (byteCount) {
		out8[pHandle->rawBufPos++] = *in8++;
		byteCount--;
		// Need to get a line of input before we can compress it
		if (pHandle->rawBufPos == lineBytes) {
			// Now see if we need to split scan line into separate components
			lil_u8 *pBuf8 = out8;
			if (!pHandle->pFormat->planar && pHandle->pFormat->numChannels > 1 &&
				pHandle->pFormat->dataFormat == LDF_SIMPLE)
			{
				if (pHandle->pFormat->numChannels != 3)
					return LE_UNSUPPORTEDFILEFEATURE;
				// Use second half of raw buffer to store reorderning of the 1st half of raw buffer
				pBuf8 += lineBytes;
				size_t pos = 0;
				for (lil_u32 iP = 0; iP < 3; iP++)
					for (lil_u32 i = iP; i < lineBytes; i += 3)
						pBuf8[pos++] = out8[i];
			}

			if (!pHandle->preImage)
				Write16be(pIo, 0);

			pHandle->rawBufPos = 0;
			pHandle->totalDecoded += lineBytes;

			for (lil_u32 i = 0; i < lineBytes; i++) {
				pHandle->encoder->AddData(pBuf8[i]);

				if (pHandle->encoder->IsReady()) {
                    while (pHandle->encoder->IsReady()) {
                        lil_i32 count = static_cast<lil_i32>(pHandle->encoder->GetCount());
                        LIL_ASSERT(count <= 128);
                        if (pHandle->encoder->IsRleMode()) {  // RLE run
                            Write8(pIo, static_cast<lil_u8>(1 - count));
                            Write8(pIo, pHandle->encoder->GetData());
                            pHandle->totalEncoded += 2;
                        }
                        else {  // Absolute run
                            Write8(pIo, static_cast<lil_u8>(count - 1));
                            for (lil_i32 j = 0; j < count; j++)
                                Write8(pIo, pHandle->encoder->GetData());
                            pHandle->totalEncoded += count + 1;
                        }
                        if (pHandle->encoder->IsEOL()) {
                            //LIL_ASSERT(i == lineBytes - 1);
                            if (pHandle->pRowBytes)
                                pHandle->pRowBytes[pHandle->yPos++] = pHandle->totalEncoded - pHandle->currRowEnd;
                            pHandle->currRowEnd = pHandle->totalEncoded;
                            if (!pHandle->preImage) {
                                lil_i32 bytes = static_cast<lil_i32>(pHandle->pRowBytes[pHandle->yPos - 1]);
                                pIo->seekFunc(pIo->handle, -bytes - 2, LS_RELATIVE);
                                Write16be(pIo, static_cast<lil_u16>(bytes));
                                pIo->seekFunc(pIo->handle, bytes, LS_RELATIVE);
                            }
                        }
                        pHandle->encoder->Reset();
                    }
				}
			}

		}
	}

	if (pHandle->totalDecoded == CalcByteSize(pHandle->pFormat, pHandle->pImage->data.width, pHandle->pImage->data.height) && pHandle->preImage) {
		lil_u32 rowCounts = pHandle->pImage->data.height *
			(pHandle->preImage ? pHandle->pFormat->numChannels : 1);
		pIo->seekFunc(pIo->handle, pHandle->rowByteStart, LS_SET);
		for (lil_u32 i = 0; i < rowCounts; i++)
			Write16be(pIo, static_cast<lil_u16>(pHandle->pRowBytes[i]));
	}

	return LE_NONE;
}

} // namespace



//-----------------------------------------------------------------------------------------------------------
// CODEC RLE BMP
//-----------------------------------------------------------------------------------------------------------
LilCodecRleBmp::LilCodecRleBmp(const LilIoDesc *pIoInput, const LilImageDesc *pImage, const LilFormatDesc *pFormat, bool bits4) :
	pIoInput(pIoInput), pImage(pImage), pFormat(pFormat ? pFormat : &pImage->format), shift(bits4 ? 1 : 0), encoder(0),
	y(0), x(0), totalDecoded(0), bufSize(0), bytesPending(0), ready(false)
{
	encoder = new RleEncoder(255 >> (bits4 ? 1 : 0), (pImage->data.width + shift) >> shift, 1);
	ioOutput.handle = reinterpret_cast<lil_io_handle>(this);
	ioOutput.readFunc = LilIoRleBmpRead;
	ioOutput.writeFunc = LilIoRleBmpWrite;
	ioOutput.seekFunc = 0;
	ioOutput.tellFunc = 0;
}

//-----------------------------------------------------------------------------------------------------------
LilCodecRleBmp::~LilCodecRleBmp() {
	delete encoder;
}

//-----------------------------------------------------------------------------------------------------------
// CODEC RLE TGA
//-----------------------------------------------------------------------------------------------------------
// Differences from PackBits
// - Pixel units, not Byte units
// - Slight algorithmic change which can be seen in code
//   (Mask 7 bits and add 1 to get rep value.  8th bit is rle or direct)
LilCodecRleTga::LilCodecRleTga(const LilIoDesc *pIoInput, const LilImageDesc *pImage, const LilFormatDesc *pFormat) :
	pIoInput(pIoInput), pImage(pImage), pFormat(pFormat ? pFormat : &pImage->format), encoder(0),
	totalDecoded(0), pixBytes(0), ready(false), bytesPending(0), pixByte(1)
{
	pixBytes = CalcBpp(pFormat) / 8;

	encoder = new RleEncoder(128, pImage->data.width, pixBytes);
	ioOutput.handle = reinterpret_cast<lil_io_handle>(this);
	ioOutput.readFunc = LilIoRleTgaRead;
	ioOutput.writeFunc = LilIoRleTgaWrite;
	ioOutput.seekFunc = 0;
	ioOutput.tellFunc = 0;
}

//-----------------------------------------------------------------------------------------------------------
LilCodecRleTga::~LilCodecRleTga() {
	delete encoder;
}

//-----------------------------------------------------------------------------------------------------------
// CODEC PACKBITS
//-----------------------------------------------------------------------------------------------------------
LilCodecPackbits::LilCodecPackbits(const LilIoDesc *pIoInput, const LilImageDesc *pImage,
								   const LilFormatDesc *pFormat, lil_u32 *pRowBytes, bool preImage)
	: pIoInput(pIoInput), pImage(pImage), pFormat(pFormat ? pFormat : &pImage->format), encoder(0), 
	totalEncoded(0), totalDecoded(0), ready(false), bytesPending(0), resetAtZero(0), pRowBytes(pRowBytes),
	preImage(preImage), currRowEnd(0), rawBufPos(0), rowByteStart(0), yPos(0)
{
	if (pImage->format.bitsPerChannel < 8) {
		lil_u32 w = (pImage->data.width * pImage->format.bitsPerChannel + 7) / 8;
		encoder = new RleEncoder(127, w, 1);
	}
	else {
		LIL_ASSERT((pImage->format.bitsPerChannel == 8 && pImage->format.dataFormat == LDF_SIMPLE) ||
			(CalcBpp(pFormat) == 16 && pImage->format.dataFormat == LDF_SIMPLEBITS));
		// ALWAYS used planes
		lil_u32 w = pImage->data.width;// * (pImage->format.planar ? 1 : pImage->format.numChannels);
		encoder = new RleEncoder(127, w, pImage->format.dataFormat == LDF_SIMPLE ? 1 : 2);
	}
	ioOutput.handle = reinterpret_cast<lil_io_handle>(this);
	ioOutput.readFunc = LilIoPackbitsRead;
	ioOutput.writeFunc = LilIoPackbitsWrite;
	ioOutput.seekFunc = 0;
	ioOutput.tellFunc = 0;

	// Need enough room to swizzle if need-be.
	rawBuf.Realloc(&pImage->memFuncs, CalcBpl(pFormat, pImage->data.width) * 2);
}

//-----------------------------------------------------------------------------------------------------------
LilCodecPackbits::~LilCodecPackbits() {
	delete encoder;
}

}  // namespace lil
