#include <algorithm>
#include "LilConvert.h"
#include "LilColorspace.h"
#include "LilMem.h"
#include "LilBitManip.h"
#include "LilUtilFormat.h"

namespace lil {

//-----------------------------------------------------------------------------------------------------------
namespace {
	// For each type (this is the dest)
	//   For each type (this is the source)
	//     Mapping of channel to channel for the number of possible channels
	const lil_u8 DestToSrcLut[15][15][LIL_MAX_CHANNELS] = {
		{
			// LCO_UNKNOWN dest
			{ 0, 1, 2, 3, 4, 5, 6, 7 }, // LCO_UNKNOWN src
			{ 0, 0, 0, 0, 0, 0, 0, 0 }, // LCO_SINGLE src
			{ LIL_CHANNEL_ERROR }, // LCO_YA src
			{ LIL_CHANNEL_ERROR }, // LCO_RGB src
			{ LIL_CHANNEL_ERROR }, // LCO_BGR src
			{ LIL_CHANNEL_ERROR }, // LCO_ARGB src
			{ LIL_CHANNEL_ERROR }, // LCO_ABGR src
			{ LIL_CHANNEL_ERROR }, // LCO_RGBA src
			{ LIL_CHANNEL_ERROR }, // LCO_BGRA src
			{ LIL_CHANNEL_ERROR }, // LCO_CMY src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYA src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYK src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYKA src
			{ LIL_CHANNEL_ERROR }, // LCO_YCBCR src
			{ LIL_CHANNEL_ERROR }, // LCO_YCBCRA src
		},
		{
			// LCO_SINGLE dest
			{ LIL_CHANNEL_ERROR }, // LCO_UNKNOWN src
			{ 0 }, // LCO_SINGLE src
			{ 0 }, // LCO_YA src
			{ LIL_CHANNEL_CONVERT }, // LCO_RGB src
			{ LIL_CHANNEL_CONVERT }, // LCO_BGR src
			{ LIL_CHANNEL_CONVERT }, // LCO_ARGB src
			{ LIL_CHANNEL_CONVERT }, // LCO_ABGR src
			{ LIL_CHANNEL_CONVERT }, // LCO_RGBA src
			{ LIL_CHANNEL_CONVERT }, // LCO_BGRA src
			{ LIL_CHANNEL_ERROR }, // LCO_CMY src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYA src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYK src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYKA src
			{ 0 }, // LCO_YCBCR src
			{ 0 }, // LCO_YCBCRA src
		},
		{
			// LCO_YA dest
			{ LIL_CHANNEL_ERROR }, // LCO_UNKNOWN src
			{ 0, LIL_CHANNEL_NONE }, // LCO_SINGLE src
			{ 0, 1 }, // LCO_YA src
			{ LIL_CHANNEL_CONVERT, LIL_CHANNEL_NONE }, // LCO_RGB src
			{ LIL_CHANNEL_CONVERT, LIL_CHANNEL_NONE }, // LCO_BGR src
			{ LIL_CHANNEL_CONVERT, 0 }, // LCO_ARGB src
			{ LIL_CHANNEL_CONVERT, 0 }, // LCO_ABGR src
			{ LIL_CHANNEL_CONVERT, 3 }, // LCO_RGBA src
			{ LIL_CHANNEL_CONVERT, 3 }, // LCO_BGRA src
			{ LIL_CHANNEL_ERROR }, // LCO_CMY src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYA src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYK src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYKA src
			{ 0, LIL_CHANNEL_NONE }, // LCO_YCBCR src
			{ 0, 3 }, // LCO_YCBCRA src
		},
		{
			// LCO_RGB dest
			{ LIL_CHANNEL_ERROR }, // LCO_UNKNOWN src
			{ 0, 0, 0 }, // LCO_SINGLE src
			{ 0, 0, 0 }, // LCO_YA src
			{ 0, 1, 2 }, // LCO_RGB src
			{ 2, 1, 0 }, // LCO_BGR src
			{ 1, 2, 3 }, // LCO_ARGB src
			{ 3, 2, 1 }, // LCO_ABGR src
			{ 0, 1, 2 }, // LCO_RGBA src
			{ 2, 1, 0 }, // LCO_BGRA src
			{ LIL_CHANNEL_ERROR }, // LCO_CMY src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYA src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYK src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYKA src
			{ LIL_CHANNEL_ERROR }, // LCO_YCBCR src
			{ LIL_CHANNEL_ERROR }, // LCO_YCBCRA src
		},
		{
			// LCO_BGR dest
			{ LIL_CHANNEL_ERROR },
			{ 0, 0, 0 },
			{ 0, 0, 0 },
			{ 2, 1, 0 },
			{ 0, 1, 2 },
			{ 3, 2, 1 },
			{ 1, 2, 3 },
			{ 2, 1, 0 },
			{ 0, 1, 2 },
			{ LIL_CHANNEL_ERROR }, // LCO_CMY src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYA src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYK src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYKA src
			{ LIL_CHANNEL_ERROR }, // LCO_YCBCR src
			{ LIL_CHANNEL_ERROR }, // LCO_YCBCRA src
		},
		{
			// LCO_ARGB dest
			{ LIL_CHANNEL_ERROR }, // LCO_UNKNOWN src
			{ LIL_CHANNEL_NONE, 0, 0, 0 }, // LCO_SINGLE src
			{ 1, 0, 0, 0 }, // LCO_YA src
			{ LIL_CHANNEL_NONE, 0, 1, 2 }, // LCO_RGB src
			{ LIL_CHANNEL_NONE, 2, 1, 0 }, // LCO_BGR src
			{ 0, 1, 2, 3 }, // LCO_ARGB src
			{ 0, 3, 2, 1 }, // LCO_ABGR src
			{ 3, 0, 1, 2 }, // LCO_RGBA src
			{ 3, 2, 1, 0 }, // LCO_BGRA src
			{ LIL_CHANNEL_ERROR }, // LCO_CMY src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYA src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYK src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYKA src
			{ LIL_CHANNEL_ERROR }, // LCO_YCBCR src
			{ LIL_CHANNEL_ERROR }, // LCO_YCBCRA src
		},
		{
			// LCO_ABGR dest
			{ LIL_CHANNEL_ERROR },
			{ LIL_CHANNEL_NONE, 0, 0, 0 },
			{ 1, 0, 0, 0 },
			{ LIL_CHANNEL_NONE, 2, 1, 0 },
			{ LIL_CHANNEL_NONE, 0, 1, 2 },
			{ 0, 3, 2, 1 },
			{ 0, 1, 2, 3 },
			{ 3, 2, 1, 0 },
			{ 3, 0, 1, 2 },
			{ LIL_CHANNEL_ERROR }, // LCO_CMY src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYA src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYK src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYKA src
			{ LIL_CHANNEL_ERROR }, // LCO_YCBCR src
			{ LIL_CHANNEL_ERROR }, // LCO_YCBCRA src
		},
		{
			// LCO_RGBA dest
			{ LIL_CHANNEL_ERROR },
			{ 0, 0, 0, LIL_CHANNEL_NONE },
			{ 0, 0, 0, 1 },
			{ 0, 1, 2, LIL_CHANNEL_NONE },
			{ 2, 1, 0, LIL_CHANNEL_NONE },
			{ 1, 2, 3, 0 },
			{ 3, 2, 1, 0 },
			{ 0, 1, 2, 3 },
			{ 2, 1, 0, 3 },
			{ LIL_CHANNEL_ERROR }, // LCO_CMY src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYA src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYK src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYKA src
			{ LIL_CHANNEL_ERROR }, // LCO_YCBCR src
			{ LIL_CHANNEL_ERROR }, // LCO_YCBCRA src
		},
		{
			// LCO_BGRA dest
			{ LIL_CHANNEL_ERROR },
			{ 0, 0, 0, LIL_CHANNEL_NONE },
			{ 0, 0, 0, 1 },
			{ 2, 1, 0, LIL_CHANNEL_NONE },
			{ 0, 1, 2, LIL_CHANNEL_NONE },
			{ 3, 2, 1, 0 },
			{ 1, 2, 3, 0 },
			{ 2, 1, 0, 3 },
			{ 0, 1, 2, 3 },
			{ LIL_CHANNEL_ERROR }, // LCO_CMY src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYA src
			{ LIL_CHANNEL_ERROR }, // LCO_CMYK src
			{ LIL_CHANNEL_CONVERT, LIL_CHANNEL_CONVERT, LIL_CHANNEL_CONVERT, 4 }, // LCO_CMYKA src
			{ LIL_CHANNEL_ERROR }, // LCO_YCBCR src
			{ LIL_CHANNEL_ERROR }, // LCO_YCBCRA src
		}
	};

	//-------------------------------------------------------------------------------------------------------
	// Assumes destSrcMap is large enough...
	void BuildDestSrcMap(lil_u8 *destSrcMap, LilChannelOrderType srcOrder, LilChannelOrderType destOrder)
	{
		for (int iCh = 0; iCh < LIL_NUM_CHANNELS[destOrder]; iCh++)
			destSrcMap[iCh] = DestToSrcLut[destOrder][srcOrder][iCh];
	}

	//-------------------------------------------------------------------------------------------------------
	// lutBitOffset has to be the beginning bit of the 1st entry for the current destination plane
	// (Remember at this point the destination format and LUT format are identical)
	void ConvertBigToBigLutPlane(BitIoMsb *pSrc, lil_u8 srcBitCount, lil_u8 srcBitSkip,
		BitIoMsb *pDest, lil_u8 destBitCount, lil_u8 destBitSkip,
		lil_u32 numPixels, const lil_u8 *pLut, lil_u8 lutBitOffset, lil_u8 lutBpp)
	{
		BitIoMsb pLutBits(pLut);

		// Do the conversion
		for (lil_u32 iPix = 0; iPix < numPixels; iPix++) {
			// Read src channel index.  MAY be alpha, not just index.
			lil_u32 srcIndex = pSrc->ReadBits(srcBitCount);
			pSrc->SkipBits(srcBitSkip);

			// Write dest channel (singular)
			// @@ See comment in non-planar version
			pLutBits.SetBitPos(lutBpp * srcIndex + lutBitOffset);
			pDest->WriteBits(destBitCount, pLutBits.ReadBits(destBitCount));
			pDest->SkipBits(destBitSkip);
		}
	}

	//-------------------------------------------------------------------------------------------------------
	void ConvertBigToBigLut(BitIoMsb *pSrc, const lil_u8 *pSrcBitCounts, lil_u8 numSrcCh,
		BitIoMsb *pDest, const lil_u8 *pDestBitCounts, const lil_u8 *pDestSrcChMap, lil_u8 numDestCh,
		lil_u32 numPixels, const lil_u8 *pLut)
	{
		// Fast paths
		if (numSrcCh == 1 && numDestCh == 4 && pDestBitCounts[0] == 8 && pDestBitCounts[1] == 8 &&
			pDestBitCounts[2] == 8 && pDestBitCounts[3] == 8 && pSrcBitCounts[0] == 8)
		{
			if (pDestSrcChMap[1] == 0 && pDestSrcChMap[2] == 0) {
				if (pDestSrcChMap[0] == 0 && pDestSrcChMap[3] == LIL_CHANNEL_NONE) {
					for (lil_u32 iPix = 0; iPix < numPixels; iPix++) {
						lil_u32 i = pSrc->ReadBits(8) * 4;
						pDest->WriteBits(32, (pLut[i] << 24) | (pLut[i + 1] << 16) | (pLut[i + 2] << 8) | 0xff);
					}
					return;
				}
			}
		}

		// General path
		lil_u32 srcIndex[LIL_MAX_CHANNELS];

		lil_u32 palBits[LIL_MAX_CHANNELS + 1];
		palBits[0] = 0;
		for (int i = 1; i <= numDestCh; i++)
			palBits[i] = palBits[i - 1] + pDestBitCounts[i - 1];
		BitIoMsb pLutBits(pLut);

		// Do the conversion
		for (lil_u32 iPix = 0; iPix < numPixels; iPix++) {
			// Read src channel indices.  Could be more than one if alpha or something like that
			// is present
			for (int iCh = 0; iCh < numSrcCh; iCh++)
				srcIndex[iCh] = pSrc->ReadBits(pSrcBitCounts[iCh]);

			// @@ Only works if channel 0 of source is the index.  Need more general method
			//    if we will use AY and YA color order, or if lut can span channels.
			//    IF we do anything about this, it needs to be handled in planar version as well.
			//    So... this shouldn't convert channels that are direct src alpha to dest alpha
			//    in the case of say, YA or something.  But it does currently...
			// Write dest channels
			for (int iCh = 0; iCh < numDestCh; iCh++) {
				// We assume the pLut is already the correct format
				lil_u32 srcChan = pDestSrcChMap[iCh];
				if (srcChan != LIL_CHANNEL_NONE) {
					if (srcChan == 0) {
						pLutBits.SetBitPos(palBits[numDestCh] * srcIndex[srcChan] + palBits[iCh]);
						pDest->WriteBits(pDestBitCounts[iCh], pLutBits.ReadBits(pDestBitCounts[iCh]));
					}
					else
						pDest->WriteBits(pDestBitCounts[iCh], srcIndex[srcChan]);
				}
				else
					pDest->WriteBits(pDestBitCounts[iCh], 0xffffffff);
			}
		}
	}

	//-------------------------------------------------------------------------------------------------------
	void ConvertBigToBigUpPlane(BitIoMsb *pSrc, lil_u8 srcBitCount, lil_u8 srcBitSkip,
		BitIoMsb *pDest, lil_u8 destBitCount, lil_u8 destBitSkip, lil_u32 numPixels, bool scaleBits)
	{
		// A shifter for each destination channel
		lil_u32 mult = 1;
		if (scaleBits)
			mult = ((1 << destBitCount) - 1) / ((1 << srcBitCount) - 1);

		// Do the conversion
		for (lil_u32 iPix = 0; iPix < numPixels; iPix++) {
			// Read src channel
			lil_u32 srcVal = pSrc->ReadBits(srcBitCount) * mult;
			pSrc->SkipBits(srcBitSkip);

			// Write dest channel
			pDest->WriteBits(destBitCount, srcVal);
			pDest->SkipBits(destBitSkip);
		}
	}

	//-------------------------------------------------------------------------------------------------------
	void ConvertBigToBigUp(BitIoMsb *pSrc, const lil_u8 *pSrcBitCounts, lil_u8 numSrcCh,
		BitIoMsb *pDest, const lil_u8 *pDestBitCounts, const lil_u8 *pDestSrcChMap, lil_u8 numDestCh,
		lil_u32 numPixels, bool scaleBits)
	{
		lil_u32 srcVal[LIL_MAX_CHANNELS];

		// A multiplier for each destination channel
		lil_u32 mult[LIL_MAX_CHANNELS];
		for (int iCh = 0; iCh < numDestCh; iCh++) {
			if (pDestSrcChMap[iCh] == LIL_CHANNEL_NONE)
				continue;
			if (scaleBits)
				mult[pDestSrcChMap[iCh]] = ((1 << pDestBitCounts[iCh]) - 1) /
					((1 << pSrcBitCounts[pDestSrcChMap[iCh]]) - 1);
			else
				mult[pDestSrcChMap[iCh]] = 1;
		}
		/*lil_u32 shup[LIL_MAX_CHANNELS], shdown[LIL_MAX_CHANNELS];
		for (int iCh = 0; iCh < numDestCh; iCh++) {
			if (pDestSrcChMap[iCh] == LIL_CHANNEL_NONE)
				continue;
			if (scaleBits) {
				shup[pDestSrcChMap[iCh]] = pDestBitCounts[iCh] - pSrcBitCounts[pDestSrcChMap[iCh]];
				shdown[pDestSrcChMap[iCh]] = pSrcBitCounts[pDestSrcChMap[iCh]] - shup[pDestSrcChMap[iCh]];
			}
			else
				shup[pDestSrcChMap[iCh]] = shdown[pDestSrcChMap[iCh]] = 0;
		}*/

		// Do the conversion
		for (lil_u32 iPix = 0; iPix < numPixels; iPix++) {
			// Read src channels
			for (int iCh = 0; iCh < numSrcCh; iCh++) {
				srcVal[iCh] = pSrc->ReadBits(pSrcBitCounts[iCh]) * mult[iCh];
				//lil_u32 val = pSrc->ReadBits(pSrcBitCounts[iCh]);
				//srcVal[iCh] = (val << shup[iCh]) | (val >> shdown[iCh]);
			}

			// Write dest channels
			for (int iCh = 0; iCh < numDestCh; iCh++) {
				if (pDestSrcChMap[iCh] != LIL_CHANNEL_NONE)
					pDest->WriteBits(pDestBitCounts[iCh], srcVal[pDestSrcChMap[iCh]]);
				else
					pDest->WriteBits(pDestBitCounts[iCh], 0xffffffff);
			}
		}
	}

	//-------------------------------------------------------------------------------------------------------
	void ConvertBigToBigDownPlane(BitIoMsb *pSrc, lil_u8 srcBitCount, lil_u8 srcBitSkip,
		BitIoMsb *pDest, lil_u8 destBitCount, lil_u8 destBitSkip, lil_u32 numPixels, bool scaleBits)
	{
		// A shifter for each destination channel
		lil_u32 shift = 0;
		if (scaleBits)
			shift = srcBitCount - destBitCount;

		// Do the conversion
		for (lil_u32 iPix = 0; iPix < numPixels; iPix++) {
			// Read src channel
			lil_u32 srcVal = pSrc->ReadBits(srcBitCount) >> shift;
			pSrc->SkipBits(srcBitSkip);

			// Write dest channel
			pDest->WriteBits(destBitCount, srcVal);
			pDest->SkipBits(destBitSkip);
		}
	}

	//-------------------------------------------------------------------------------------------------------
	// For non-planar data that only needs to swap byte values
	void ConvertByteSwap(const lil_u8 *pSrc, lil_u8 numSrcCh, lil_u8 *pDest, const lil_u8 *pDestSrcChMap,
		lil_u8 numDestCh, lil_u32 numPixels)
	{
		lil_u32 destSrcFast = 0;  // will contain destsrchmap compacted to 00112233 in bits
		for (lil_u32 i = 0; i < numDestCh; i++)
			destSrcFast |= (pDestSrcChMap[i] << (24 - 8 * i));
		if (numSrcCh == 1 && numDestCh == 3 && destSrcFast == 0)
			for (lil_u32 iPix = 0; iPix < numPixels; iPix++) {
				*pDest++ = *pSrc;
				*pDest++ = *pSrc;
				*pDest++ = *pSrc;
				pSrc++;
			}
		else if (numSrcCh == 1 && numDestCh == 4 && destSrcFast == LIL_CHANNEL_NONE)
			for (lil_u32 iPix = 0; iPix < numPixels; iPix++) {
				*pDest++ = *pSrc;
				*pDest++ = *pSrc;
				*pDest++ = *pSrc;
				*pDest++ = 0xff;
				pSrc++;
			}
		else if (numSrcCh == 1 && numDestCh == 4 && destSrcFast == LIL_CHANNEL_NONE << 24)
			for (lil_u32 iPix = 0; iPix < numPixels; iPix++) {
				*pDest++ = 0xff;
				*pDest++ = *pSrc;
				*pDest++ = *pSrc;
				*pDest++ = *pSrc;
				pSrc++;
			}
		else if (numSrcCh == 3 && numDestCh == 3 && destSrcFast == 0x02010000)
			for (lil_u32 iPix = 0; iPix < numPixels; iPix++) {
				*pDest++ = pSrc[2];
				*pDest++ = pSrc[1];
				*pDest++ = pSrc[0];
				pSrc += 3;
			}
		else if (numSrcCh == 3 && numDestCh == 4 && destSrcFast == (0x02010000 | LIL_CHANNEL_NONE))
			for (lil_u32 iPix = 0; iPix < numPixels; iPix++) {
				*pDest++ = pSrc[2];
				*pDest++ = pSrc[1];
				*pDest++ = pSrc[0];
				*pDest++ = 0xff;
				pSrc += 3;
			}
		else if (numSrcCh == 3 && numDestCh == 4 && destSrcFast == (0x00010200 | LIL_CHANNEL_NONE))
			for (lil_u32 iPix = 0; iPix < numPixels; iPix++) {
				*pDest++ = *pSrc++;
				*pDest++ = *pSrc++;
				*pDest++ = *pSrc++;
				*pDest++ = 0xff;
			}
		else if (numSrcCh == 3 && numDestCh == 4 && destSrcFast == (0x00020100 | (LIL_CHANNEL_NONE << 24)))
			for (lil_u32 iPix = 0; iPix < numPixels; iPix++) {
				*pDest++ = 0xff;
				*pDest++ = pSrc[2];
				*pDest++ = pSrc[1];
				*pDest++ = pSrc[0];
				pSrc += 3;
			}
		else if (numSrcCh == 3 && numDestCh == 4 && destSrcFast == (0x00000102 | (LIL_CHANNEL_NONE << 24)))
			for (lil_u32 iPix = 0; iPix < numPixels; iPix++) {
				*pDest++ = 0xff;
				*pDest++ = *pSrc++;
				*pDest++ = *pSrc++;
				*pDest++ = *pSrc++;
			}
		else if (numSrcCh == 4 && numDestCh == 4 && (destSrcFast & 0xfcfcfcfc) == 0)
			for (lil_u32 iPix = 0; iPix < numPixels; iPix++) {
				*pDest++ = pSrc[pDestSrcChMap[0]];
				*pDest++ = pSrc[pDestSrcChMap[1]];
				*pDest++ = pSrc[pDestSrcChMap[2]];
				*pDest++ = pSrc[pDestSrcChMap[3]];
				pSrc += 4;
			}
		else  // Generic version
			for (lil_u32 iPix = 0; iPix < numPixels; iPix++) {
				// Write dest channels
				for (int iCh = 0; iCh < numDestCh; iCh++) {
					if (pDestSrcChMap[iCh] != LIL_CHANNEL_NONE)
						*pDest++ = pSrc[pDestSrcChMap[iCh]];
					else
						*pDest++ = 0xff;
				}
				pSrc += numSrcCh;
			}
	}

	//-------------------------------------------------------------------------------------------------------
	// The num***Ch vars are a channel count when non-planar, and the channel number itself when planar.
	void ConvertBigToBigDown(BitIoMsb *pSrc, const lil_u8 *pSrcBitCounts, lil_u8 numSrcCh,
		BitIoMsb *pDest, const lil_u8 *pDestBitCounts, const lil_u8 *pDestSrcChMap, lil_u8 numDestCh,
		lil_u32 numPixels, bool scaleBits)
	{
		lil_u32 srcVal[LIL_MAX_CHANNELS];

		// A shifter for each destination channel
		lil_u32 shift[LIL_MAX_CHANNELS];
		for (int iCh = 0; iCh < numDestCh; iCh++) {
			if (pDestSrcChMap[iCh] == LIL_CHANNEL_NONE)
				continue;
			if (scaleBits)
				shift[pDestSrcChMap[iCh]] = pSrcBitCounts[pDestSrcChMap[iCh]] - pDestBitCounts[iCh];
			else
				shift[pDestSrcChMap[iCh]] = 0;
		}

		// Do the conversion
		for (lil_u32 iPix = 0; iPix < numPixels; iPix++) {
			// Read src channels
			for (int iCh = 0; iCh < numSrcCh; iCh++)
				srcVal[iCh] = pSrc->ReadBits(pSrcBitCounts[iCh]) >> shift[iCh];

			// Write dest channels
			for (int iCh = 0; iCh < numDestCh; iCh++) {
				if (pDestSrcChMap[iCh] != LIL_CHANNEL_NONE)
					pDest->WriteBits(pDestBitCounts[iCh], srcVal[pDestSrcChMap[iCh]]);
				else
					pDest->WriteBits(pDestBitCounts[iCh], 0xffffffff);
			}
		}
	}

	//-------------------------------------------------------------------------------------------------------
	// Some channels may have to convert up, while others may have to convert down or not at all.
	// Seems unlikely, but it's there just in case - like 1:5:5:5 to 4:4:4:4 ARGB.
	void ConvertBigToBigMixed(BitIoMsb *pSrc, const lil_u8 *pSrcBitCounts, lil_u8 numSrcCh,
		BitIoMsb *pDest, const lil_u8 *pDestBitCounts, const lil_u8 *pDestSrcChMap, lil_u8 numDestCh,
		lil_u32 numPixels, bool scaleBits)
	{
		lil_u32 srcVal[LIL_MAX_CHANNELS];

		// A shifter for each destination channel
		lil_u32 shift[LIL_MAX_CHANNELS];
		lil_u32 mult[LIL_MAX_CHANNELS];
		bool doMult[LIL_MAX_CHANNELS];
		for (int iCh = 0; iCh < numDestCh; iCh++) {
			if (pDestSrcChMap[iCh] == LIL_CHANNEL_NONE)
				continue;
			if (pDestBitCounts[iCh] <= pSrcBitCounts[pDestSrcChMap[iCh]]) {
				if (scaleBits)
					shift[pDestSrcChMap[iCh]] = pSrcBitCounts[pDestSrcChMap[iCh]] - pDestBitCounts[iCh];
				else
					shift[pDestSrcChMap[iCh]] = 0;
				doMult[pDestSrcChMap[iCh]] = false;
			}
			else {
				if (scaleBits)
					mult[pDestSrcChMap[iCh]] = ((1 << pDestBitCounts[iCh]) - 1) /
						((1 << pSrcBitCounts[pDestSrcChMap[iCh]]) - 1);
				else
					mult[pDestSrcChMap[iCh]] = 1;
				doMult[pDestSrcChMap[iCh]] = true;
			}
		}

		// Do the conversion
		for (lil_u32 iPix = 0; iPix < numPixels; iPix++) {
			// Read src channels
			for (int iCh = 0; iCh < numSrcCh; iCh++) {
				srcVal[iCh] = pSrc->ReadBits(pSrcBitCounts[iCh]);
				if (doMult[iCh])
					srcVal[iCh] *= mult[iCh];
				else
					srcVal[iCh] >>= shift[iCh];
			}

			// Write dest channels
			for (int iCh = 0; iCh < numDestCh; iCh++) {
				if (pDestSrcChMap[iCh] != LIL_CHANNEL_NONE)
					pDest->WriteBits(pDestBitCounts[iCh], srcVal[pDestSrcChMap[iCh]]);
				else
					pDest->WriteBits(pDestBitCounts[iCh], 0xffffffff);
			}
		}
	}
} // namespace



//-----------------------------------------------------------------------------------------------------------
LilErrorType FillChannel(const void *pSrcData, void *pDestData, const LilFormatDesc *pDestFormat,
	 lil_u32 destCh, lil_u32 width, lil_u32 height)
{
	LilErrorType err = LE_NONE;
	if (!IsFormatValid(pDestFormat, &err))
		return err;

	// Compensate for planar conversion
	if (destCh == LIL_CHANNEL_NONE)
		return LE_NONE;
	if (destCh == LIL_CHANNEL_ERROR)
		return LE_INTERNALERROR;

	// Calc dest manip offsets
	lil_u32 destBpl = CalcBitsPerLine(pDestFormat, width);
	lil_u32 destOffset = 0;
	lil_u32 destBitSkip = 0;
	lil_u8 destBits = pDestFormat->bitsPerChannel;
	if (pDestFormat->planar)
		destOffset = destBpl * height * destCh;
	else {
		switch (pDestFormat->dataFormat) {
		case LDF_SIMPLE:
			destOffset = pDestFormat->bitsPerChannel * destCh;
			destBitSkip = pDestFormat->bitsPerChannel * (pDestFormat->numChannels - 1);
			break;
		case LDF_SIMPLEBITS:
			for (lil_u32 i = 0; i < destCh; i++)
				destOffset += pDestFormat->bitCounts[i];
			for (lil_u8 i = 0; i < pDestFormat->numChannels; i++) {
				if (i != destCh)
					destBitSkip += pDestFormat->bitCounts[i];
			}
			destBits = pDestFormat->bitCounts[destCh];
			break;
		default:
			return LE_UNSUPPORTEDCONVERSION;
		}
	}

	BitIoMsb srcData(pSrcData);
	lil_u32 srcVal = srcData.ReadBits(destBits);

	BitIoMsb destData(pDestData);

	for (lil_u32 iY = 0; iY < height; iY ++) {
		destData.SetBitPos(iY * destBpl + destOffset);

		for (lil_u32 iPix = 0; iPix < width; iPix++) {
			// Write dest channel
			destData.WriteBits(destBits, srcVal);
			destData.SkipBits(destBitSkip);
		}
	}

	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType FillAlpha(void *pDestData, LilImageDesc *pImage) {
	// Some formats use transparency correctly, while some SAY they have an alpha channel
	// (looking at some 32 bit .bmp images here) but it is treated as undefined, and some SAY
	// they have an undefined channel (even some 32 bit images) and treat it as truly alpha.
	// The best general solution is to not use alpha unfortunately...
	// I'd love to be wrong and have missed something and can fix this later!
	if (LIL_ALPHA_CHANNEL[pImage->format.channelOrder] != LIL_CHANNEL_NONE) {
		lil_u32 opaque = 0xffffffff;

		return FillChannel(&opaque, pDestData, &pImage->format, LIL_ALPHA_CHANNEL[pImage->format.channelOrder],
			pImage->data.width, pImage->data.height);
	}

	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType ConvertChannelToChannel(const void *pSrcData, const LilFormatDesc *pSrcFormat, lil_u32 srcCh,
	void *pDestData, const LilFormatDesc *pDestFormat, lil_u32 destCh, lil_u32 width, lil_u32 height)
{
	LilErrorType err = LE_NONE;
	if (!IsFormatValid(pSrcFormat, &err) || !IsFormatValid(pDestFormat, &err))
		return err;

	// Compensate for planar conversion
	if (srcCh == LIL_CHANNEL_NONE || destCh == LIL_CHANNEL_NONE)
		return LE_NONE;
	if (srcCh == LIL_CHANNEL_ERROR || destCh == LIL_CHANNEL_ERROR)
		return LE_INTERNALERROR;

	// Calc src manip offsets
	lil_u32 srcBpl = CalcBitsPerLine(pSrcFormat, width);
	lil_u64 srcOffset = 0;
	lil_u8 srcBitSkip = 0;
	lil_u8 srcBits = pSrcFormat->bitsPerChannel;
	if (pSrcFormat->planar)
		srcOffset = static_cast<lil_u64>(srcBpl) * height * srcCh;
	else {
		switch (pSrcFormat->dataFormat) {
		case LDF_SIMPLE:
			srcOffset = pSrcFormat->bitsPerChannel * srcCh;
			srcBitSkip = pSrcFormat->bitsPerChannel * (pSrcFormat->numChannels - 1);
			break;
		case LDF_SIMPLEBITS:
			for (lil_u32 i = 0; i < srcCh; i++)
				srcOffset += pSrcFormat->bitCounts[i];
			for (lil_u8 i = 0; i < pSrcFormat->numChannels; i++)
				if (i != srcCh)
					srcBitSkip += pSrcFormat->bitCounts[i];
			srcBits = pSrcFormat->bitCounts[srcCh];
			break;
		default:
			return LE_UNSUPPORTEDCONVERSION;
		}
	}

	// Calc dest manip offsets
	lil_u32 destBpl = CalcBitsPerLine(pDestFormat, width);
	lil_u64 destOffset = 0;
	lil_u8 destBitSkip = 0;
	lil_u8 destBits = pDestFormat->bitsPerChannel;
	if (pDestFormat->planar)
		destOffset = static_cast<lil_u64>(destBpl) * height * destCh;
	else {
		switch (pDestFormat->dataFormat) {
		case LDF_SIMPLE:
			destOffset = pDestFormat->bitsPerChannel * destCh;
			destBitSkip = pDestFormat->bitsPerChannel * (pDestFormat->numChannels - 1);
			break;
		case LDF_SIMPLEBITS:
			for (lil_u32 i = 0; i < destCh; i++)
				destOffset += pDestFormat->bitCounts[i];
			for (lil_u8 i = 0; i < pDestFormat->numChannels; i++)
				if (i != destCh)
					destBitSkip += pDestFormat->bitCounts[i];
			destBits = pDestFormat->bitCounts[destCh];
			break;
		default:
			return LE_UNSUPPORTEDCONVERSION;
		}
	}

	// Image iteration
	// See comment on LCB_LINEAR_SCALE in Lil.h
	// Maybe we should force to true???
	bool scaleBits = (globalConfig.convertBits == LCB_LINEAR_SCALE);

	BitIoMsb srcData(pSrcData);
	BitIoMsb destData(pDestData);
	// Check for flip during conversion
	lil_u32 yStart = 0, yEnd = height, yDir = 1;
	if (pSrcFormat->flipped != pDestFormat->flipped) {
		yEnd = yStart - 1;
		yStart = height - 1;
#pragma warning(disable: 4146)  // negating an unsigned integer gives unsigned result
		yDir = -yDir;
#pragma warning(default: 4146)
	}
	for (lil_u32 iY = 0, iY2 = yStart; iY < height; iY ++, iY2 += yDir) {
		srcData.SetBitPos(iY * srcBpl + srcOffset);
		destData.SetBitPos(iY2 * destBpl + destOffset);

		if (destBits > srcBits)
			ConvertBigToBigUpPlane(&srcData, srcBits, srcBitSkip,
				&destData, destBits, destBitSkip,
				width, scaleBits);
		else
			ConvertBigToBigDownPlane(&srcData, srcBits, srcBitSkip,
				&destData, destBits, destBitSkip,
				width, scaleBits);
	}

	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
void MakeBigEndian(void *pData, lil_u32 width, lil_u32 height, const LilFormatDesc *format) {
	lil_u32 inBpp = CalcBpp(format);
	lil_u32 inLineBytes = CalcBpl(format, width);
	// Only matters if not already in big endian format.
	if (format->dataFormat == LDF_SIMPLEBITS) {
		if (inBpp == 16) {
			LIL_ASSERT(inLineBytes / 2 * 2 == inLineBytes);
			ByteSwapBuf16(pData, height * inLineBytes / 2);
		}
		if (inBpp == 32) {
			LIL_ASSERT(inLineBytes / 4 * 4 == inLineBytes);
			ByteSwapBuf32(pData, height * inLineBytes / 4);
		}
	}
	if (format->dataFormat == LDF_SIMPLE) {
		lil_u32 Bpc = inBpp / format->numChannels;
		lil_u32 wMult = format->numChannels;
		lil_u32 hMult = 1;
		if (format->planar)
			Swap(wMult, hMult);
		if (Bpc == 16) {
			LIL_ASSERT(inLineBytes / 2 * 2 == inLineBytes);
			ByteSwapBuf16(pData, height * hMult * inLineBytes / 2);
		}
		else if (Bpc == 32) {
			LIL_ASSERT(inLineBytes / 4 * 4 == inLineBytes);
			ByteSwapBuf32(pData, height * hMult * inLineBytes / 4);
		}
		else if (!format->planar && format->bitsPerChannel & 7) {
			if (inBpp == 16) {
				LIL_ASSERT(inLineBytes / 2 * 2 == inLineBytes);
				ByteSwapBuf16(pData, height * hMult * inLineBytes / 2);
			}
			else if (inBpp == 32) {
				LIL_ASSERT(inLineBytes / 4 * 4 == inLineBytes);
				ByteSwapBuf32(pData, height * hMult * inLineBytes / 4);
			}
		}
	}
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType ConvertData(const LilMemDesc *pMemFuncs,
	const void *pSrcData, const LilFormatDesc *pSrcFormat, const LilPaletteDesc *pSrcPalette,
	void *pDestData, const LilFormatDesc *pDestFormat, LilPaletteDesc *pDestPalette,
	lil_u32 width, lil_u32 height, bool isLittleEndian)
{
	LilErrorType err = LE_NONE;
	if (!IsFormatValid(pSrcFormat, &err) || !IsFormatValid(pDestFormat, &err))
		return err;

	// --- CONVERT PALETTE FIRST ! ---

	// Palette check
	// ?? Should be able to apply lookup
	// tables for RGB images... triple level palettes essentially.
	// ... maybe ...
	bool hasPalette =
		pSrcPalette && pSrcPalette->numEntries && (globalConfig.convertPalette == LCP_USE_PALETTE) &&
		(pSrcFormat->channelOrder == LCO_SINGLE || pSrcFormat->channelOrder == LCO_YA);
	if (hasPalette && !IsFormatValid(&pSrcPalette->format, 0))
		return LE_INVALIDPALETTEFORMAT;
	if (hasPalette && pDestFormat->channelOrder == LCO_SINGLE &&
		pDestFormat->bitsPerChannel < pSrcFormat->bitsPerChannel)
		return LE_NOPALETTESHRINKING;

	// If we match the palette format to the dest image format, it will be a much
	// better LUT huh?
	// @@ What should we do with the palette once we're done with it though?  Free it...
	void *pLut = 0;
	lil_mem_handle tempPal = 0;

	if ((pSrcFormat->numChannels != pDestFormat->numChannels ||
		pSrcFormat->channelOrder != pDestFormat->channelOrder ||
		pSrcFormat->dataFormat != pDestFormat->dataFormat) &&
		hasPalette)
	{
		lil_u32 paletteBytes = CalcBpl(&pSrcPalette->format, pSrcPalette->numEntries);
		pLut = pSrcPalette->lut;
		if (IsFormatDifferent(&pSrcPalette->format, pDestFormat))
		{
			lil_u32 newPaletteBytes = CalcBpl(pDestFormat, pSrcPalette->numEntries);
			lil_mem_handle memNewLut = 0;
			err = pMemFuncs->memReallocFunc(&memNewLut, newPaletteBytes);
			if (err != LE_NONE)
				return err;

			if (isLittleEndian)
				MakeBigEndian(pSrcPalette->lut, pSrcPalette->numEntries, 1, &pSrcPalette->format);
			ConvertData(pMemFuncs, pLut, &pSrcPalette->format, 0, memNewLut, pDestFormat, 0, pSrcPalette->numEntries, 1, isLittleEndian);
#ifdef LIL_ENDIAN_LITTLE
			MakeBigEndian(pSrcPalette->lut, pSrcPalette->numEntries, 1, &pSrcPalette->format);
#endif

			// Free src palette.....
			if (pDestPalette) {
				pMemFuncs->memReallocFunc(&pDestPalette->lut, 0);
				pDestPalette->lut = memNewLut;
                pDestPalette->format = *pDestFormat;
                pDestPalette->numEntries = pSrcPalette->numEntries;
			}
			else
				tempPal = memNewLut;
			pLut = memNewLut;
		}
	}


	// --- CONVERT IMAGE DATA COLORSPACE ---

	LilFormatDesc colorSpaceSrc;
	LilBuffer colorSpaceBuf;
	if (IsColorSpaceDifferent(pSrcFormat->channelOrder, pDestFormat->channelOrder)) {
		BitIoMsb srcIo[LIL_MAX_CHANNELS];
		BitIoMsb destIo[LIL_MAX_CHANNELS];

		colorSpaceSrc = *pSrcFormat;
		// @@ See note in todo.txt on this planar/chunky difference...
		if (pSrcFormat->planar)
			colorSpaceSrc.channelOrder = GetDefaultColorspace(pDestFormat->channelOrder);
		else
			colorSpaceSrc.channelOrder = pDestFormat->channelOrder;
		colorSpaceSrc.numChannels = LIL_NUM_CHANNELS[colorSpaceSrc.channelOrder];
		if (!IsFormatValid(&colorSpaceSrc, &err))
			return err;

		// @@ See note in todo.txt on this planar/chunky difference...
		ConvertFunc func;
		if (pSrcFormat->planar)
			func = CONVERT_FUNCTION[GetDefaultColorspace(pSrcFormat->channelOrder)][colorSpaceSrc.channelOrder];
		else
			func = CONVERT_FUNCTION[pSrcFormat->channelOrder][colorSpaceSrc.channelOrder];
		if (!func)
			return LE_UNSUPPORTEDCONVERSION;

		// There are a few cases that are naturally handled by conversion below (such as LCO_SINGLE to LCO_RGB)
		if (func != ConvertVoid) {
			if (!colorSpaceBuf.Realloc(pMemFuncs, CalcByteSize(&colorSpaceSrc, width, height)))
				return LE_NOTENOUGHMEMORY;
			if (pSrcFormat->planar)
				for (lil_u32 i = 0; i < pSrcFormat->numChannels; i++) {
					const lil_u8 *pSrc = reinterpret_cast<const lil_u8 *>(pSrcData);
					srcIo[i].Reset(pSrc + CalcBpl(pSrcFormat, width) * height * i);
					lil_u8 *pDest = colorSpaceBuf.GetPtr<lil_u8>();
					destIo[i].Reset(pDest + CalcBpl(&colorSpaceSrc, width) * height * i);
				}
			else {
				srcIo[0].Reset(pSrcData);
				destIo[0].Reset(colorSpaceBuf.GetPtr<lil_u8>());
			}

			LIL_ASSERT(pSrcFormat->dataFormat == LDF_SIMPLE);
			lil_u32 oldBpl = CalcBitsPerLine(pSrcFormat, width);
			lil_u32 newBpl = CalcBitsPerLine(&colorSpaceSrc, width);
			for (lil_u32 iY = 0; iY < height; iY ++) {
				for (lil_u32 i = 0; i < pSrcFormat->numChannels; i++)
					srcIo[i].SetBitPos(iY * oldBpl);
				for (lil_u32 i = 0; i < colorSpaceSrc.numChannels; i++)
					destIo[i].SetBitPos(iY * newBpl);
				func(srcIo, destIo, pSrcFormat->bitsPerChannel, HasAlpha(pSrcFormat), HasAlpha(pDestFormat),
					width, pSrcFormat->planar);
			}

			pSrcFormat = &colorSpaceSrc;
			pSrcData = colorSpaceBuf.GetPtr<const void>();
		}
	}


	// --- SET UP CONVERSION VARS ---

	BitIoMsb srcData(pSrcData);
	BitIoMsb destData(pDestData);

	// Build masks and channel transfer
	lil_u8 srcBits[LIL_MAX_CHANNELS];
	lil_u8 destBits[LIL_MAX_CHANNELS];
	lil_u8 destSrcMap[LIL_MAX_CHANNELS];
	if (pSrcFormat->dataFormat == LDF_SIMPLEBITS)
		for (int iCh = 0; iCh < pSrcFormat->numChannels; iCh++)
			srcBits[iCh] = pSrcFormat->bitCounts[iCh];
	else
		for (int iCh = 0; iCh < pSrcFormat->numChannels; iCh++)
			srcBits[iCh] = pSrcFormat->bitsPerChannel;
	if (pDestFormat->dataFormat == LDF_SIMPLEBITS)
		for (int iCh = 0; iCh < pDestFormat->numChannels; iCh++)
			destBits[iCh] = pDestFormat->bitCounts[iCh];
	else
		for (int iCh = 0; iCh < pDestFormat->numChannels; iCh++)
			destBits[iCh] = pDestFormat->bitsPerChannel;
	BuildDestSrcMap(destSrcMap, pSrcFormat->channelOrder, pDestFormat->channelOrder);

	// See comment on LCB_LINEAR_SCALE in Lil.h
	bool scaleBits = (globalConfig.convertBits == LCB_LINEAR_SCALE) && (!hasPalette);

	// Image iteration
	lil_u64 srcBitsPerLine = CalcBitsPerLine(pSrcFormat, width);
	lil_u64 srcBytesPerLine = CalcBpl(pSrcFormat, width);
	lil_u64 destBitsPerLine = CalcBitsPerLine(pDestFormat, width);
	lil_u64 destBytesPerLine = CalcBpl(pDestFormat, width);

	// Check for flip during conversion
	lil_u32 yStart = 0, yEnd = height, yDir = 1;
	if (pSrcFormat->flipped != pDestFormat->flipped) {
		yEnd = yStart - 1;
		yStart = height - 1;
#pragma warning(disable: 4146)  // negating an unsigned integer gives unsigned result
		// wouldn't this be yDir = ~(yDir - 1);
		yDir = -yDir;
#pragma warning(default: 4146)
	}


	// --- CONVERT IMAGE DATA BITS ---

	// Conversion of all channels are same bit depth
	if (pSrcFormat->dataFormat == LDF_SIMPLE &&
		pDestFormat->dataFormat == LDF_SIMPLE)
	{
		// Do Planar conversion
		if (pSrcFormat->planar || pDestFormat->planar) {
			for (lil_u8 iDestCh = 0; iDestCh < pDestFormat->numChannels; iDestCh++) {
				// Compensate for planar conversion
				lil_u8 iSrcCh = destSrcMap[iDestCh];
				if (iSrcCh == LIL_CHANNEL_NONE)
					continue;
				if (iSrcCh == LIL_CHANNEL_ERROR || iSrcCh == LIL_CHANNEL_CONVERT)
					return LE_INTERNALERROR;
				lil_u64 srcOffset = static_cast<lil_u64>(pSrcFormat->bitsPerChannel) * iSrcCh;
				lil_u8 srcBitSkip = pSrcFormat->bitsPerChannel * (pSrcFormat->numChannels - 1);
				if (pSrcFormat->planar) {
					srcOffset = static_cast<lil_u64>(srcBitsPerLine) * height * iSrcCh;
					srcBitSkip = 0;
				}
				lil_u64 destOffset = static_cast<lil_u64>(pDestFormat->bitsPerChannel) * iDestCh;
				lil_u8 destBitSkip = pDestFormat->bitsPerChannel * (pDestFormat->numChannels - 1);
				if (pDestFormat->planar) {
					destOffset = static_cast<lil_u64>(destBitsPerLine) * height * iDestCh;
					destBitSkip = 0;
				}
				for (lil_u32 iY = 0, iY2 = yStart; iY < height; iY ++, iY2 += yDir) {
					srcData.SetBitPos(iY * srcBitsPerLine + srcOffset);
					destData.SetBitPos(iY2 * destBitsPerLine + destOffset);

					if (pLut)
						ConvertBigToBigLutPlane(&srcData, srcBits[iSrcCh], srcBitSkip,
							&destData, destBits[iDestCh], destBitSkip,
							width, reinterpret_cast<lil_u8 *>(pLut),
							pDestFormat->bitsPerChannel * iDestCh,
							pDestFormat->bitsPerChannel * pDestFormat->numChannels);
					else if (pDestFormat->bitsPerChannel > pSrcFormat->bitsPerChannel)
						ConvertBigToBigUpPlane(&srcData, srcBits[iSrcCh], srcBitSkip,
							&destData, destBits[iDestCh], destBitSkip,
							width, scaleBits);
					else
						ConvertBigToBigDownPlane(&srcData, srcBits[iSrcCh], srcBitSkip,
							&destData, destBits[iDestCh], destBitSkip,
							width, scaleBits);
				}
			}
		}
		// Do non-planar conversion
		else {
			for (lil_u32 iY = 0, iY2 = yStart; iY < height; iY ++, iY2 += yDir) {
				srcData.SetBitPos(iY * srcBitsPerLine);
				destData.SetBitPos(iY2 * destBitsPerLine);

				if (pLut)
					ConvertBigToBigLut(&srcData, srcBits, pSrcFormat->numChannels,
						&destData, destBits, destSrcMap, pDestFormat->numChannels,
						width, reinterpret_cast<lil_u8 *>(pLut));
				else if (pDestFormat->bitsPerChannel == 8 && pSrcFormat->bitsPerChannel == 8)
					ConvertByteSwap(reinterpret_cast<const lil_u8 *>(pSrcData) + iY * srcBytesPerLine,
						pSrcFormat->numChannels, reinterpret_cast<lil_u8 *>(pDestData) + iY2 * destBytesPerLine,
						destSrcMap, pDestFormat->numChannels, width);
				else if (pDestFormat->bitsPerChannel > pSrcFormat->bitsPerChannel)
					ConvertBigToBigUp(&srcData, srcBits, pSrcFormat->numChannels,
						&destData, destBits, destSrcMap, pDestFormat->numChannels,
						width, scaleBits);
				else
					ConvertBigToBigDown(&srcData, srcBits, pSrcFormat->numChannels,
						&destData, destBits, destSrcMap, pDestFormat->numChannels,
						width, scaleBits);
			}
		}
	}
	// Conversion if channels in one or both of the images aren't all the same bit depth,
	// but are still integral.
	else if ((pSrcFormat->dataFormat == LDF_SIMPLEBITS &&
		pDestFormat->dataFormat == LDF_SIMPLEBITS) ||
		(pSrcFormat->dataFormat == LDF_SIMPLE &&
		pDestFormat->dataFormat == LDF_SIMPLEBITS) ||
		(pSrcFormat->dataFormat == LDF_SIMPLEBITS &&
		pDestFormat->dataFormat == LDF_SIMPLE))
	{
		for (lil_u32 iY = 0, iY2 = yStart; iY < height; iY ++, iY2 += yDir) {
			srcData.SetBitPos(iY * srcBitsPerLine);
			destData.SetBitPos(iY2 * destBitsPerLine);

			ConvertBigToBigMixed(&srcData, srcBits, pSrcFormat->numChannels,
				&destData, destBits, destSrcMap, pDestFormat->numChannels,
				width, scaleBits);
		}
	}
	else
		return LE_UNSUPPORTEDCONVERSION;


	// --- HANDLE ANY UNHANDLED ALPHA CHANNEL STUFF ---
	// If planar, the dest alpha channel won't be filled with anything yet...
	if ((pSrcFormat->planar || pDestFormat->planar) && !hasPalette) {
		lil_u8 alphaSrc = LIL_ALPHA_CHANNEL[pSrcFormat->channelOrder];
		lil_u8 alphaDest = LIL_ALPHA_CHANNEL[pDestFormat->channelOrder];
		if (alphaSrc == LIL_CHANNEL_NONE && alphaDest != LIL_CHANNEL_NONE) {
			lil_u64 opaque = 0xffffffffffffffffULL;
			FillChannel(&opaque, pDestData, pDestFormat, alphaDest, width, height);
		}
	}
	// If source has both lut indexing and separate alpha, destination alpha will possibly
	// be messed up.  Unfortunately this is an extra pass over the data...
	if (pLut && LIL_ALPHA_CHANNEL[pSrcFormat->channelOrder] != LIL_CHANNEL_NONE
		&& LIL_ALPHA_CHANNEL[pDestFormat->channelOrder] != LIL_CHANNEL_NONE)
	{
		ConvertChannelToChannel(pSrcData, pSrcFormat, LIL_ALPHA_CHANNEL[pSrcFormat->channelOrder],
			pDestData, pDestFormat, LIL_ALPHA_CHANNEL[pDestFormat->channelOrder], width, height);
	}

	if (tempPal) {
		LIL_ASSERT(!pDestPalette);
		pMemFuncs->memReallocFunc(&tempPal, 0);
	}
	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType ConvertPalette(LIL_CONST LilImageDesc *pImage, LilPaletteDesc *pDest, LilBuffer *pTempBuf) {
	LilErrorType err = LE_NONE;

    pDest->numEntries = pImage->palette.numEntries;
    lil_u32 inSize = CalcByteSize(&pImage->palette.format, pImage->palette.numEntries, 1);
    lil_u32 outSize = CalcByteSize(&pDest->format, pDest->numEntries, 1);
    if (pTempBuf) {
        if (!pTempBuf->Realloc(&pImage->memFuncs, outSize))
			return LE_NOTENOUGHMEMORY;
        pDest->lut = pTempBuf->GetPtr<void>();
    }
    else {
        err = pImage->memFuncs.memReallocFunc(&pDest->lut, outSize);
		if (err != LE_NONE)
			return err;
	}

#ifdef LIL_ENDIAN_LITTLE
	bool littleEndian = true;
#else
	bool littleEndian = false;
#endif

	err = ConvertData(&pImage->memFuncs, pImage->palette.lut, &pImage->palette.format, 0,
		pDest->lut, &pDest->format, 0, pDest->numEntries, 1, littleEndian);

    return err;
}

}  // namespace lil
