#include "LilPlatform.h"

#ifdef LIL_SUPPORTS_BMP

#include <memory>
#include "LilCodecRle.h"
#include "LilUtilFormat.h"
#include "LilMem.h"
#include "LilConvert.h"
#include "LilBitManip.h"
#include "LilIo.h"

namespace lil {

//-----------------------------------------------------------------------------------------------------------
namespace {

// IDs
const lil_u16 ID_WIN_BMP = 0x4d42;  // BM
const lil_u16 ID_OS2_BMP = 0x4142;  // BA

// Headers
const lil_u32 HEADER_SIZE_V1 = 12;
const lil_u32 HEADER_SIZE_V2 = 64;
const lil_u32 HEADER_SIZE_V3 = 40;
const lil_u32 HEADER_SIZE_V4 = 108;
const lil_u32 HEADER_SIZE_V5 = 124;

const lil_i32 GOOD_H[] = { HEADER_SIZE_V1, HEADER_SIZE_V3 };
const lil_u8 N_GOOD_H = LIL_ARRAY_SIZE(GOOD_H);
const lil_i32 BAD_H[] = { HEADER_SIZE_V2, HEADER_SIZE_V4, HEADER_SIZE_V5 };
const lil_u8 N_BAD_H = LIL_ARRAY_SIZE(BAD_H);

// Compressions
const lil_i32 GOOD_C[] = { LC_BMP_NONE, LC_BMP_RLE8, LC_BMP_RLE4, LC_BMP_BITFIELDS };
const lil_u8 N_GOOD_C = LIL_ARRAY_SIZE(GOOD_C);
const lil_i32 BAD_C[] = { LC_BMP_JPEG, LC_BMP_PNG };
const lil_u8 N_BAD_C = LIL_ARRAY_SIZE(BAD_C);

// Valid formats
const LilFormatRange ranges[] = {
	{ LDF_SIMPLE, { {1}, {4}, {8} }, 3, 1, 4, FLAG_NO, LCO_SINGLE, FLAG_YESNO },
	{ LDF_SIMPLE, { {8} }, 1, 3, 4, FLAG_NO, LCO_BGR, FLAG_YESNO },
	{ LDF_SIMPLE, { {8} }, 1, 4, 4, FLAG_NO, LCO_BGRA, FLAG_YESNO },
	{ LDF_SIMPLEBITS, { {1,5,5,5} }, 1, 4, 4, FLAG_NO, LCO_ARGB, FLAG_YESNO },
};
const lil_u8 N_RANGES = LIL_ARRAY_SIZE(ranges);
const LilSimpleBitsRange ranges2[] = {
	{ 16, 4, 4, FLAG_NO, LCO_ARGB, FLAG_YESNO },
	{ 16, 3, 4, FLAG_NO, LCO_RGB, FLAG_YESNO },
	{ 32, 4, 4, FLAG_NO, LCO_ARGB, FLAG_YESNO },
	{ 32, 3, 4, FLAG_NO, LCO_ARGB, FLAG_YESNO },
};
const lil_u8 N_RANGES2 = LIL_ARRAY_SIZE(ranges2);

//LilFormatDesc PALETTE_FORMAT_V1 = { LDF_SIMPLE, 8, {}, 3, 4, false, LCO_BGR, false };
LilFormatDesc PALETTE_FORMAT_V3 = { LDF_SIMPLE, 8, {}, 4, 4, false, LCO_BGRA, false };

//-----------------------------------------------------------------------------------------------------------
bool IsValidId(lil_u16 id, LilErrorType *pErr) {
	if (id == ID_WIN_BMP || id == ID_OS2_BMP) {
		if (pErr)
			*pErr = LE_NONE;
		return true;
	}
	if (pErr)
		*pErr = LE_BADHEADER;
	return false;
}

} // namespace



//-----------------------------------------------------------------------------------------------------------
bool IsExtensionBmp(const char *fileName)
{
	if (CheckExtension(fileName, ".bmp") || CheckExtension(fileName, ".dib"))
		return true;
	return false;
}

//-----------------------------------------------------------------------------------------------------------
bool IsDetectableBmp(const LilIoDesc *io)
{
	io->seekFunc(io->handle, 0, LS_SET);
	lil_u16 id = Read16le(io);
	io->seekFunc(io->handle, 0, LS_SET);
	return IsValidId(id, 0);
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType LoadBmp(const LilIoDesc *io, LilImageDesc *image, LilLoadType request, const LilFormatDesc *imageFormat)
{
	//------------------------
	// 1) VARIABLES TO BE READ
	LilErrorType err = LE_NONE;

	lil_i32 fileSize = 0, reserved = 0, dataOffset = 0, headerSize = 0;
	lil_i32 width = 0, height = 0, compression = LC_ANY, imageSize = 0, ppmX = 0, ppmY = 0;
	lil_i32 colors = 0, importantColors = 0;
	lil_u16 id = 0, numPlanes = 0, bitsPerPixel = 0;
	lil_u8 palChannels = 4; // Default - OS2 bitmaps are different though
	lil_u32 rMask = 0, gMask = 0, bMask = 0;
	lil_u32 skipHeaderBytes = 0;  // There's some stuff we just won't read.

	// We're going to ignore imageSize and fileSize because some programs save corrupted versions of these

	//------------------------
	// 2a) READ FILE FORMAT HEADER
	io->seekFunc(io->handle, 0, LS_SET);
	ReadFle(io, "24444", &id, &fileSize, &reserved, &dataOffset, &headerSize);
	// Validate
	if (!IsValidId(id, &err))
		return err;
	LIL_ASSERT(id != ID_OS2_BMP);  // I want to see when this happens to see if it actually works

	//------------------------
	// 2b) READ SELECTED IMAGE INFO
	bool littleEndian = false;
	if (!CheckValue(headerSize, GOOD_H, N_GOOD_H, BAD_H, N_BAD_H, LE_UNSUPPORTEDFILEFORMAT, LE_BADHEADER, &err))
		return err;
	if (headerSize == HEADER_SIZE_V1) {
		lil_u16 w, h;
		ReadFle(io, "2222", &w, &h, &numPlanes, &bitsPerPixel);
		width = static_cast<lil_i32>(w); height = static_cast<lil_i32>(h);
		palChannels = 3;
	}
	else
		ReadFle(io, "4422 444444", &width, &height, &numPlanes, &bitsPerPixel,
			&compression, &imageSize, &ppmX, &ppmY, &colors, &importantColors);
//		case HEADER_SIZE_V5: // Untested - Can't find any!
//			skipHeaderBytes += HEADER_SIZE_V5 - HEADER_SIZE_V4;
//		case HEADER_SIZE_V4: // Untested - Can't find any!
//			skipHeaderBytes += HEADER_SIZE_V4 - HEADER_SIZE_V3;

	// The compression enum values for bmps happen to be +1 from the values in the files
	compression++;

	if (compression == LC_BMP_BITFIELDS)
		ReadFle(io, "444", &rMask, &gMask, &bMask);

	io->seekFunc(io->handle, skipHeaderBytes, LS_RELATIVE);

	//------------------------
	// 2c) READ PALETTE INFO
	if (bitsPerPixel <= 8) {
		lil_u32 currPos = static_cast<lil_u32>(io->tellFunc(io->handle));
		lil_u32 numColors = (dataOffset - currPos) / palChannels;
		if (numColors > (1u << bitsPerPixel))
			numColors = 1 << bitsPerPixel;
		if (!ReadPaletteWin(io, image, numColors, palChannels, &err))
			return err;
		io->seekFunc(io->handle, dataOffset, LS_SET);
	}

	//------------------------
	// 3) SET IMAGE, FORMAT, EXTENDED INFO
	if (!SetImageSize(&image->data, width, abs(height), &err))
		return err;
	if (!SetFormatFromBppWin(&image->format, static_cast<lil_u8>(bitsPerPixel), height > 0, &err))
		return err;
	if (!SetExtendedInfo(&image->extended, 1, compression, LD_METERS, ppmX, ppmY, &err))
		return err;

	// UGH Bitfields compression is kind of weird between different bmp files
	if (!CheckValue(compression, GOOD_C, N_GOOD_C, BAD_C, N_BAD_C, LE_UNSUPPORTEDCOMPRESSION, LE_BADCOMPRESSION, &err))
		return err;
	if (compression == LC_BMP_NONE || compression == LC_BMP_RLE8 || compression == LC_BMP_RLE4) {
		if (CalcBpp(&image->format) == 16)
			littleEndian = true;
	}
	else if (compression == LC_BMP_BITFIELDS) {
		if (bitsPerPixel != 32 && bitsPerPixel != 16)
			return LE_BADBPP;
		if (skipHeaderBytes)
			skipHeaderBytes -= 12;
		// Some bmps just have really crazy values so we attempt to compensate.  Some may have
		// higher bit masks than the max possible value so we just ignore higher bits with the
		// "max" variable.  Others actually overlap masks, so we give red priority and go from
		// there as you can see in the code...
		lil_u32 max = bitsPerPixel == 32 ? 0xffffffff : (1 << bitsPerPixel) - 1;
		lil_u8 rBits = BitGetCount(rMask & max);
		max = ((1 << BitGetLsb(rMask)) - 1);
		lil_u8 gBits = BitGetCount(gMask & max);
		max = ((1 << BitGetLsb(gMask)) - 1);
		lil_u8 bBits = BitGetCount(bMask & max);
		//max = ((1 << BitGetLsb(bMask)) - 1);
		lil_u8 xBits = static_cast<lil_u8>(bitsPerPixel - rBits - gBits - bBits);
		//xBitsLow = max ? true : false;

		int chOffset = xBits ? 1 : 0;
		image->format.bitCounts[0] = xBits;
		image->format.bitCounts[0 + chOffset] = rBits;
		image->format.bitCounts[1 + chOffset] = gBits;
		image->format.bitCounts[2 + chOffset] = bBits;
		if (!SetFormat(&image->format, LDF_SIMPLEBITS, 0, static_cast<lil_u8>(3 + chOffset), 4, false,
			chOffset ? LCO_ARGB : LCO_RGB, height > 0, &err))
			return err;
		littleEndian = true;
	}

	// This is the only possible way to simplify channels in this case...
	if (SimplifyFormat(&image->format)) {
		if (image->format.bitsPerChannel == 8) {
			image->format.channelOrder = LCO_BGRA;
			littleEndian = false;
		}
	}

	//------------------------
	// 4) LOAD IMAGE DATA
	if (request == LL_INFO)
		return LE_NONE;

	if (compression == LC_BMP_RLE8 || compression == LC_BMP_RLE4) {
		LilCodecRleBmp codec(io, image, &image->format, compression == LC_BMP_RLE4);
		return ReadImageData(codec.GetIoPtr(), image, imageFormat, littleEndian, true);
	}

	return ReadImageData(io, image, imageFormat, littleEndian, true);
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType SaveBmp(const LilIoDesc *io, const LilImageDesc *image, const LilFormatDesc *imageFormat)
{
	//------------------------
	// 1) VARIABLES TO BE WRITTEN
	LilErrorType err = LE_NONE;

	LilFormatDesc outFormat = imageFormat ? *imageFormat : image->format;
	lil_u16 bpp = static_cast<lil_u16>(CalcBpp(&outFormat));
	lil_u32 imageSize = static_cast<lil_u32>(CalcByteSize(&outFormat, image->data.width, image->data.height));
	bool detectComp = image->extended.compression == LC_ANY;
	LilCompressionType compression = detectComp ? LC_BMP_NONE : image->extended.compression;
	lil_u32 importantColors = 0;
	lil_i32 height = image->data.height;
	LilBuffer tempPalBuf;
	LilPaletteDesc tempPal = {0};
	lil_u32 fileSize = 0;

	lil_u32 palUseType = 2, headerSize = HEADER_SIZE_V3;
	lil_u16 headerType = ID_WIN_BMP;

	const lil_u32 reserved = 0;  // Must be 0
	const lil_u16 numPlanes = 1; // Must be 1

	//------------------------
	// 2) GET IMAGE, FORMAT, EXTENDED INFO
	// Check format that we will be saving.  Compensate for any possibility to use LC_BMP_BITFIELDS.
	if (!MatchFormat(&outFormat, ranges, N_RANGES, ranges2, N_RANGES2, &err))
		return err;
	// We know we found a good format, so we test against defaults, and if it doesn't work we
	// know to use LC_BMP_BITFIELDS
	LilFormatDesc compareFormat;
	bool needBitFields = true;
	if (SetFormatFromBppWin(&compareFormat, static_cast<lil_u8>(bpp), true, &err)) {
		if (!IsFormatDifferent(&outFormat, &compareFormat, IFD_IGNORE_FLIPPED | IFD_IGNORE_PLANAR | IFD_IGNORE_STRIDE))
			needBitFields = false;
	}
	if (needBitFields)
		compression = LC_BMP_BITFIELDS;
	if (compression == LC_BMP_BITFIELDS && outFormat.dataFormat == LDF_SIMPLE) {
		for (lil_u32 i = 0; i < outFormat.numChannels; i++)
			outFormat.bitCounts[i] = outFormat.bitsPerChannel;
		outFormat.bitsPerChannel = 0;
		outFormat.dataFormat = LDF_SIMPLEBITS;
	}
	if (detectComp) {
		if (bpp == 8)
			compression = LC_BMP_RLE8;
		if (bpp == 4)
			compression = LC_BMP_RLE4;
		if (bpp == 8 || bpp == 4)
			outFormat.flipped = true;
	}

	tempPal.numEntries = 0;
	if (bpp <= 8) {
        lil_u32 colors = image->palette.numEntries;
        if (colors > (1u << bpp))
            colors = 1 << bpp;
        if (!colors)
            colors = 1 << bpp;
        if (!SetPalette(&tempPal, image, &PALETTE_FORMAT_V3, colors, &tempPalBuf, &err))
            return err;
    }

	// Needed all previous information to calculate these
	lil_u32 dataOffset = 14 + headerSize + tempPal.numEntries * tempPal.format.numChannels +
        (compression == LC_BMP_BITFIELDS ? 12 : 0);
	fileSize = dataOffset + imageSize;

	//------------------------
	// 3a) WRITE FILE FORMAT HEADER
	// 3b) WRITE IMAGE INFO
	// 3c) WRITE PALETTE INFO
	io->seekFunc(io->handle, 0, LS_SET);

	WriteFle(io, "24444", headerType, fileSize, reserved, dataOffset, headerSize);

	if (palUseType == 1) {
		// OS2 has 16 bit fields for w and h
		lil_u16 w = static_cast<lil_u16>(image->data.width);
		lil_u16 h = static_cast<lil_u16>(image->data.height);
		WriteFle(io, "2222", w, h, numPlanes, bpp);
	}
	else {
		WriteFle(io, "4422 444444", image->data.width, outFormat.flipped ? height : -height,
			numPlanes, bpp, compression - 1, imageSize,
			image->extended.dpuX, image->extended.dpuY,
			tempPal.numEntries, importantColors);
		if (compression == LC_BMP_BITFIELDS) {
			lil_u32 rMask = 0, gMask = 0, bMask = 0;
			lil_u32 addC = outFormat.numChannels == 4 ? 1 : 0;
			bMask = (1 << outFormat.bitCounts[2 + addC]) - 1;
			gMask = ((1 << outFormat.bitCounts[1 + addC]) - 1) << outFormat.bitCounts[2 + addC];
			rMask = ((1 << outFormat.bitCounts[0 + addC]) - 1) << (outFormat.bitCounts[2 + addC] + outFormat.bitCounts[1 + addC]);
			WriteFle(io, "444", rMask, gMask, bMask);
		}
	}

	WriteN(io, tempPal.numEntries * tempPal.format.numChannels, tempPal.lut);

	//------------------------
	// 4) SAVE IMAGE DATA
	if (compression == LC_BMP_RLE8 || compression == LC_BMP_RLE4) {
		LilBuffer buf;
		if (!buf.Realloc(&image->memFuncs, imageSize))
			return LE_NOTENOUGHMEMORY;
		LilCodecRleBmp codec(buf.GetIoDescPtr(), image, &outFormat, compression == LC_BMP_RLE4);
		err = WriteImageData(codec.GetIoPtr(), image, &outFormat, true);
		// There's this weird undocumented behavior in windows which prohibits compressed bmp data from being
		// larger in size than what the raw image data would be.  When this happens, windows truncates the
		// data being read.
		// Soooo... we write to memory and if we go past the raw image size, we'll just write the data raw.
		if (err == LE_NONE) {
			lil_u32 compSize = static_cast<lil_u32>(buf.GetIoDescPtr()->tellFunc(buf.GetIoDescPtr()->handle));
			WriteN(io, compSize, buf.GetPtr<void>());
			io->seekFunc(io->handle, 2, LS_SET);
			Write32le(io, dataOffset + compSize);
			return LE_NONE;
		}
		if (err != LE_IOOUTOFRANGE)
			return err;
		
		// Ok so here we need to re-write the compression used to none, since RLE compression won't work.
		// @@ At this point the only thing we potentially lost was the desire to have image data
		// UNflipped.  Compressed bitmaps must be flipped (apparently)
		io->seekFunc(io->handle, 30, LS_SET);
		Write32le(io, LC_BMP_NONE - 1);
		io->seekFunc(io->handle, dataOffset, LS_SET);
	}

	return WriteImageData(io, image, &outFormat, true);
}

} // namespace lil

#endif // LIL_SUPPORTS_BMP
