#include "LilPlatform.h"

#ifdef LIL_SUPPORTS_GIF

#include <memory>
#include "LilCodecLzw.h"
#include "LilUtilFormat.h"
#include "LilMem.h"
#include "LilConvert.h"
#include "LilBitManip.h"
#include "LilIo.h"

namespace lil {

//-----------------------------------------------------------------------------------------------------------
namespace {

const lil_u8 SYMBOL_IMAGE_START = 0x2c;
const lil_u8 SYMBOL_FILE_END = 0x3b;
const lil_u8 SYMBOL_BLOCK_END = 0x00;
const lil_u8 SYMBOL_BLOCK_START_89 = 0x21;
const lil_u8 SYMBOL_GRAPHICS_CONTROL_89 = 0xf9;
const lil_u8 SYMBOL_PLAIN_TEXT_89 = 0x01;
const lil_u8 SYMBOL_APP_89 = 0xff;
const lil_u8 SYMBOL_COMMENT_89 = 0xfe;

//-----------------------------------------------------------------------------------------------------------
const LilFormatRange ranges[] = {
	{ LDF_SIMPLE, { {1}, {2}, {3}, {4}, {5}, {6}, {7}, {8} }, 8, 1, 1, FLAG_NO, LCO_SINGLE, FLAG_NO }
};
const lil_u8 N_RANGES = LIL_ARRAY_SIZE(ranges);

const LilFormatDesc PALETTE_FORMAT = { LDF_SIMPLE, 8, {}, 3, 4, false, LCO_RGB, false };


//-----------------------------------------------------------------------------------------------------------
bool IsValidId(lil_i8 sig[3], lil_i8 version[3]) {
	if (sig[0] == 'G' && sig[1] == 'I' && sig[2] == 'F'
		&& version[0] == '8' && (version[1] == '7' || version[1] == '9') && version[2] == 'a')
	{
		return true;
	}
	return false;
}

//-----------------------------------------------------------------------------------------------------------
// Create an LilIoDesc for gif image sub-blocks
LilErrorType LilIoGifRead(lil_io_handle handle, size_t byteCount, void *outMemPtr);
LilErrorType LilIoGifWrite(lil_io_handle handle, size_t byteCount, const void *inMemPtr);

struct GifIoHandle {
	GifIoHandle(const LilIoDesc *io) : pIoInput(io), bytesLeft(0) {
		ioOutput.handle = reinterpret_cast<lil_io_handle>(this);
		ioOutput.readFunc = &LilIoGifRead;
		ioOutput.writeFunc = &LilIoGifWrite;
		ioOutput.seekFunc = 0;
		ioOutput.tellFunc = 0;
	}
	const LilIoDesc *GetIoPtr() const { return &ioOutput; }
	LilErrorType FlushWrite() {
	    if (bytesLeft) {
            Write8(pIoInput, bytesLeft);
            LilErrorType err = WriteN(pIoInput, bytesLeft, writeBuf);
            bytesLeft = 0;
            return err;
	    }
	    return LE_NONE;
	}

	const LilIoDesc *pIoInput;
	size_t bytesLeft;  // in sub-block
	lil_u8 writeBuf[255];

private:
	LilIoDesc ioOutput;
};

LilErrorType LilIoGifRead(lil_io_handle handle, size_t byteCount, void *outMemPtr) {
	lil_u8 *out8 = reinterpret_cast<lil_u8 *>(outMemPtr);
	GifIoHandle *hGif = reinterpret_cast<GifIoHandle *>(handle);
	while (byteCount) {
		if (hGif->bytesLeft == 0) {
			hGif->bytesLeft = Read8(hGif->pIoInput);
			if (hGif->bytesLeft == 0)
				return LE_IOOUTOFRANGE;  // No more sub blocks - shouldn't be trying to read more!
		}
		size_t readBytes = GetMin(byteCount, hGif->bytesLeft);
		LilErrorType err = hGif->pIoInput->readFunc(hGif->pIoInput->handle, readBytes, out8);
		if (err != LE_NONE)
			return err;
		out8 += readBytes;
		byteCount -= readBytes;
		hGif->bytesLeft -= readBytes;
	}
	return LE_NONE;
}

LilErrorType LilIoGifWrite(lil_io_handle handle, size_t byteCount, const void *inMemPtr) {
	const lil_u8 *in8 = reinterpret_cast<const lil_u8 *>(inMemPtr);
	GifIoHandle *hGif = reinterpret_cast<GifIoHandle *>(handle);
	while (byteCount) {
		lil_u32 allowed = 255 - hGif->bytesLeft;
		lil_u32 added = allowed < byteCount ? allowed : byteCount;
		memcpy(hGif->writeBuf + hGif->bytesLeft, in8, added);
		hGif->bytesLeft += added;
		in8 += added;
		byteCount -= added;
		if (hGif->bytesLeft == 255) {
            hGif->bytesLeft = 0;
            Write8(hGif->pIoInput, 255);
            LilErrorType err = WriteN(hGif->pIoInput, 255, hGif->writeBuf);
            if (err != LE_NONE)
                return err;
        }
	}
	return LE_NONE;
}

} // namespace



//-----------------------------------------------------------------------------------------------------------
bool IsExtensionGif(const char *fileName)
{
	if (CheckExtension(fileName, ".GIF"))
		return true;
	return false;
}

//-----------------------------------------------------------------------------------------------------------
bool IsDetectableGif(const LilIoDesc *io)
{
	io->seekFunc(io->handle, 0, LS_SET);
	lil_i8 sig[3], ver[3];
	ReadN(io, 3, sig);
	ReadN(io, 3, ver);
	return IsValidId(sig, ver);
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType LoadGif(const LilIoDesc *io, LilImageDesc *image, LilLoadType request, const LilFormatDesc *imageFormat)
{
	// Read file to memory first...
	lil_u64 size = GetIoSize(io);
	LilBuffer buf;
	if (!buf.Realloc(&image->memFuncs, size))
		return LE_NOTENOUGHMEMORY;
	io->readFunc(io->handle, size, buf.GetPtr<lil_u8>());
	const LilIoDesc *pMemIo = io;//buf.GetIoDescPtr();

	//------------------------
	// 1) VARIABLES TO BE READ
	LilErrorType err = LE_NONE;

	lil_i8 sig[3], ver[3];

	bool hasPalette = false;
	lil_u16 width = 0, height = 0;
	lil_u8 packed = 0, bgColor = 0, aspectRatio = 0, channelDepth = 0;
	lil_u32 paletteSize = 0;
	lil_u32 numImages = 0, selImageOffset = 0;  // So we can quickly find it later when we load.

	lil_u8 blockId = 0;
	lil_u8 sizeByte = 0, end = 0;
	bool done = false;

	//------------------------
	// 2a) READ FILE FORMAT HEADER
	pMemIo->seekFunc(pMemIo->handle, 0, LS_SET);
	ReadN(pMemIo, 3, sig);
	ReadN(pMemIo, 3, ver);
	if (!IsValidId(sig, ver))
		return LE_BADHEADER;

	//------------------------
	// 2b) READ SELECTED IMAGE INFO

	// These will be the same for ALL images inside the gif file
	ReadFle(pMemIo, "22 111", &width, &height, &packed, &bgColor, &aspectRatio);

	// For packed...
	// Bits 0-2 Size of the Global Color Table
	// Bit 3 Color Table Sort Flag
	// Bits 4-6 Color Resolution
	// Bit 7 Global Color Table Flag
	paletteSize = 2 << (packed & 0x07);
	channelDepth = 1 + ((packed & 0x70) >> 4);
	hasPalette = (packed & 0x80) ? true : false;

	//------------------------
	// 2c) READ PALETTE INFO

	// ignore bg color if no palette
	// Not sure: DO we or DON'T we use channelDepth to calculate these next 2 lines?
	if (hasPalette) {
		ReadHandle(&image->memFuncs, pMemIo, 3 * paletteSize, &image->palette.lut, 0);
		image->palette.numEntries = paletteSize;
		if (!SetFormat(&image->palette.format, LDF_SIMPLE, 8, 3, 0, false, LCO_RGB, false, &err))
			return err;
	}

	//------------------------
	// 3) SET IMAGE, FORMAT, EXTENDED INFO
	while (!done) {
		blockId = Read8(pMemIo);
		switch (blockId) {
		case SYMBOL_IMAGE_START:
			{
				lil_u16 left = 0, top = 0;
				ReadFle(pMemIo, "22221", &left, &top, &width, &height, &packed);
				lil_u32 lPaletteSize = 2 << (packed & 0x07);
				bool interlaced = (packed & 0x40) ? true : false;
				bool hasLPalette = (packed & 0x80) ? true : false;
				if (numImages == image->extended.pageIndex) {
					if (hasLPalette) {
						// Overwrite with a local palette if it has one.
						ReadHandle(&image->memFuncs, pMemIo, 3 * lPaletteSize, &image->palette.lut, 0);
						image->palette.numEntries = paletteSize;
						if (!SetFormat(&image->palette.format, LDF_SIMPLE, 8, 3, 0, false, LCO_RGB, false, &err))
							return err;
					}
					// After getting info we'll know where in file to start reading the actual image.
					selImageOffset = pMemIo->tellFunc(pMemIo->handle);
					image->data.width = width;
					image->data.height = height;
					if (!SetFormat(&image->format, LDF_SIMPLE, channelDepth, 1, 0, false, LCO_SINGLE, false, &err))
						return err;
					image->extended.fileFormat.gif.interlaced = interlaced;
					image->extended.fileFormat.gif.bgColorIndex = bgColor;
				}
				else if (hasLPalette)
					pMemIo->seekFunc(pMemIo->handle, 3 * lPaletteSize, LS_RELATIVE);
				// Now skip over the image for now.....
				// We remember where in the file it was in selImageOffset if it is the one
				// we are/were looking for.
				pMemIo->seekFunc(pMemIo->handle, 1, LS_RELATIVE);  // minCodeSize
				sizeByte = Read8(pMemIo);
				while (sizeByte != SYMBOL_BLOCK_END) {
					err = pMemIo->seekFunc(pMemIo->handle, sizeByte, LS_RELATIVE);
                    if (err != LE_NONE)
                        return err;
					sizeByte = Read8(pMemIo);
				}
				numImages++;
			}
			break;
		case SYMBOL_BLOCK_START_89:
			blockId = Read8(pMemIo);
			sizeByte = Read8(pMemIo);
			pMemIo->seekFunc(pMemIo->handle, sizeByte, LS_RELATIVE);
			switch (blockId) {
			case SYMBOL_GRAPHICS_CONTROL_89:
				end = Read8(pMemIo);
				if (end != SYMBOL_BLOCK_END)
					return LE_BADHEADER;
				break;
			case SYMBOL_PLAIN_TEXT_89:
				// Has subblocks of text data
				sizeByte = Read8(pMemIo);
				while (sizeByte != SYMBOL_BLOCK_END) {
					pMemIo->seekFunc(pMemIo->handle, sizeByte, LS_RELATIVE);
					sizeByte = Read8(pMemIo);
				}
				break;
			case SYMBOL_APP_89:
				// Has subblocks of application/user defined data - could be anything
				sizeByte = Read8(pMemIo);
				while (sizeByte != SYMBOL_BLOCK_END) {
					pMemIo->seekFunc(pMemIo->handle, sizeByte, LS_RELATIVE);
					sizeByte = Read8(pMemIo);
				}
				break;
			case SYMBOL_COMMENT_89:
				// Note, for comment block, we've already read the first part and possibly only part of
				// the comment!  At least with the current impl above.
				sizeByte = Read8(pMemIo);
				while (sizeByte != SYMBOL_BLOCK_END) {
					pMemIo->seekFunc(pMemIo->handle, sizeByte, LS_RELATIVE);
					sizeByte = Read8(pMemIo);
				}
				break;
			default:
				return LE_BADHEADER;
			}
			break;
		case SYMBOL_FILE_END:
			done = true;
			break;
		default:
			return LE_BADHEADER;
		}
	}

	if (!SetExtendedInfo(&image->extended, numImages, LC_GIF_LZW, LD_INCHES, 0, 0, &err))
		return err;

	//------------------------
	// 4) LOAD IMAGE DATA
	if (request == LL_INFO)
		return LE_NONE;

	pMemIo->seekFunc(pMemIo->handle, selImageOffset, LS_SET);

	// Get min code size.  Ensure min code size is according to spec.  I don't think it can
	// ever actually be anything different than what this checks for?
	lil_u8 minCodeSize = Read8(pMemIo);
	if (image->format.bitsPerChannel == 1) {
		if (minCodeSize != 2)
			return LE_BADCOMPRESSION;
	}
	else if (minCodeSize != image->format.bitsPerChannel) {
		//LIL_ASSERT(0);
		//return LE_BADCOMPRESSION;
	}

	GifIoHandle hGif(pMemIo);
	std::auto_ptr<LilCodecLzw> codec(new LilCodecLzw(hGif.GetIoPtr(), image,
		&image->format, image->extended.fileFormat.gif.interlaced, minCodeSize));
	const LilIoDesc *pIo = codec->GetIoPtr();

	return ReadImageData(pIo, image, imageFormat, false, false);
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType SaveGif(const LilIoDesc *io, const LilImageDesc *image, const LilFormatDesc *imageFormat)
{
	//------------------------
	// 1) VARIABLES TO BE WRITTEN
	LilErrorType err = LE_NONE;

	LilFormatDesc outFormat = imageFormat ? *imageFormat : image->format;

	LilPaletteDesc tempPal;
	LilBuffer tempPalBuf;
	lil_u16 w = static_cast<lil_u16>(image->data.width), h = static_cast<lil_u16>(image->data.height);
	lil_u8 packed = 0, bgColor = 0, aspectRatio = 0;
	lil_u32 colors = 0;

	//------------------------
	// 2) GET IMAGE, FORMAT, EXTENDED INFO
	if (!MatchFormat(&outFormat, ranges, N_RANGES, 0, 0, &err))
		return err;
	if (image->data.width > 65535 || image->data.height > 65535)
		return LE_UNSUPPORTEDSIZE;

	// Setup palette...
    colors = 1 << outFormat.bitsPerChannel;
    if (!SetPalette(&tempPal, image, &PALETTE_FORMAT, colors, &tempPalBuf, &err))
        return err;

	// For packed...
	// Bits 0-2 Size of the Global Color Table
	// Bit 3 Color Table Sort Flag
	// Bits 4-6 Color Resolution
	// Bit 7 Global Color Table Flag
	packed = (colors ? 0x80 | BitGetMsb(colors - 1) : 0);
	packed |= (outFormat.bitsPerChannel - 1) << 4;
	bgColor = colors ? image->extended.fileFormat.gif.bgColorIndex : 0;
	aspectRatio = 0;

	//------------------------
	// 3a) WRITE FILE FORMAT HEADER
	// 3b) WRITE IMAGE INFO
	// 3c) WRITE PALETTE INFO
	io->seekFunc(io->handle, 0, LS_SET);
	WriteFle(io, "111111", 'G', 'I', 'F', '8', '9', 'a');

	// These will be the same for ALL images inside the gif file
	WriteFle(io, "22 111", w, h, packed, bgColor, aspectRatio);
	WriteN(io, colors * 3, tempPal.lut);

	Write8(io, SYMBOL_IMAGE_START);

	// @@ Interlaced = no for now.
	if (image->extended.fileFormat.gif.interlaced)
		return LE_UNSUPPORTEDFILEFEATURE;
	// @@ Also no local palette for now - at least not until multiple image savings are implemented
	packed = 0;

	WriteFle(io, "22221", ZERO_16, ZERO_16, w, h, packed);

	//------------------------
	// 4) SAVE IMAGE DATA

	// Min code size
	lil_u8 minCodeSize = outFormat.bitsPerChannel == 1 ? 2 : outFormat.bitsPerChannel;
	Write8(io, minCodeSize);

	GifIoHandle hGif(io);
	LilCodecLzw codec(hGif.GetIoPtr(), image, &outFormat, image->extended.fileFormat.gif.interlaced, minCodeSize);

	err = WriteImageData(codec.GetIoPtr(), image, &outFormat, true);
	if (err != LE_NONE)
        return err;
	codec.codeWriter.FlushByte();
	err = hGif.FlushWrite();
	if (err == LE_NONE) {
		Write8(io, SYMBOL_BLOCK_END);
		Write8(io, SYMBOL_FILE_END);
	}

	return err;
}

}  // namespace lil

#endif // LIL_SUPPORTS_GIF
