#include "LilPlatform.h"

#if defined(LIL_SUPPORTS_ICO) || defined(LIL_SUPPORTS_CUR)

#include <memory>
#include "LilCodecRle.h"
#include "LilUtilFormat.h"
#include "LilMem.h"
#include "LilConvert.h"
#include "LilBitManip.h"
#include "LilIo.h"
#include "Lil.h"

namespace lil {

//-----------------------------------------------------------------------------------------------------------
namespace {

const lil_u32 HEADER_DIR_SIZE = 16;
const lil_u32 HEADER_ID_ICO = 0x00010000;
const lil_u32 HEADER_ID_CUR = 0x00020000;
const lil_u32 BMP_HEADER_SIZE_V3 = 40;
const lil_u32 BMP_NO_COMPRESSION = 0;

// Compressions
const lil_i32 GOOD_C_ICO[] = { LC_ICO_NONE, LC_ICO_PNG };
const lil_u8 N_GOOD_C_ICO = LIL_ARRAY_SIZE(GOOD_C_ICO);
const lil_i32 GOOD_C_CUR[] = { LC_CUR_NONE };
const lil_u8 N_GOOD_C_CUR = LIL_ARRAY_SIZE(GOOD_C_CUR);

// Valid formats
const LilFormatRange RANGES_ICO[] = {
	{ LDF_SIMPLE, { {1}, {4}, {8} }, 3, 1, 4, FLAG_NO, LCO_SINGLE, FLAG_NO },
	{ LDF_SIMPLE, { {1} }, 1, 2, 4, FLAG_NO, LCO_YA, FLAG_NO },
	{ LDF_SIMPLE, { {8} }, 1, 3, 4, FLAG_NO, LCO_BGR, FLAG_NO },
	{ LDF_SIMPLE, { {8} }, 1, 4, 4, FLAG_NO, LCO_BGRA, FLAG_NO },
};
const lil_u8 N_RANGES_ICO = LIL_ARRAY_SIZE(RANGES_ICO);

LilFormatDesc PALETTE_FORMAT = { LDF_SIMPLE, 8, {}, 4, 4, false, LCO_BGRA, false };

//-----------------------------------------------------------------------------------------------------------
bool IsValidId(lil_u32 id, bool isIco) {
	if (isIco && id == HEADER_ID_ICO)
		return true;
	if (!isIco && id == HEADER_ID_CUR)
		return true;
	return false;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType LoadIcoCur(const LilIoDesc *io, LilImageDesc *image, LilLoadType request, const LilFormatDesc *imageFormat, bool isIco)
{
	//------------------------
	// 1) VARIABLES TO BE READ
	LilErrorType err = LE_NONE;

	lil_i16 numImages = 0;
	lil_u8 reserved8 = 0, width = 0, height = 0, colorCount = 0;  // colors == 0 if >= 8bpp
	lil_u16 planes = 0, bpp = 0;
	lil_u32 id = 0, byteSize = 0, filePos = 0;
	lil_u32 paletteSize	= 0;
	lil_i32 headerSize = 0, width2 = 0, height2 = 0, compression = 0, imageSize = 0,
		ppmX = 0, ppmY = 0, colors = 0, importantColors = 0;
	lil_i16 numPlanes2 = 0, bitsPerPixel2 = 0;
	lil_u32 bpl = 0;

	//------------------------
	// 2a) READ FILE FORMAT HEADER
	io->seekFunc(io->handle, 0, LS_SET);
	ReadFle(io, "42", &id, &numImages);
	if (!IsValidId(id, isIco))
		return LE_BADHEADER;

	//------------------------
	// 2b) READ SELECTED IMAGE INFO
	io->seekFunc(io->handle, image->extended.pageIndex * HEADER_DIR_SIZE, LS_RELATIVE);

	// HEADER_STRUCT_SIZE is the combination of this entire read op.
	ReadFle(io, "1111 22 44", &width, &height, &colorCount, &reserved8, &planes, &bpp, &byteSize, &filePos);
	paletteSize = (bpp == 8 ? 256 : colorCount);

	io->seekFunc(io->handle, filePos, LS_SET);

	ReadFle(io, "4 4422 444444", &headerSize, &width2, &height2, &numPlanes2, &bitsPerPixel2,
		&compression, &imageSize, &ppmX, &ppmY, &colors, &importantColors);

	// Check if it's a vista-png-icon!
	// *** THIS LOADS THE ICON AND RETURNS FROM THE FUNCTION IF TRUE ***
	//137 80 78 71 13 10 26 10
	if (headerSize == 0x474e5089 && width2 == 0x0a1a0a0d) {
		io->seekFunc(io->handle, -40, LS_RELATIVE);

		RaiiIo ioSub(IoCloseSub);
		IoOpenSub(&ioSub, io, io->tellFunc(io->handle), byteSize);
		return LilImageLoad(LFF_PNG, &ioSub, image, request, imageFormat);
	}

	//------------------------
	// 2c) READ PALETTE INFO
	if (!ReadPaletteWin(io, image, paletteSize, 4, &err))
		return err;

	//------------------------
	// 3) SET IMAGE, FORMAT, EXTENDED INFO
	if (!SetImageSize(&image->data, width, height, &err))
		return err;
	if (!SetFormatFromBppWin(&image->format, static_cast<lil_u8>(bpp), false, &err))
		return err;
	if (!SetExtendedInfo(&image->extended, numImages, LC_ICO_NONE, LD_INCHES, 0, 0, &err))
		return err;

	// If ico's support 16 bit - I haven't seen one yet (though I didn't look SUPER hard).
	// Just need to add support if we end up here one day...
	if (image->format.dataFormat == LDF_SIMPLEBITS)
		return LE_UNSUPPORTEDFILEFEATURE;

	//------------------------
	// 4) LOAD IMAGE DATA
	if (request == LL_INFO)
		return LE_NONE;

	bpl = LilImageGetBpl(image);
	LilBuffer buf, buf4;
	if (!buf.Realloc(&image->memFuncs, bpl * height))
		return LE_NOTENOUGHMEMORY;
	RaiiIo memIo(IoCloseMem);
	ReadN(io, bpl * height, buf.GetPtr<lil_u8>());

	// Convert the standard .ico and mask into an alpha channel
	if (image->format.numChannels < 4) {
		LilFormatDesc alphaFormat, oldFormat;
		oldFormat = image->format;
		if (!SetFormat(&alphaFormat, LDF_SIMPLE, 1, 1, 4, false, LCO_SINGLE, oldFormat.flipped, &err))
			return err;

		// This might convert grayscale to YA type as well...
		image->format.numChannels ++;
		if (image->format.channelOrder == LCO_BGR)
			image->format.channelOrder = LCO_BGRA;
		if (image->format.channelOrder == LCO_SINGLE)
			image->format.channelOrder = LCO_YA;

		// Convert old to new...
		if (!buf4.Realloc(&image->memFuncs, static_cast<size_t>(LilImageGetByteSize(image))))
			return LE_NOTENOUGHMEMORY;
		IoOpenMem(&memIo, buf4.GetPtr<lil_u8>(), buf4.GetSize());
		ConvertData(&image->memFuncs, buf.GetPtr<lil_u8>(), &oldFormat, 0,
			buf4.GetPtr<lil_u8>(), &image->format, 0, width, height, true);

		// Now copy over new image->data and fill it W/alpha channel as well.
		lil_u32 andSize = (RoundUpToNearest(width, 32) / 8) * height;
		LilBuffer temp;
		temp.Realloc(&image->memFuncs, andSize);
		ReadN(io, andSize, temp.GetPtr<lil_u8>());
		BitInvertBuf(temp.GetPtr<lil_u8>(), andSize);  // AND mask for dest...
		LilErrorType err = ConvertChannelToChannel(temp.GetPtr<lil_u8>(), &alphaFormat, 0,
			buf4.GetPtr<lil_u8>(), &image->format, image->format.numChannels - 1, width, height);
		if (err != LE_NONE)
			return err;
	}
	else
		IoOpenMem(&memIo, buf.GetPtr<lil_u8>(), buf.GetSize());

	return ReadImageData(&memIo, image, imageFormat, true, true);
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType SaveIcoCur(const LilIoDesc *io, const LilImageDesc *image, const LilFormatDesc *imageFormat, bool isIco)
{
	//------------------------
	// 1) VARIABLES TO BE WRITTEN
	LilErrorType err = LE_NONE;

	LilFormatDesc outFormat = imageFormat ? *imageFormat : image->format;
	lil_u16 bpp = 0;
	lil_u32 imageSize = 0;
	bool detectComp = image->extended.compression == LC_ANY;
	LilCompressionType compression = detectComp ? LC_ICO_NONE : image->extended.compression;
	LilBuffer tempPalBuf;
	LilPaletteDesc tempPal = {0};

	const lil_u16 numPlanes = 1; // Must be 1
	lil_u16 numPages = static_cast<lil_u16>(lil::GetMax(image->extended.numPages, image->extended.pageIndex + 1));

	//------------------------
	// 2) GET IMAGE, FORMAT, EXTENDED INFO
	// Check format that we will be saving.
	if (!MatchFormat(&outFormat, RANGES_ICO, N_RANGES_ICO, 0, 0, &err))
		return err;
	if (outFormat.channelOrder == LCO_YA) {
		outFormat.channelOrder = LCO_SINGLE;
		outFormat.numChannels = 1;
	};
	bpp = static_cast<lil_u16>(CalcBpp(&outFormat));
	if (image->data.width > 256 || image->data.width == 0 ||
		image->data.height > 256 || image->data.height == 0)
	{
		return LE_BADIMAGESIZE;
	}

	if (detectComp) {
		if (bpp == 32)
			compression = LC_ICO_PNG;
	}

	tempPal.numEntries = 0;
	if (bpp <= 8) {
        lil_u32 colors = image->palette.numEntries;
        if (colors > (1u << bpp))
            colors = 1 << bpp;
        if (!colors)
            colors = 1 << bpp;
        if (!SetPalette(&tempPal, image, &PALETTE_FORMAT, colors, &tempPalBuf, &err))
            return err;
    }

	imageSize = static_cast<lil_u32>(CalcByteSize(&outFormat, image->data.width, image->data.height));

	// Needed all previous information to calculate these
//	lil_u32 dataOffset = 14 + headerSize + tempPal.numEntries * tempPal.format.numChannels +
//        (compression == LC_BMP_BITFIELDS ? 12 : 0);
//	fileSize = dataOffset + imageSize;

	//------------------------
	// 3a) WRITE FILE FORMAT HEADER
	// 3b) WRITE IMAGE INFO
	// 3c) WRITE PALETTE INFO
	io->seekFunc(io->handle, 0, LS_SET);

	WriteFle(io, "42", isIco ? HEADER_ID_ICO : HEADER_ID_CUR, numPages);

	io->seekFunc(io->handle, image->extended.pageIndex * HEADER_DIR_SIZE, LS_RELATIVE);

	// HEADER_STRUCT_SIZE is the combination of this entire write op.
	lil_u32 filePos = static_cast<lil_u32>(GetIoSize(io) + HEADER_DIR_SIZE);
	if (isIco)
		WriteFle(io, "1111 22 44", image->data.width, image->data.height, tempPal.numEntries,
			ZERO_8, numPlanes, bpp, imageSize, filePos);
	else
		WriteFle(io, "1111 22 44", image->data.width, image->data.height, tempPal.numEntries,
			ZERO_8, image->extended.fileFormat.cur.hotPointX, image->extended.fileFormat.cur.hotPointY,
			imageSize, filePos);

	io->seekFunc(io->handle, filePos, LS_SET);

	if (compression != LC_ICO_PNG) {
		WriteFle(io, "4 4422 444444", BMP_HEADER_SIZE_V3, image->data.width, image->data.height * 2,
			numPlanes, bpp, BMP_NO_COMPRESSION, imageSize, 0, 0, tempPal.numEntries, 0);
		WriteN(io, tempPal.numEntries * tempPal.format.numChannels, tempPal.lut);
	}

	//------------------------
	// 4) SAVE IMAGE DATA
	if (compression == LC_ICO_PNG) {
		RaiiIo ioSub(IoCloseSub);
		IoOpenSub(&ioSub, io, filePos, imageSize);
		return LilImageSave(LFF_PNG, &ioSub, image, imageFormat);
	}

	lil_u32 andSize = (RoundUpToNearest(image->data.width, 32) / 8) * image->data.height;
	LilBuffer andData;
	if (!andData.Realloc(&image->memFuncs, andSize))
		return LE_NOTENOUGHMEMORY;

	// Either pull out alpha channel or create a solid alpha channel
	if (HasAlpha(&image->format)) {
		LilBuffer buf4;
		if (!buf4.Realloc(&image->memFuncs, static_cast<size_t>(LilImageGetByteSize(image))))
			return LE_NOTENOUGHMEMORY;
		LilFormatDesc alphaFormat;
		if (!SetFormat(&alphaFormat, LDF_SIMPLE, 1, 1, 4, false, LCO_SINGLE, false, &err))
			return err;
		WriteImageData(buf4.GetIoDescPtr(), image, &image->format, true);
		LilErrorType err = ConvertChannelToChannel(buf4.GetPtr<lil_u8>(), &image->format,
			LIL_ALPHA_CHANNEL[image->format.channelOrder], andData.GetPtr<lil_u8>(), &alphaFormat, 0,
			image->data.width, image->data.height);
		if (err != LE_NONE)
			return err;
	}
	else
		memset(andData.GetPtr<lil_u8>(), 0, andData.GetSize());

	err = WriteImageData(io, image, &outFormat, true);
	if (err != LE_NONE)
		return err;

	WriteN(io, andData.GetSize(), andData.GetPtr<lil_u8>());

	return LE_NONE;
}

}  // namespace



#ifdef LIL_SUPPORTS_ICO

//-----------------------------------------------------------------------------------------------------------
bool IsExtensionIco(const char *fileName)
{
	if (CheckExtension(fileName, ".ico"))
		return true;
	return false;
}

//-----------------------------------------------------------------------------------------------------------
bool IsDetectableIco(const LilIoDesc *io)
{
	io->seekFunc(io->handle, 0, LS_SET);
	lil_u32 id = Read32le(io);
	io->seekFunc(io->handle, 0, LS_SET);
	return IsValidId(id, true);
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType LoadIco(const LilIoDesc *io, LilImageDesc *image, LilLoadType request, const LilFormatDesc *imageFormat)
{
	return LoadIcoCur(io, image, request, imageFormat, true);
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType SaveIco(const LilIoDesc *io, const LilImageDesc *image, const LilFormatDesc *imageFormat)
{
	return SaveIcoCur(io, image, imageFormat, true);
}

#endif  // LIL_SUPPORTS_ICO

#ifdef LIL_SUPPORTS_CUR

//-----------------------------------------------------------------------------------------------------------
bool IsExtensionCur(const char *fileName)
{
	if (CheckExtension(fileName, ".cur"))
		return true;
	return false;
}

//-----------------------------------------------------------------------------------------------------------
bool IsDetectableCur(const LilIoDesc *io)
{
	io->seekFunc(io->handle, 0, LS_SET);
	lil_u32 id = Read32le(io);
	io->seekFunc(io->handle, 0, LS_SET);
	return IsValidId(id, false);
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType LoadCur(const LilIoDesc *io, LilImageDesc *image, LilLoadType request, const LilFormatDesc *imageFormat)
{
	return LoadIcoCur(io, image, request, imageFormat, false);
}


//-----------------------------------------------------------------------------------------------------------
LilErrorType SaveCur(const LilIoDesc *io, const LilImageDesc *image, const LilFormatDesc *imageFormat)
{
	return SaveIcoCur(io, image, imageFormat, false);
}

#endif  // LIL_SUPPORTS_CUR

}  // namespace lil

#endif  // LIL_SUPPORTS_ICO || LIL_SUPPORTS_CUR
