#include "LilPlatform.h"

#ifdef LIL_SUPPORTS_PNG

#include "LilConvert.h"
#include "LilMem.h"
#include "LilBitManip.h"
#include "LilIo.h"
#include "LilUtilFormat.h"
#include "LilCodec.h"

namespace lil {

namespace {

	//-----------------------------------------------------------------------------------------------------------
	const lil_u32 ID1 = 0x89504e47;
	const lil_u32 ID2 = 0x0d0a1a0a;

	bool IsValidId(lil_u32 id1, lil_u32 id2) {
		//137 80 78 71 13 10 26 10
		if (id1 == ID1 && id2 == ID2)
			return true;
		return false;
	}

	// '*' means can have multiple
	const lil_u32 CHUNK_IHDR = 'IHDR';  //0x49484452; // first

	const lil_u32 CHUNK_cHRM = 'cHRM';  //0x6348524d; // before PLTE
	const lil_u32 CHUNK_gAMA = 'gAMA';  //0x67414d41; // before PLTE
	const lil_u32 CHUNK_iCCP = 0; // before PLTE
	const lil_u32 CHUNK_sBIT = 'sBIT';  //0x73424954; // before PLTE
	const lil_u32 CHUNK_sRGB = 1; // before PLTE

	const lil_u32 CHUNK_PLTE = 'PLTE';  //0x504c5445; // before IDAT

	const lil_u32 CHUNK_bKGD = 'bKGD';  //0x624b4744; // between PLTE and IDAT
	const lil_u32 CHUNK_hIST = 'hIST';  //0x68495354; // between PLTE and IDAT
	const lil_u32 CHUNK_tRNS = 'tRNS';  //0x74524e53; // between PLTE and IDAT
	const lil_u32 CHUNK_pHYs = 'pHYs';  //0x70485973; // before IDAT
	const lil_u32 CHUNK_sPLT = 2; // before IDAT *

	const lil_u32 CHUNK_IDAT = 'IDAT';  //0x49444154; // >1 must be consecutive! *

	const lil_u32 CHUNK_tIME = 'tIME';  //0x74494d45; // no order
	const lil_u32 CHUNK_tEXt = 'tEXt';  //0x74455874; // no order *
	const lil_u32 CHUNK_zTXt = 'zTXt';  //0x7a545874; // no order *
	const lil_u32 CHUNK_iTXt = 3; // no order *

	const lil_u32 CHUNK_IEND = 'IEND';  //0x49454e44; // last

	//-----------------------------------------------------------------------------------------------------------
	const lil_u8 COMPRESSION_DEFLATE = 0;

	//-----------------------------------------------------------------------------------------------------------
	const lil_u8 INTERLACE_NONE = 0;
	const lil_u8 INTERLACE_ADAM7 = 1;

	//-----------------------------------------------------------------------------------------------------------
	const lil_u8 FILTER_NONE = 0;
	const lil_u8 FILTER_SUB = 1;
	const lil_u8 FILTER_UP = 2;
	const lil_u8 FILTER_AVG = 3;
	const lil_u8 FILTER_PAETH = 4;

	//-----------------------------------------------------------------------------------------------------------
	const LilFormatRange ranges[] = {
		{ LDF_SIMPLE, { {1}, {2}, {4}, {8}, {16} }, 5, 1, 1, FLAG_NO, LCO_SINGLE, FLAG_NO },
		{ LDF_SIMPLE, { {8}, {16} }, 2, 2, 1, FLAG_NO, LCO_YA, FLAG_NO },
		{ LDF_SIMPLE, { {8}, {16} }, 2, 3, 1, FLAG_NO, LCO_RGB, FLAG_NO },
		{ LDF_SIMPLE, { {8}, {16} }, 2, 4, 1, FLAG_NO, LCO_RGBA, FLAG_NO },
	};
	const lil_u8 N_RANGES = LIL_ARRAY_SIZE(ranges);

	LilFormatDesc PALETTE_FORMAT = { LDF_SIMPLE, 8, {}, 3, 1, false, LCO_RGB, false };


	//-----------------------------------------------------------------------------------------------------------
	LIL_INLINE lil_u8 Paeth(lil_u8 a, lil_u8 b, lil_u8 c) {
		lil_u32 p = a + b - c;
		lil_u32 pa = abs(static_cast<int>(p - a));
		lil_u32 pb = abs(static_cast<int>(p - b));
		lil_u32 pc = abs(static_cast<int>(p - c));
		if (pa <= pb && pa <= pc)
			return a;
		else if (pb <= pc)
			return b;
		return c;
	}

	//-----------------------------------------------------------------------------------------------------------
	// Does 16 bit swap if need-be as well.
	LilErrorType UnfilterImageData(lil_u8 *pOut, lil_u8 *pData, lil_u32 lineBytes, lil_u32 pixelBytesCeil, lil_u32 height, bool ByteSwap16) {
		// Do in-place reverse filtering
		lil_u8 *pPosFilt = pData;
		// First line treats all aboves as 0
		for (lil_u32 iY = 0; iY < height; iY++) {
			lil_u8 filter = *pPosFilt++;
			lil_u8 *pLeft = pOut;  // stays static until pixelBytesCeil read
			lil_u8 *pUp = pOut - lineBytes;  // moves with pOut
			if (filter == FILTER_NONE || (iY == 0 && filter == FILTER_UP))
				for (lil_u32 iB = 0; iB < lineBytes; iB++)
					*pOut++ = *pPosFilt++;
			else if (filter == FILTER_SUB || (iY == 0 && filter == FILTER_PAETH)) {
				for (lil_u32 iB = 0; iB < pixelBytesCeil; iB++)
					*pOut++ = *pPosFilt++;
				for (lil_u32 iB = pixelBytesCeil; iB < lineBytes; iB++)
					*pOut++ = (*pPosFilt++ + *pLeft++) & 0xff;
			}
			else if (iY == 0 && filter == FILTER_AVG) {
				for (lil_u32 iB = 0; iB < pixelBytesCeil; iB++)
					*pOut++ = *pPosFilt++;
				for (lil_u32 iB = pixelBytesCeil; iB < lineBytes; iB++)
					*pOut++ = (*pPosFilt++ + (*pLeft++ >> 1)) & 0xff;
			}
			// At this point any filter that needs to go *up* a scanline will be ok to do so.
			else if (filter == FILTER_UP)
				for (lil_u32 iB = 0; iB < lineBytes; iB++)
					*pOut++ = (*pPosFilt++ + *pUp++) & 0xff;
			else if (filter == FILTER_AVG) {
				for (lil_u32 iB = 0; iB < pixelBytesCeil; iB++)
					*pOut++ = (*pPosFilt++ + (*pUp++ >> 1)) & 0xff;
				for (lil_u32 iB = pixelBytesCeil; iB < lineBytes; iB++)
					*pOut++ = (*pPosFilt++ + ((*pLeft++ + *pUp++) >> 1)) & 0xff;
			}
			else if (filter == FILTER_PAETH) {
				lil_u8 *pUpLeft = pOut - lineBytes;  // stays static until pixelBytesCeil read
				for (lil_u32 iB = 0; iB < pixelBytesCeil; iB++)
					*pOut++ = (*pPosFilt++ + *pUp++) & 0xff;
				for (lil_u32 iB = pixelBytesCeil; iB < lineBytes; iB++)
					*pOut++ = (*pPosFilt++ + Paeth(*pLeft++, *pUp++, *pUpLeft++)) & 0xff;
			}
			else
				return LE_BADIMAGEDATA;

	#ifndef LIL_ENDIAN_BIG
			// Swap prev line, but not the current one just created because next scanline's filter
			// may need this one - and it will need to be in the original byte order.
			if (ByteSwap16 && iY > 0)
				ByteSwapBuf16(pOut - lineBytes * 2, lineBytes / 2);
	#endif
		}
	#ifndef LIL_ENDIAN_BIG
		// Ok to do last line now
		if (ByteSwap16)
			ByteSwapBuf16(pOut - lineBytes, lineBytes / 2);
	#endif
		return LE_NONE;
	}

	//-----------------------------------------------------------------------------------------------------------
	LilErrorType FilterImageData(const LilMemDesc *pMem, lil_u8 *pOut, lil_u8 *pData, lil_u32 lineBytes, lil_u32 pixelBytesCeil, lil_u32 height) {
		pOut += (height - 1) * (lineBytes + 1);  // Needed for proper pointer offset in loop
		lil_mem_handle mh = 0;
		LilErrorType err = LE_NONE;
		if (!SetError(pMem->memReallocFunc(&mh, (lineBytes + 1) * 5), &err))
			return err;

		lil_u8 *lineBuf = reinterpret_cast<lil_u8 *>(mh);
		lil_u32 endY = height - 1;
		for (lil_u32 iY = 0; iY <= endY; iY++) {
			lil_u8 best = 0;
			lil_u32 bestVal = 0xffffffff;
			for (lil_u8 filter = 0; filter < 5; filter++ ) {
				lil_u8 *pPosFilt = pData + height * lineBytes - 1 - lineBytes * iY;
				lil_u8 *pLineOut = lineBuf + filter * (lineBytes + 1) + lineBytes;
				lil_u8 *pLeft = pPosFilt - pixelBytesCeil;
				lil_u8 *pUp = pPosFilt - lineBytes;

				if (filter == FILTER_NONE || (iY == endY && filter == FILTER_UP))
					for (lil_u32 iB = 0; iB < lineBytes; iB++)
						*pLineOut-- = *pPosFilt--;
				else if (filter == FILTER_SUB || (iY == endY && filter == FILTER_PAETH)) {
					for (lil_u32 iB = pixelBytesCeil; iB < lineBytes; iB++)
						*pLineOut-- = (*pPosFilt-- - *pLeft--) & 0xff;
					for (lil_u32 iB = 0; iB < pixelBytesCeil; iB++)
						*pLineOut-- = *pPosFilt--;
				}
				else if (iY == endY && filter == FILTER_AVG) {
					for (lil_u32 iB = pixelBytesCeil; iB < lineBytes; iB++)
						*pLineOut-- = (*pPosFilt-- - (*pLeft-- >> 1)) & 0xff;
					for (lil_u32 iB = 0; iB < pixelBytesCeil; iB++)
						*pLineOut-- = *pPosFilt--;
				}
				// At this point any filter that needs to go *up* a scanline will be ok to do so.
				else if (filter == FILTER_UP)
					for (lil_u32 iB = 0; iB < lineBytes; iB++)
						*pLineOut-- = (*pPosFilt-- - *pUp--) & 0xff;
				else if (filter == FILTER_AVG) {
					for (lil_u32 iB = pixelBytesCeil; iB < lineBytes; iB++)
						*pLineOut-- = (*pPosFilt-- - ((*pLeft-- + *pUp--) >> 1)) & 0xff;
					for (lil_u32 iB = 0; iB < pixelBytesCeil; iB++)
						*pLineOut-- = (*pPosFilt-- - (*pUp-- >> 1)) & 0xff;
				}
				else if (filter == FILTER_PAETH) {
					lil_u8 *pUpLeft = pPosFilt - lineBytes - pixelBytesCeil;
					for (lil_u32 iB = pixelBytesCeil; iB < lineBytes; iB++)
						*pLineOut-- = (*pPosFilt-- - Paeth(*pLeft--, *pUp--, *pUpLeft--)) & 0xff;
					for (lil_u32 iB = 0; iB < pixelBytesCeil; iB++)
						*pLineOut-- = (*pPosFilt-- - *pUp--) & 0xff;
				}
				lil_u32 sum = 0;
				for (lil_u32 iB = 1; iB <= lineBytes; iB++)
					sum += abs(static_cast<lil_i8>(pLineOut[iB]));
				if (sum < bestVal) {
					bestVal = sum;
					best = filter;
				}

				*pLineOut = filter;
			}
			memcpy(pOut, lineBuf + best * (lineBytes + 1), lineBytes + 1);
			pOut -= lineBytes + 1;
		}
		pMem->memReallocFunc(&mh, 0);
		return LE_NONE;
	}

	//-----------------------------------------------------------------------------------------------------------
	// Create an LilIoDesc for consecutive png image blocks
	LilErrorType LilIoPngRead(lil_io_handle handle, size_t byteCount, void *outMemPtr);
	LilErrorType LilIoPngWrite(lil_io_handle handle, size_t byteCount, const void *outMemPtr);

	struct PngIoHandle {
		PngIoHandle(const LilImageDesc *pImage, const LilIoDesc *io) :
			pIoInput(io), bytesLeft(0), pImage(pImage)
		{
			ioOutput.handle = reinterpret_cast<lil_io_handle>(this);
			ioOutput.readFunc = &LilIoPngRead;
			ioOutput.writeFunc = &LilIoPngWrite;
			ioOutput.seekFunc = 0;
			ioOutput.tellFunc = 0;
		}
		const LilIoDesc *GetIoPtr() const { return &ioOutput; }

		LilErrorType FlushWrite() {
			// This has to be lil_u32 since that's all that can be written in the png
			lil_u32 dataLen = static_cast<lil_u32>(mem.GetSize() - 4 - bytesLeft);
			if (dataLen > 0) {
				Write32be(pIoInput, dataLen);

				const lil_u8 *pData = mem.GetPtr<const lil_u8>();
				WriteN(pIoInput, dataLen + 4, pData);
				lil_u32 crc = CalcCrc32(pData, dataLen + 4);
				Write32be(pIoInput, crc);
				bytesLeft = mem.GetSize() - 4;
			}
			return LE_NONE;
		}

		const LilIoDesc *pIoInput;
		size_t bytesLeft;  // in sub-block
		LilBuffer mem;
		const LilImageDesc *pImage;

	private:
		LilIoDesc ioOutput;
	};

	LilErrorType LilIoPngRead(lil_io_handle handle, size_t byteCount, void *outMemPtr) {
		lil_u8 *out8 = reinterpret_cast<lil_u8 *>(outMemPtr);
		PngIoHandle *hPng = reinterpret_cast<PngIoHandle *>(handle);
		const LilIoDesc *pMemIo = hPng->mem.GetIoDescPtr();
		while (byteCount) {
			if (hPng->bytesLeft == 0) {
				lil_u32 dataLen = Read32be(hPng->pIoInput);

				if (!hPng->mem.Realloc(&hPng->pImage->memFuncs, dataLen + 4))
					return LE_NOTENOUGHMEMORY;
				// Data + ChunkId go into crc calc.
				lil_u8 *pData = hPng->mem.GetPtr<lil_u8>();
				ReadN(hPng->pIoInput, 4, pData);

				// Mem IO pointing to output of current IO
				lil_u32 chunkType = Read32be(pMemIo);  // now memIo points to data only

				if (chunkType != CHUNK_IDAT)
					return LE_BADCOMPRESSION;
				else {
					ReadN(hPng->pIoInput, dataLen, pData + 4);
					lil_u32 crc = Read32be(hPng->pIoInput);
					if (crc != CalcCrc32(pData, dataLen + 4))
						return LE_BADCRC;
					hPng->bytesLeft = dataLen;
				}
			}
			size_t readBytes = GetMin(byteCount, hPng->bytesLeft);
			LilErrorType err = ReadN(pMemIo, readBytes, out8);
			if (err != LE_NONE)
				return err;
			out8 += readBytes;
			byteCount -= readBytes;
			hPng->bytesLeft -= readBytes;
		}
		return LE_NONE;
	}

	LilErrorType LilIoPngWrite(lil_io_handle handle, size_t byteCount, const void *inMemPtr) {
		const lil_u8 *in8 = reinterpret_cast<const lil_u8 *>(inMemPtr);
		PngIoHandle *hPng = reinterpret_cast<PngIoHandle *>(handle);
		const LilIoDesc *pMemIo = hPng->mem.GetIoDescPtr();
		if (hPng->mem.GetSize() == 0) {
			if (!hPng->mem.Realloc(&hPng->pImage->memFuncs, 1024 * 1024 * 512 + 4))
				return LE_NOTENOUGHMEMORY;
			hPng->bytesLeft = hPng->mem.GetSize() - 4;
			Write32be(pMemIo, CHUNK_IDAT);
		}
		while (byteCount) {
			size_t writeBytes = GetMin(byteCount, hPng->bytesLeft);
			LilErrorType err = WriteN(pMemIo, writeBytes, in8);
			if (err != LE_NONE)
				return err;
			in8 += writeBytes;
			byteCount -= writeBytes;
			hPng->bytesLeft -= writeBytes;

			if (hPng->bytesLeft == 0) {
				// This has to be lil_u32 since that's all that can be written in the png
				lil_u32 dataLen = static_cast<lil_u32>(hPng->mem.GetSize() - 4);
				Write32be(hPng->pIoInput, dataLen);

				const lil_u8 *pData = hPng->mem.GetPtr<const lil_u8>();
				WriteN(hPng->pIoInput, dataLen + 4, pData);
				lil_u32 crc = CalcCrc32(pData, dataLen + 4);
				Write32be(hPng->pIoInput, crc);
				hPng->bytesLeft = dataLen;
			}
		}
		return LE_NONE;
	}

} // namespace



//-----------------------------------------------------------------------------------------------------------
bool IsExtensionPng(const char *fileName)
{
	if (CheckExtension(fileName, ".PNG"))
		return true;
	return false;
}

//-----------------------------------------------------------------------------------------------------------
bool IsDetectablePng(const LilIoDesc *io)
{
	//137 80 78 71 13 10 26 10
	io->seekFunc(io->handle, 0, LS_SET);
	lil_u32 id1 = Read32be(io);
	lil_u32 id2 = Read32be(io);

	io->seekFunc(io->handle, 0, LS_SET);
	return IsValidId(id1, id2);
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType LoadPng(const LilIoDesc *io, LilImageDesc *image, LilLoadType request, const LilFormatDesc *imageFormat)
{
	//------------------------
	// 1) VARIABLES TO BE READ
	LilErrorType err = LE_NONE;
	lil_u32 id1 = 0, id2 = 0;
	lil_u64 imageOffset = 0, compBytes = 0;  // Seek location in file to actually read image stream
	bool done = false;
	lil_u8 colorType = 0, compression = 0, filter = 0, interlace = 0, bits = 0;
	// Safety flags
	bool firstIdat = true, prevIdat = false, firstChunk = true, needPal = false, hasPal = false;
	LilBuffer buf;  // holds data so it can be crc checked

	//------------------------
	// 2a) READ FILE FORMAT HEADER
	io->seekFunc(io->handle, 0, LS_SET);
	// MSB byte order

	//137 80 78 71 13 10 26 10
	ReadFbe(io, "44", &id1, &id2);
	if (!IsValidId(id1, id2))
		return LE_BADHEADER;

	//------------------------
	// 2b) READ SELECTED IMAGE INFO
	// 2c) READ PALETTE INFO
	// 3) SET IMAGE, FORMAT, EXTENDED INFO
	while (!done) {
		lil_u32 dataLen = Read32be(io);
		if (dataLen > 0x7fffffff)
			return LE_BADHEADER;

		if (!buf.Realloc(&image->memFuncs, dataLen + 4))
			return LE_NOTENOUGHMEMORY;
		lil_u8 *pData = buf.GetPtr<lil_u8>();

		// Data + ChunkId go into crc calc.
		ReadN(io, 4, pData);
		lil_u32 chunkType = Read32be(buf.GetIoDescPtr());

		if (firstChunk != (chunkType == CHUNK_IHDR))
			return LE_BADHEADER;

		// We'll read and test idat chunks when loading image, not info.
		if (chunkType != CHUNK_IDAT) {
			ReadN(io, dataLen, pData + 4);
			lil_u32 crc = Read32be(io);
			if (crc != CalcCrc32(pData, dataLen + 4))
				return LE_BADCRC;
		}
		else
			io->seekFunc(io->handle, dataLen + 4, LS_RELATIVE);  // data + crc

		switch (chunkType) {
		// CRITICAL CHUNKS
		case CHUNK_IHDR:
			{
				// The standard is very specific about what is valid and what isn't so we have
				// a good bit of extra error checking to do even though Lil could theoretically
				// support most of those odder combinations...
				// All are obvious - except index 3 is lookup table
				LilChannelOrderType lco[7] = {
					LCO_SINGLE, LCO_UNKNOWN, LCO_RGB, LCO_SINGLE, LCO_YA, LCO_UNKNOWN, LCO_RGBA };
				lil_u8 num[7] = { 1, 0, 3, 1, 2, 0, 4 };

				lil_u32 w = 0, h = 0;
				ReadFbe(buf.GetIoDescPtr(), "4411111", &w, &h,
					&bits, &colorType, &compression, &filter, &interlace);
				if (w == 0 || h == 0 || w > 0x7fffffff || h > 0x7fffffff)
					return LE_BADHEADER;
				// 1, 2, 4, 8, and 16 are valid - that's it!
				if (bits > 16 || BitGetCount(bits) > 1)
					return LE_BADBPP;
				if (colorType == 1 || colorType == 5 || colorType > 6)
					return LE_BADCOLORSPACE;
				if (num[colorType] > 1 && (bits & 7))
					return LE_BADBPP;
				if (colorType == 3 && bits > 8)
					return LE_BADBPP;
				if (compression != COMPRESSION_DEFLATE)  // standard allows for future expansion but so far nothing
					return LE_BADCOMPRESSION;
				if (interlace != INTERLACE_NONE && interlace != INTERLACE_ADAM7)
					return LE_BADHEADER;
				if (filter != 0)
					return LE_BADHEADER;

				if (!SetImageSize(&image->data, w, h, &err))
					return err;
				if (!SetFormat(&image->format, LDF_SIMPLE, bits, num[colorType], 1, false, lco[colorType], false, &err))
					return err;
				if (!SetExtendedInfo(&image->extended, 1, LC_PNG_ZIP, LD_INCHES, 0, 0, &err))
					return err;

				image->extended.fileFormat.png.interlaced = (interlace == INTERLACE_NONE ? false : true);
				if (colorType == 3)
					needPal = true;
			}
			break;
		case CHUNK_PLTE:
			if (colorType == 0 || colorType == 4)
				return LE_BADPALETTE;
			// For RGB images this is suggested palette if quantization needed on older displays...
			if (dataLen % 3 != 0 || dataLen / 3 > static_cast<lil_u32>(1 << bits) || dataLen == 0)
				return LE_BADPALETTE;
			hasPal = true;
			ReadHandle(&image->memFuncs, buf.GetIoDescPtr(), dataLen, &image->palette.lut, 0);
			image->palette.numEntries = dataLen / 3;
			if (!SetFormat(&image->palette.format, LDF_SIMPLE, 8, 3, 1, false, LCO_RGB, false, &err))
				return err;
			break;
		case CHUNK_IDAT:
			// pos - crc(4) - data - chunk type(4) - data len(4)
			// Remember we had skipped all the data for CHUNK_IDAT above when testing the CRC...
			if (firstIdat)
				imageOffset = io->tellFunc(io->handle) - 4 - dataLen - 4 - 4;
			if (!firstIdat && !prevIdat) // prevIdat set at end of switch
				return LE_BADIMAGEDATA;  // All IDAT chunks must be consecutive
			firstIdat = false;
			compBytes += dataLen;
			break;
		case CHUNK_IEND:  // (should be) empty chunk
			done = true;
			break;

		// ANCILLARY CHUNKS
		case CHUNK_cHRM:  // dependant on image data
		case CHUNK_gAMA:  // dependant on image data
		case CHUNK_sBIT:  // dependant on image data
		case CHUNK_iCCP:  // dependant on image data
		case CHUNK_sRGB:  // dependant on image data

		case CHUNK_pHYs:
			image->extended.dpuX = Read32be(buf.GetIoDescPtr());
			image->extended.dpuY = Read32be(buf.GetIoDescPtr());
			image->extended.dotsPerWhat = Read8(buf.GetIoDescPtr()) ? LD_METERS : LD_UNKNOWN;
			break;
		case CHUNK_bKGD:  // dependant on image data - background color to render image against
		case CHUNK_hIST:  // dependant on image data
		case CHUNK_tRNS:  // dependant on image data
			// Convert palette to 4 channel
			// Must precede idat, but proceed plte
		case CHUNK_sPLT:  // dependant on image data

		case CHUNK_tIME:  // dependant on image data
		case CHUNK_tEXt:  // safe to copy as is
		case CHUNK_zTXt:  // safe to copy as is
		case CHUNK_iTXt:  // safe to copy as is
		default:
			// Is it a critical chunk???
			if ((chunkType & 0x20000000) == 0)  // uppercase first letter
				return LE_UNSUPPORTEDFILEFEATURE;
			if (chunkType & 0x20) {  // lowercase last letter
				// @@ This is an ancillary chunk plus it is ok to copy if we save to another png file...
				// We should perhaps store it somewhere in meta-data land.
			}
			break;
		}

		prevIdat = (chunkType == CHUNK_IDAT);
		firstChunk = false;
	}

	// No idat block found
	if (firstIdat)
		return LE_BADIMAGEDATA;
	if (firstChunk)
		return LE_BADHEADER;
	if (needPal && !hasPal)
		return LE_BADPALETTE;

	//------------------------
	// 4) LOAD IMAGE DATA
	if (request == LL_INFO)
		return LE_NONE;

	//---------------------------------
	// Final combined image
	lil_u32 lineBytes = LilImageGetBpl(image);
	// Code doesn't currently support images > 4gb
	lil_u32 imageBytes = static_cast<lil_u32>(LilImageGetByteSize(image));
	lil_u32 pixelBytesCeil = CalcMinBytesNeeded(&image->format, 1);

	// If interlaced then each scanline-change in interlace data has its own filter.
	// 6 passes go on the rounded-up half of the scanlines, and 1 pass goes on the
	// rounded-down half.  Thus we really end up with something like height * 3.5.
	lil_u32 extra = image->data.height * (interlace ? 4 : 1);  // filter bytes
	if (!buf.Realloc(&image->memFuncs, imageBytes + extra))
		return LE_NOTENOUGHMEMORY;
	lil_u8 *pData = buf.GetPtr<lil_u8>();

	// Find image
	io->seekFunc(io->handle, imageOffset, LS_SET);

	PngIoHandle hPng(image, io);

	// Decode
	err = DecodeZlib(pData, hPng.GetIoPtr(), compBytes, &image->memFuncs);
	if (err != LE_NONE)
		return err;

	if (!interlace) {
		// Non-interlaced reverse filtering
		err = UnfilterImageData(pData, pData, lineBytes, pixelBytesCeil,
			image->data.height, image->format.bitsPerChannel == 16);
		if (err != LE_NONE)
			return err;
	}
	else {
		// Interlaced reverse filtering
		lil_u32 xStart[7] = { 0, 4, 0, 2, 0, 1, 0 };
		lil_u32 yStart[7] = { 0, 0, 4, 0, 2, 0, 1 };
		lil_u32 xSkip[7] = { 8, 8, 4, 4, 2, 2, 1 };
		lil_u32 ySkip[7] = { 8, 8, 8, 4, 4, 2, 2 };
		lil_u8 *pDataStart = pData;
		LilBuffer buf2;
		if (!buf2.Realloc(&image->memFuncs, imageBytes + extra))
			return LE_NOTENOUGHMEMORY;
		lil_u8 *pOut = buf2.GetPtr<lil_u8>();
		for (int i = 0; i < 7; i++) {
			if (image->data.width <= xStart[i] || image->data.height <= yStart[i])
				continue;
			lil_u32 passW = (image->data.width - xStart[i] - 1) / xSkip[i] + 1;
			lil_u32 passH = (image->data.height - yStart[i] - 1) / ySkip[i] + 1;
			lil_u32 passLineBytes = CalcMinBytesNeeded(&image->format, passW);
			err = UnfilterImageData(pOut, pDataStart, passLineBytes, pixelBytesCeil,
				passH, image->format.bitsPerChannel == 16);
			// Remember to account for filter byte here
			pDataStart += (passLineBytes + 1) * passH;
			pOut += passLineBytes * passH;
			if (err != LE_NONE)
				return err;
		}

		// Deinterlace
		// This is slightly inconvenient if bpp < 8.
		// Regardless this needs better impl.
		pDataStart = buf2.GetPtr<lil_u8>();
		for (int i = 0; i < 7; i++) {
			if (image->data.width <= xStart[i] || image->data.height <= yStart[i])
				continue;
			pOut = pData + yStart[i] * lineBytes;
			lil_u32 passW = (image->data.width - xStart[i] - 1) / xSkip[i] + 1;
			lil_u32 passH = (image->data.height - yStart[i] - 1) / ySkip[i] + 1;
			lil_u32 passLineBytes = CalcMinBytesNeeded(&image->format, passW);
			for (lil_u32 iY = 0; iY < passH; iY++) {
				BitInMsb reader(pDataStart, passLineBytes);
				pDataStart += passLineBytes;
				BitIoMsb writer(pOut + xStart[i] * CalcBpp(&image->format) / 8);
				writer.SetBitPos((xStart[i] * CalcBpp(&image->format)) & 7);
				pOut += lineBytes * ySkip[i];
				lil_u32 value = 0;
				for (lil_u32 iX = 0; iX < passW; iX++) {
					for (lil_u32 j = 0; j < image->format.numChannels; j++) {
						reader.ReadBits(value, image->format.bitsPerChannel);
						writer.WriteBits(image->format.bitsPerChannel, value);
					}
					writer.SkipBits(CalcBpp(&image->format) * (xSkip[i] - 1));
				}
			}
		}
	}

	return ReadImageData(buf.GetIoDescPtr(), image, imageFormat, true, false);
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType SavePng(const LilIoDesc *io, const LilImageDesc *image, const LilFormatDesc *imageFormat)
{
	//------------------------
	// 1) VARIABLES TO BE WRITTEN
	LilErrorType err = LE_NONE;
	LilFormatDesc outFormat = imageFormat ? *imageFormat : image->format;
	LilBuffer buf;  // holds data so it can be crc calced
    const LilIoDesc *pMemIo = buf.GetIoDescPtr();
    lil_u8 interlacing = INTERLACE_NONE, colorType = 7;
    lil_u32 crc = 0, len = 0, bpp = CalcBpp(&outFormat);
    lil_u32 colors = image->palette.numEntries;
    LilPaletteDesc tempPal;
    LilBuffer tempPalBuf;

	//------------------------
	// 2) GET IMAGE, FORMAT, EXTENDED INFO
	if (!MatchFormat(&outFormat, ranges, N_RANGES, 0, 0, &err))
		return err;
	if (image->data.width > 0x7fffffff || image->data.height > 0x7fffffff)
		return LE_UNSUPPORTEDSIZE;
    if (image->extended.fileFormat.png.interlaced) {
        interlacing = INTERLACE_ADAM7;
        return LE_UNSUPPORTEDFILEFEATURE;
    }

    if (colors) {
        if (bpp <= 8)
            colors = 1 << bpp;
        else
            colors = 0;
        if (!SetPalette(&tempPal, image, &PALETTE_FORMAT, colors, &tempPalBuf, &err))
            return err;
    }

	//------------------------
	// 3a) WRITE FILE FORMAT HEADER
	// 3b) WRITE IMAGE INFO
	// 3c) WRITE PALETTE INFO
	io->seekFunc(io->handle, 0, LS_SET);
	WriteFbe(io, "44", ID1, ID2);

	// ********* CHUNK_IHDR **********
	len = 13;
	if (!buf.Realloc(&image->memFuncs, len + 4))
		return LE_NOTENOUGHMEMORY;
	Write32be(io, len);
    // len = data only
    // crc is calced on *chunktype* + data
	Write32be(pMemIo, CHUNK_IHDR);
    LilChannelOrderType lco[7] = {
        LCO_SINGLE, LCO_UNKNOWN, LCO_RGB, LCO_SINGLE, LCO_YA, LCO_UNKNOWN, LCO_RGBA };
    lil_u8 num[7] = { 1, 0, 3, 1, 2, 0, 4 };
    for (colorType = 0; colorType < 7; colorType++)
        if (outFormat.channelOrder == lco[colorType] && outFormat.numChannels == num[colorType])
            break;
    if (colorType == 1 || colorType == 5 || colorType > 6)
        return LE_INTERNALERROR;  // This should have been handled in MatchFormat
    if (colorType == 0 && colors)
        colorType = 3;

    WriteFbe(pMemIo, "4411111", image->data.width, image->data.height,
        outFormat.bitsPerChannel, colorType, COMPRESSION_DEFLATE, ZERO_8, interlacing);
    crc = CalcCrc32(buf.GetPtr<lil_u8>(), len + 4);
    WriteN(io, len + 4, buf.GetPtr<lil_u8>());
    Write32be(io, crc);

	//------------------------
	// 4) SAVE IMAGE DATA

    // ********* CHUNK_PLTE **********
    if (colorType == 3 && colors) {
        len = static_cast<lil_u32>(tempPalBuf.GetSize());
        if (!buf.Realloc(&image->memFuncs, len + 4))
			return LE_NOTENOUGHMEMORY;
        Write32be(pMemIo, CHUNK_PLTE);
        WriteN(pMemIo, len, tempPalBuf.GetPtr<lil_u8>());
        crc = CalcCrc32(buf.GetPtr<lil_u8>(), len + 4);
        Write32be(io, len);
        WriteN(io, len + 4, buf.GetPtr<lil_u8>());
        Write32be(io, crc);
    }

	// ********* CHUNK_IDAT **********
	lil_u32 pixelBytesCeil = CalcMinBytesNeeded(&outFormat, 1);
	lil_u32 lineBytes = CalcBpl(&outFormat, image->data.width);
	lil_u32 decodedBytes = (lineBytes + 1) * image->data.height;

	// Convert data to output format
	if (!buf.Realloc(&image->memFuncs, decodedBytes))
		return LE_NOTENOUGHMEMORY;
	err = WriteImageData(buf.GetIoDescPtr(), image, &outFormat, false);
    if (err != LE_NONE)
        return err;

    // Filter it to ready it for compression
    FilterImageData(&image->memFuncs, buf.GetPtr<lil_u8>(), buf.GetPtr<lil_u8>(), lineBytes, pixelBytesCeil, image->data.height);

    // Now compress it
    PngIoHandle hPng(image, io);

	// Encode
	err = EncodeZlib(buf.GetPtr<lil_u8>(), hPng.GetIoPtr(), decodedBytes, LCL_MAX, &image->memFuncs);
	if (err != LE_NONE)
		return err;

    err = hPng.FlushWrite();
    if (err != LE_NONE)
        return err;

	// ********* CHUNK_pHYs **********
	
	if (image->extended.dpuX && image->extended.dpuY) {
		lil_u32 dpuX = image->extended.dpuX, dpuY = image->extended.dpuY;
		LilDpuType type = image->extended.dotsPerWhat;
		if (type == LD_INCHES) {
			dpuX = static_cast<lil_u32>(static_cast<double>(dpuX) / 0.0254);
			dpuY = static_cast<lil_u32>(static_cast<double>(dpuY) / 0.0254);
			type = LD_METERS;
		}
		if (type == LD_CENTIMETERS) {
			dpuX *= 100;
			dpuY *= 100;
			type = LD_METERS;
		}
		
        if (!buf.Realloc(&image->memFuncs, 9 + 4))
			return LE_NOTENOUGHMEMORY;
        Write32be(pMemIo, CHUNK_pHYs);
		Write32be(pMemIo, dpuX);
		Write32be(pMemIo, dpuY);
		Write8(pMemIo, static_cast<lil_u8>(type));
        crc = CalcCrc32(buf.GetPtr<lil_u8>(), 9 + 4);
        Write32be(io, 9);
        WriteN(io, 9 + 4, buf.GetPtr<lil_u8>());
        Write32be(io, crc);
	}

	// ********* CHUNK_IEND **********
	len = 0;
	Write32be(io, len);
	lil_u32 iend = CHUNK_IEND;
#ifdef LIL_ENDIAN_LITTLE
	ByteSwap32(iend);
#endif
	crc = CalcCrc32(reinterpret_cast<const lil_u8 *>(&iend), 4);
	Write32be(io, CHUNK_IEND);
	Write32be(io, crc);

	return err;
}

}  // namespace lil

#endif // LIL_SUPPORTS_PNG
