#include "LilPlatform.h"

#ifdef LIL_SUPPORTS_PNM

#include <memory>
#include "LilUtilFormat.h"
#include "LilMem.h"
#include "LilConvert.h"
#include "LilIo.h"
#include "LilBitManip.h"

namespace lil {

//-----------------------------------------------------------------------------------------------------------
namespace {

	// Compressions
	const lil_i32 GOOD_C[] = { LC_PNM_NONE, LC_PNM_BINARY };
	const lil_u8 N_GOOD_C = LIL_ARRAY_SIZE(GOOD_C);

	// Valid formats
	const LilFormatRange rangesText[] = {
		{ LDF_SIMPLE, { {1}, {2}, {3}, {4}, {5}, {6}, {7}, {8}, {9}, {10}, {11}, {12}, {13}, {14}, {15}, {16}, 
			{17}, {18}, {19}, {20}, {21}, {22}, {23}, {24}, {25}, {26}, {27}, {28}, {29}, {30}, {31}, {32} }, 32, 1, 1, FLAG_NO, LCO_SINGLE, FLAG_NO },
		{ LDF_SIMPLE, { {1}, {2}, {3}, {4}, {5}, {6}, {7}, {8}, {9}, {10}, {11}, {12}, {13}, {14}, {15}, {16}, 
			{17}, {18}, {19}, {20}, {21}, {22}, {23}, {24}, {25}, {26}, {27}, {28}, {29}, {30}, {31}, {32} }, 32, 3, 1, FLAG_NO, LCO_RGB, FLAG_NO },
	};
	const LilFormatRange rangesBinary[] = {
		{ LDF_SIMPLE, { {1}, {8} }, 2, 1, 1, FLAG_NO, LCO_SINGLE, FLAG_NO },
		{ LDF_SIMPLE, { {8} }, 1, 3, 1, FLAG_NO, LCO_RGB, FLAG_NO },
	};

	const lil_u8 N_RANGES_TEXT = LIL_ARRAY_SIZE(rangesText);
	const lil_u8 N_RANGES_BINARY = LIL_ARRAY_SIZE(rangesBinary);

	//-----------------------------------------------------------------------------------------------------------
	char SkipSpace(char currChar, const LilIoDesc *io) {
		while (IsSpace(currChar))
			currChar = Read8(io);
		return currChar;
	}

	//-----------------------------------------------------------------------------------------------------------
	char SkipExtra(char currChar, const LilIoDesc *io) {
		while (currChar == '#' || IsSpace(currChar)) {
			if (currChar == '#') {
				while (currChar != '\n' && currChar != '\r')
					currChar = Read8(io);
			}
			while (IsSpace(currChar))
				currChar = Read8(io);
		}
		return currChar;
	}

	//-----------------------------------------------------------------------------------------------------------
	char ReadValue(char currChar, const LilIoDesc *io, int *pVal) {
		int num = 0;
		while (IsNum(currChar)) {
			num = num * 10 + currChar - '0';
			currChar = Read8(io);
		}
		*pVal = num;
		return currChar;
	}

	//-----------------------------------------------------------------------------------------------------------
	void WriteValue(const LilIoDesc *io, lil_u32 val) {
		char buf[11];
		int len = 0;
		do {
			lil_u32 div = val / 10;
			lil_u32 rem = val - div * 10;
			buf[len++] = '0' + rem;
			val = div;
		} while (val);
		while (len--)
			Write8(io, buf[len]);
	}

	//-----------------------------------------------------------------------------------------------------------
	bool IsValidId(lil_u8 p, lil_u8 num) {
		return (p == 'p' || p == 'P') && num >= '1' && num <= '6';
	}

	//-----------------------------------------------------------------------------------------------------------
	struct LilIoPnmHandle {
		const LilIoDesc *pIo;  // Actual source data
		char currChar;
		int bits;
	};

	//-----------------------------------------------------------------------------------------------------------
	LilErrorType LilIoPnmRead(lil_io_handle handle, size_t byteCount, void *outMemPtr) {
		LilIoPnmHandle *p = reinterpret_cast<LilIoPnmHandle *>(handle);
		int val = 0;
		p->currChar = SkipSpace(p->currChar, p->pIo);
		p->currChar = ReadValue(p->currChar, p->pIo, &val);

		return LE_NONE;
	}

	//-----------------------------------------------------------------------------------------------------------
	LilErrorType LilIoPnmWrite(lil_io_handle handle, size_t byteCount, const void *inMemPtr) {
		LilIoPnmHandle *p = reinterpret_cast<LilIoPnmHandle *>(handle);
		int val = 0;
		p->currChar = SkipSpace(p->currChar, p->pIo);
		p->currChar = ReadValue(p->currChar, p->pIo, &val);

		return LE_NONE;
	}

} // namespace



//-----------------------------------------------------------------------------------------------------------
bool IsExtensionPnm(const char *fileName) {
	return (CheckExtension(fileName, ".PNM") || CheckExtension(fileName, ".PGM") || 
		CheckExtension(fileName, ".PBM") || CheckExtension(fileName, ".PPM"));
}

//-----------------------------------------------------------------------------------------------------------
bool IsDetectablePnm(const LilIoDesc *io) {
	io->seekFunc(io->handle, 0, LS_SET);
	lil_u8 p = Read8(io);
	lil_u8 num = Read8(io);
	return IsValidId(p, num);
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType LoadPnm(const LilIoDesc *io, LilImageDesc *image, LilLoadType request, const LilFormatDesc *imageFormat) {
	//------------------------
	// 1) VARIABLES TO BE READ
	LilErrorType err = LE_NONE;
	lil_i32 w = 0, h = 0, maxVal = 0;
	lil_u8 type = 0, bits = 0, ch = 0;

	//------------------------
	// 2a) READ FILE FORMAT HEADER
	// 2b) READ SELECTED IMAGE INFO
	io->seekFunc(io->handle, 0, LS_SET);
	
	lil_u8 p = Read8(io);
	type = Read8(io);
	if (!IsValidId(p, type))
		return LE_BADHEADER;
	lil_u8 currChar = Read8(io);

	currChar = SkipExtra(currChar, io);
	if (!IsNum(currChar))
		return LE_BADHEADER;
	currChar = ReadValue(currChar, io, &w);

	currChar = SkipExtra(currChar, io);
	if (!IsNum(currChar))
		return LE_BADHEADER;
	currChar = ReadValue(currChar, io, &h);

	bits = 1;
	ch = 1;
	if (type == '2' || type == '3' || type == '5' || type == '6') {
		currChar = SkipExtra(currChar, io);
		if (!IsNum(currChar))
			return LE_BADHEADER;
		currChar = ReadValue(currChar, io, &maxVal);
		bits = 1 << (BitGetMsb(maxVal) + 1);
	}
	if (type == '3' || type == '6')
		ch = 3;
	bool isBinary = (type >= '4');
	if (isBinary && bits > 8)
		return LE_INVALIDFORMAT;
	if (isBinary && type >= '5')
		bits = 8;

	//------------------------
	// 3) SET IMAGE, FORMAT, EXTENDED INFO
	if (!SetImageSize(&image->data, w, h, &err))
		return err;
	if (!SetFormat(&image->format, LDF_SIMPLE, bits, ch, 4, false, LCO_SINGLE, false, &err))
		return err;
	if (!SetExtendedInfo(&image->extended, 1, isBinary ? LC_PNM_BINARY : LC_PNM_NONE, LD_UNKNOWN, 0, 0, &err))
		return err;

	SimplifyFormat(&image->format);

	//------------------------
	// 4) LOAD IMAGE DATA
	if (request == LL_INFO)
		return LE_NONE;

	const LilIoDesc *pIo = io;
	LilIoDesc textIo;
	LilIoPnmHandle pnmh;
	if (isBinary) {
		if (IsSpace(currChar))
			Read8(io);
	}
	else {
		currChar = SkipExtra(currChar, io);
		pnmh.bits = bits;
		pnmh.pIo = io;
		pnmh.currChar = currChar;
		textIo.handle = &pnmh;
		textIo.readFunc = LilIoPnmRead;
		pIo = &textIo;
	}

	err = ReadImageData(pIo, image, imageFormat, true, true);

	return err;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType SavePnm(const LilIoDesc *io, const LilImageDesc *image, const LilFormatDesc *imageFormat) {
	//------------------------
	// 1) VARIABLES TO BE WRITTEN
	LilErrorType err = LE_NONE;
	
	LilFormatDesc outFormat = imageFormat ? *imageFormat : image->format;
	LilCompressionType compression = image->extended.compression == LC_ANY ? LC_PNM_NONE : image->extended.compression;
	bool isBinary = image->extended.compression == LC_PNM_BINARY;
	lil_i32 bits = outFormat.bitsPerChannel, ch = outFormat.numChannels;

	if (!CheckValue(compression, GOOD_C, N_GOOD_C, 0, 0, LE_INTERNALERROR, LE_BADCOMPRESSION, &err))
		return err;

	//------------------------
	// 2) GET IMAGE, FORMAT, EXTENDED INFO
	if (isBinary && !MatchFormat(&outFormat, rangesBinary, N_RANGES_BINARY, 0, 0, &err))
		return err;
	if (!isBinary && !MatchFormat(&outFormat, rangesText, N_RANGES_TEXT, 0, 0, &err))
		return err;

	//------------------------
	// 3a) WRITE FILE FORMAT HEADER
	// 3b) WRITE IMAGE INFO
	io->seekFunc(io->handle, 0, LS_SET);
	lil_u8 num = (isBinary ? '4' : '1') + (ch == 3 ? 2 : (bits == 1 ? 0 : 1));
	WriteFle(io, "111", 'P', num, '\n');

	WriteValue(io, image->data.width);
	Write8(io, '\n');
	WriteValue(io, image->data.height);
	Write8(io, '\n');
	if (num != '1' && num != '4') {
		WriteValue(io, (1 << bits) - 1);
		Write8(io, '\n');
	}

	//------------------------
	// 4) SAVE IMAGE DATA
	const LilIoDesc *pIo = io;
	LilIoDesc textIo;
	LilIoPnmHandle pnmh;
	if (!isBinary) {
		pnmh.bits = bits;
		pnmh.pIo = io;
		textIo.handle = &pnmh;
		textIo.writeFunc = LilIoPnmWrite;
		pIo = &textIo;
	}
		
	return WriteImageData(pIo, image, &outFormat, true);
}

}  // namespace lil

#endif // LIL_SUPPORTS_PNM
