#include "LilPlatform.h"

#ifdef LIL_SUPPORTS_PSD

#include <memory>
#include "LilCodecRle.h"
#include "LilConvert.h"
#include "LilUtilFormat.h"
#include "LilMem.h"
#include "LilBitManip.h"
#include "LilIo.h"

namespace lil {

namespace {

//-----------------------------------------------------------------------------------------------------------
// IDs
const lil_u32 PSD_ID = 0x38425053;  // '8BPS'
const lil_u32 PSD_RES = 0x3842494d;  // '8BIM'

const lil_u16 PSD_VER = 1;

// Color Types
const lil_i16 COLOR_BITMAP = 0;
const lil_i16 COLOR_GRAYSCALE = 1;
const lil_i16 COLOR_INDEXED = 2;
const lil_i16 COLOR_RGB = 3;
const lil_i16 COLOR_CMYK = 4;
const lil_i16 COLOR_MULTICHANNEL = 7;
const lil_i16 COLOR_DUOTONE = 8;
const lil_i16 COLOR_LAB = 9;
const lil_i32 GOOD_CLR[] = { COLOR_BITMAP, COLOR_GRAYSCALE, COLOR_INDEXED, COLOR_RGB, COLOR_DUOTONE };
const lil_u8 N_GOOD_CLR = LIL_ARRAY_SIZE(GOOD_CLR);
const lil_i32 BAD_CLR[] = { COLOR_CMYK, COLOR_MULTICHANNEL, COLOR_LAB };
const lil_u8 N_BAD_CLR = LIL_ARRAY_SIZE(BAD_CLR);

// Compressions
const lil_i16 COMPRESSION_NONE = 0;
const lil_i16 COMPRESSION_RLE = 1;
const lil_i16 COMPRESSION_ZIPWO = 2;
const lil_i16 COMPRESSION_ZIPWITH = 3;
const lil_i32 GOOD_COMP[] = { COMPRESSION_NONE, COMPRESSION_RLE };
const lil_u8 N_GOOD_COMP = LIL_ARRAY_SIZE(GOOD_COMP);
const lil_i32 BAD_COMP[] = { COMPRESSION_ZIPWO, COMPRESSION_ZIPWITH };
const lil_u8 N_BAD_COMP = LIL_ARRAY_SIZE(BAD_COMP);

// Valid Formats
const LilFormatRange ranges[] = {
	{ LDF_SIMPLE, { {1}, {8}, {16} }, 3, 1, 1, FLAG_YES, LCO_SINGLE, FLAG_NO },
	{ LDF_SIMPLE, { {1}, {8}, {16} }, 3, 2, 1, FLAG_YES, LCO_YA, FLAG_NO },
	{ LDF_SIMPLE, { {1}, {8}, {16} }, 3, 3, 1, FLAG_YES, LCO_RGB, FLAG_NO },
	{ LDF_SIMPLE, { {1}, {8}, {16} }, 3, 4, 1, FLAG_YES, LCO_RGBA, FLAG_NO },
	//{ LDF_SIMPLE, { {1}, {8}, {16} }, 3, 4, 1, FLAG_YES, LCO_CMYK, FLAG_NO },
	//{ LDF_SIMPLE, { {1}, {8}, {16} }, 3, 5, 1, FLAG_YES, LCO_CMYKA, FLAG_NO },
};
const lil_u8 N_RANGES = LIL_ARRAY_SIZE(ranges);

const LilFormatDesc PALETTE_FORMAT = { LDF_SIMPLE, 8, {}, 3, 4, true, LCO_RGB, false };

//-----------------------------------------------------------------------------------------------------------
bool IsValidId(lil_u32 id, lil_u16 version)
{
	// Specs say do NOT try to read the file if these 2 things don't match
	if (id == PSD_ID && version == PSD_VER)
		return true;
	return false;
}

} // namespace



//-----------------------------------------------------------------------------------------------------------
bool IsExtensionPsd(const char *fileName)
{
	if (CheckExtension(fileName, ".PSD"))
		return true;
	return false;
}

//-----------------------------------------------------------------------------------------------------------
bool IsDetectablePsd(const LilIoDesc *io)
{
	io->seekFunc(io->handle, 0, LS_SET);
	lil_u32 id = Read32be(io);
	lil_u16 version = Read16be(io);
	io->seekFunc(io->handle, 0, LS_SET);
	return IsValidId(id, version);
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType LoadPsd(const LilIoDesc *io, LilImageDesc *image, LilLoadType request,
								 const LilFormatDesc *imageFormat)
{
	// File Header
	// Color Mode Data
	// Image Resources
	// Layer and Mask Information
	// Image Data

	//------------------------
	// 1) VARIABLES TO BE READ
	LilErrorType err = LE_NONE;
	lil_u32 dummy32 = 0, id = 0, width = 0, height = 0, colorDataSize = 0;
	lil_i16 dummy16 = 0, version = 0, numChannels = 0, channelDepth = 0, colorMode = 0;

	//------------------------
	// 2a) READ FILE FORMAT HEADER
	io->seekFunc(io->handle, 0, LS_SET);
	ReadFbe(io, "4242", &id, &version, &dummy32, &dummy16);
	if (!IsValidId(id, version))
		return LE_BADHEADER;

	//------------------------
	// 2b) READ SELECTED IMAGE INFO
	// 2c) READ PALETTE INFO
	ReadFbe(io, "2442 2 4", &numChannels, &height, &width, &channelDepth, &colorMode, &colorDataSize);

	// Indexed = RRR GGG BBB 256 color 8 bits per channel
	// Duotone = Adobe intentially undocumented - preserve this data when reading/writing and
	//           treat image as if just grayscale.
	// Otherwise - colordatasize needs to be 0
	if (colorDataSize != 0) {
		if (colorMode == COLOR_INDEXED) {
			if (colorDataSize != 768)
				return LE_BADHEADER;  // Docs say it will always be 768...
			ReadHandle(&image->memFuncs, io, 768, &image->palette.lut, 0);
			image->palette.numEntries = 256;
			if (!SetFormat(&image->palette.format, LDF_SIMPLE, 8, 3, 4, true, LCO_RGB, false, &err))
				return err;
		}
		else if (colorMode == COLOR_DUOTONE) {
			ReadHandle(&image->memFuncs, io, colorDataSize, &image->extended.data,
				&image->extended.fileFormat.psd.duotoneDataSize);
			image->extended.fileFormat.psd.duotoneDataOffset = 0;
		}
		else
			return LE_BADHEADER;
	}

	// Similar to TIFF directories.
	lil_i32 imageResSize = Read32be(io);
	lil_i32 totalRead = 0;
	for (;;) {
		lil_u32 type = Read32be(io);
		if (type != PSD_RES) {
			io->seekFunc(io->handle, -4, LS_RELATIVE);
			if (totalRead != imageResSize)
				return LE_BADMETADATA;
			break;
		}
		/*lil_u16 id =*/ Read16be(io);  // Currently not using this...
		lil_u8 strSize = Read8(io);
		strSize += ((strSize & 1) ^ 1);  // Padded to even size INCLUDING char coutn
		io->seekFunc(io->handle, strSize, LS_RELATIVE);
		lil_u32 size = Read32be(io);
		size += (size & 1);  // Padded to even size
		io->seekFunc(io->handle, size, LS_RELATIVE);
		totalRead += 4 + 2 + 1 + strSize + 4 + size;
	}

	// Layer AND global mask info.
	lil_i32 layerInfo = Read32be(io);
	io->seekFunc(io->handle, abs(layerInfo), LS_RELATIVE);

	lil_i16 imageCompression = Read16be(io);

	if (!CheckValue(colorMode, GOOD_CLR, N_GOOD_CLR, BAD_CLR, N_BAD_CLR, LE_UNSUPPORTEDCOLORSPACE, LE_BADCOLORSPACE, &err))
		return err;

	LilChannelOrderType lco;
	if (colorMode == COLOR_CMYK) {
		if (numChannels == 4)
			lco = LCO_CMYK;
		else if (numChannels == 5)
			lco = LCO_CMYKA;
		else
			return LE_BADCOLORSPACE;
	}
	else if (colorMode == COLOR_RGB) {
		if (numChannels == 3)
			lco = LCO_RGB;
		else if (numChannels == 4)
			lco = LCO_RGBA;
		else
			return LE_BADCOLORSPACE;
	}
	else {
		if (numChannels == 1)
			lco = LCO_SINGLE;
		else if (numChannels == 2)
			lco = LCO_YA;
		else
			return LE_BADCOLORSPACE;
	}

	//------------------------
	// 3) SET IMAGE, FORMAT, EXTENDED INFO
	// Thankfully, there's not TOO much variation in the format the data comes in!
	if (!SetImageSize(&image->data, width, height, &err))
		return err;
	if (!SetFormat(&image->format, LDF_SIMPLE, static_cast<lil_u8>(channelDepth), static_cast<lil_u8>(numChannels),
		1, true, lco, false, &err))
		return err;
	if (!SetExtendedInfo(&image->extended, 1, imageCompression + 1, LD_INCHES, 0, 0, &err))
		return err;

	if (!CheckValue(imageCompression, GOOD_COMP, N_GOOD_COMP, BAD_COMP, N_BAD_COMP, LE_UNSUPPORTEDCOMPRESSION, LE_BADCOMPRESSION, &err))
		return err;

	imageCompression++;

	//------------------------
	// 4) LOAD IMAGE DATA
	if (request == LL_INFO)
		return LE_NONE;

	const LilIoDesc *pIo = io;

	std::auto_ptr<LilCodecPackbits> codec;
	LilBuffer buf;
	if (imageCompression == LC_PSD_PACKBITS) {
        if (!buf.Realloc(&image->memFuncs, image->data.height * numChannels * 4))
			return LE_NOTENOUGHMEMORY;
		codec = std::auto_ptr<LilCodecPackbits>(new LilCodecPackbits(io, image,
			&image->format, buf.GetPtr<lil_u32>(), true));
		pIo = codec->GetIoPtr();
	}

	RaiiIo invertIo(IoCloseInvert);
	if (colorMode == COLOR_BITMAP) {
		IoOpenInvert(&invertIo, io);
        pIo = &invertIo;
	}

	err = ReadImageData(pIo, image, imageFormat, false, false);

	return err;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType SavePsd(const LilIoDesc *io, const LilImageDesc *image, const LilFormatDesc *imageFormat)
{
	//------------------------
	// 1) VARIABLES TO BE WRITTEN
	LilErrorType err = LE_NONE;

	LilFormatDesc outFormat = imageFormat ? *imageFormat : image->format;

	bool detectComp = image->extended.compression == LC_ANY;
	LilCompressionType compression = detectComp ? LC_PSD_NONE : image->extended.compression;

	lil_u16 ch = outFormat.numChannels, bits = outFormat.bitsPerChannel, colorMode = 0;
	lil_u32 colorDataSize = 0;
	LilPaletteDesc tempPal = { 0 };
	LilBuffer tempPalBuf;

	//------------------------
	// 2) GET IMAGE, FORMAT, EXTENDED INFO
	if (!MatchFormat(&outFormat, ranges, N_RANGES, 0, 0, &err))
		return err;

	if (image->data.width > 65535 || image->data.height > 65535)
		return LE_UNSUPPORTEDSIZE;
	lil_u32 oneChan = (ch == 1 || (ch == 2 && HasAlpha(&outFormat)));
	if (bits * ch == 1)
		colorMode = COLOR_BITMAP;
	else if (oneChan && (image->palette.numEntries == 0 || bits == 16))
		colorMode = COLOR_GRAYSCALE;
	else if (oneChan && image->extended.fileFormat.psd.duotoneDataSize) {
		colorMode = COLOR_DUOTONE;
		colorDataSize = static_cast<lil_u32>(image->extended.fileFormat.psd.duotoneDataSize);
		// @@Need to change WriteHandle to accept an offset, and when writing
		// palette below, use image->extended.data instead of tempPal.lut, IF
		// colorMode == COLOR_DUOTONE.
		//pColorData = static_cast<lil_u8 *>(image->extended.data);
		return LE_UNSUPPORTEDCOLORSPACE;
	}
	else if (oneChan) {
        if (!SetPalette(&tempPal, image, &PALETTE_FORMAT, 256, &tempPalBuf, &err))
            return err;
		colorMode = COLOR_INDEXED;
		colorDataSize = 768;  // Mandatory
	}
	else
		colorMode = COLOR_RGB;

	if (detectComp)
		compression = LC_PSD_PACKBITS;
	if (compression == LC_PSD_PACKBITS && outFormat.bitsPerChannel != 8)
		compression = LC_PSD_NONE;

	//------------------------
	// 3a) WRITE FILE FORMAT HEADER
	// 3b) WRITE IMAGE INFO
	// 3c) WRITE PALETTE INFO

	io->seekFunc(io->handle, 0, LS_SET);
	WriteFbe(io, "4242", PSD_ID, PSD_VER, ZERO_32, ZERO_16);  // File Header

	WriteFbe(io, "2442 2 4", ch, image->data.height, image->data.width, bits, colorMode, colorDataSize);
	WriteN(io, colorDataSize, tempPal.lut);
	Write32be(io, 0);  // Image Resources
	Write32be(io, 0);  // Layer and Mask Information
	Write16be(io, static_cast<lil_u16>(compression - 1));  // Image Data

	//------------------------
	// 4) SAVE IMAGE DATA
	std::auto_ptr<LilCodecPackbits> codec;
	const LilIoDesc *pIo = io;

	RaiiIo invertIo(IoCloseInvert);
	if (colorMode == COLOR_BITMAP) {
		IoOpenInvert(&invertIo, io);
        pIo = &invertIo;
	}

	LilBuffer buf;
	if (compression == LC_PSD_PACKBITS) {
	    if (!buf.Realloc(&image->memFuncs, image->data.height * ch * 4))
			return LE_NOTENOUGHMEMORY;
		codec = std::auto_ptr<LilCodecPackbits>(new LilCodecPackbits(pIo, image, &outFormat, buf.GetPtr<lil_u32>(), true));
		pIo = codec->GetIoPtr();
	}

	return WriteImageData(pIo, image, &outFormat, false);
}

}  // namespace lil

#endif // LIL_SUPPORTS_PSD
