#include "LilPlatform.h"

#ifdef LIL_SUPPORTS_TGA

#include <memory>
#include "LilCodecRle.h"
#include "LilUtilFormat.h"
#include "LilMem.h"
#include "LilConvert.h"
#include "LilBitManip.h"
#include "LilIo.h"

namespace lil {

//-----------------------------------------------------------------------------------------------------------
namespace {

// Colormap flags
const lil_u8 COLORMAP_NO = 0;
const lil_u8 COLORMAP_YES = 1;
const lil_i32 GOOD_M[] = { COLORMAP_NO, COLORMAP_YES };
const lil_u8 N_GOOD_M = LIL_ARRAY_SIZE(GOOD_M);

// Color data formats
const lil_u8 IMAGETYPE_NONE = 0;
const lil_u8 IMAGETYPE_COLORMAP = 1;
const lil_u8 IMAGETYPE_RGB = 2;
const lil_u8 IMAGETYPE_MONO = 3;
const lil_u8 IMAGETYPE_RLECOLORMAP = 9;
const lil_u8 IMAGETYPE_RLERGB = 10;
const lil_u8 IMAGETYPE_RLEMONO = 11;
const lil_u8 IMAGETYPE_ZIPCOLORMAP = 32;
const lil_u8 IMAGETYPE_ZIPPREDCOLORMAP = 33;

const lil_i32 GOOD_I[] = { IMAGETYPE_NONE, IMAGETYPE_COLORMAP, IMAGETYPE_RGB, IMAGETYPE_MONO,
	IMAGETYPE_RLECOLORMAP, IMAGETYPE_RLERGB, IMAGETYPE_RLEMONO };
const lil_u8 N_GOOD_I = LIL_ARRAY_SIZE(GOOD_I);
const lil_i32 BAD_I[] = { IMAGETYPE_ZIPCOLORMAP, IMAGETYPE_ZIPPREDCOLORMAP };
const lil_u8 N_BAD_I = LIL_ARRAY_SIZE(BAD_I);

// Valid formats
const LilFormatRange ranges[] = {
	{ LDF_SIMPLE, { {8} }, 1, 1, 1, FLAG_NO, LCO_SINGLE, FLAG_YESNO },
	{ LDF_SIMPLE, { {8} }, 1, 3, 1, FLAG_NO, LCO_BGR, FLAG_YESNO },
	{ LDF_SIMPLE, { {8} }, 1, 4, 1, FLAG_NO, LCO_BGRA, FLAG_YESNO },
	{ LDF_SIMPLEBITS, { {1,5,5,5} }, 1, 4, 1, FLAG_NO, LCO_ARGB, FLAG_YESNO },
};
const lil_u8 N_RANGES = LIL_ARRAY_SIZE(ranges);



//-----------------------------------------------------------------------------------------------------------
bool IsValidId(lil_u8 colorMapType, lil_u8 imageType) {
	// It's kinda a long shot, but this may be semi-reliable auto-detection.
	if ((colorMapType == 0 || colorMapType == 1)
		&& (imageType == 0 || imageType == 1 || imageType == 2 || imageType == 3
		|| imageType == 9 || imageType == 10 || imageType == 11 ||
		imageType == 32 || imageType == 33))
	{
		return true;
	}
	return false;
}

} // namespace



//-----------------------------------------------------------------------------------------------------------
bool IsExtensionTga(const char *fileName)
{
	if (CheckExtension(fileName, ".TGA"))
		return true;
	return false;
}

//-----------------------------------------------------------------------------------------------------------
bool IsDetectableTga(const LilIoDesc *io) {
	io->seekFunc(io->handle, 1, LS_SET);
	lil_u8 colorMapType = Read8(io);
	lil_u8 imageType = Read8(io);
	return IsValidId(colorMapType, imageType);
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType LoadTga(const LilIoDesc *io, LilImageDesc *image, LilLoadType request, const LilFormatDesc *imageFormat)
{
	//------------------------
	// 1) VARIABLES TO BE READ
	LilErrorType err = LE_NONE;
	lil_u8 descLen = 0, colorMap = 0, imageType = 0;   // What's in this file
	lil_u16 colorStart = 0, colorLen = 0;
	lil_u8 colorBpp = 0, bpp = 0;  // For color map loading
	lil_u16 left = 0, bottom = 0, width = 0, height = 0;
	// misc   bits 7-6 unused   5-4 img origin   3-0 alpha/attr bits per pixel
	// misc   bit 5 is top if set, bottom if reset   4 is right if set, left if reset
	lil_u8 misc = 0;

	//------------------------
	// 2a) READ FILE FORMAT HEADER
	// 2b) READ SELECTED IMAGE INFO
	io->seekFunc(io->handle, 0, LS_SET);
	ReadFle(io, "111 221 2222 11", &descLen, &colorMap, &imageType, &colorStart, &colorLen, &colorBpp,
		&left, &bottom, &width, &height, &bpp, &misc);

	io->seekFunc(io->handle, descLen, LS_RELATIVE);

	// Process "misc"
	bool alignRight = (misc & 16) ? true : false;
	if (alignRight)
		return LE_UNSUPPORTEDORIENTATION;
	bool alignBottom = !(misc & 32);
	lil_u8 attrBits = misc & 15;

	// Compression
	LilCompressionType compression = LC_TGA_NONE;
	if (imageType == IMAGETYPE_NONE) {
		SetExtendedInfo(&image->extended, 1, LC_TGA_NONE, LD_INCHES, 0, 0, &err);
		return LE_NONE;  // Not sure what else to do here?
	}
	else if (imageType == IMAGETYPE_RLECOLORMAP || imageType == IMAGETYPE_RLERGB || imageType == IMAGETYPE_RLEMONO)
		compression = LC_TGA_RLE;
	else if (imageType == IMAGETYPE_ZIPCOLORMAP)
		compression = LC_TGA_ZIP;
	else if (imageType == IMAGETYPE_ZIPPREDCOLORMAP)
		compression = LC_TGA_ZIPPRED;

	//------------------------
	// 2c) READ PALETTE INFO
	if (!CheckValue(colorMap, GOOD_M, N_GOOD_M, 0, 0, LE_NONE, LE_BADHEADER, &err))
		return err;
	if (colorMap == COLORMAP_YES) {
		// Never more than 8 bits per component for the color map in TGA
		int pixBytes = RoundUpToNearest(colorBpp, 8) / 8;
		err = image->memFuncs.memReallocFunc(&image->palette.lut, (colorStart + colorLen) * pixBytes);
		if (err != LE_NONE)
			return err;
		ReadN(io, colorLen * pixBytes, image->palette.lut);
		image->palette.numEntries = colorStart + colorLen;
		if (!SetFormatFromBppTga(&image->palette.format, attrBits, colorBpp, false, &err))
			return err;
	}

	//------------------------
	// 3) SET IMAGE, FORMAT, EXTENDED INFO
	if (!SetImageSize(&image->data, width, height, &err))
		return err;
	if (!SetFormatFromBppTga(&image->format, attrBits, bpp, alignBottom, &err))
		return err;
	if (!SetExtendedInfo(&image->extended, 1, compression, LD_INCHES, 0, 0, &err))
		return err;

	if (!CheckValue(imageType, GOOD_I, N_GOOD_I, BAD_I, N_BAD_I, LE_UNSUPPORTEDCOMPRESSION, LE_BADCOLORSPACE, &err))
		return err;

	SimplifyFormat(&image->format);

	//------------------------
	// 4) LOAD IMAGE DATA
	if (request == LL_INFO)
		return LE_NONE;

	std::auto_ptr<LilCodecRleTga> codec;
	const LilIoDesc *pIo = io;
	if (compression == LC_TGA_RLE) {
		codec = std::auto_ptr<LilCodecRleTga>(new LilCodecRleTga(io, image, &image->format));
		pIo = codec->GetIoPtr();
	}

	err = ReadImageData(pIo, image, imageFormat, true, true);

	return err;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType SaveTga(const LilIoDesc *io, const LilImageDesc *image, const LilFormatDesc *imageFormat)
{
	//------------------------
	// 1) VARIABLES TO BE WRITTEN
	LilErrorType err = LE_NONE;

	LilFormatDesc outFormat = imageFormat ? *imageFormat : image->format;
	bool detectComp = image->extended.compression == LC_ANY;
	LilCompressionType compression = detectComp ? LC_TGA_NONE : image->extended.compression;

	lil_u8 descLen = 0, colorMap = image->palette.numEntries ? 1 : 0;
	lil_u8 imageType = 0;   // What's in this file
	lil_u16 colorStart = 0, colorLen = static_cast<lil_u16>(image->palette.numEntries);
	lil_u8 colorBpp = static_cast<lil_u8>(CalcBpp(&image->palette.format));
	lil_u8 bpp = static_cast<lil_u8>(CalcBpp(&outFormat));
	lil_u16 left = 0, bottom = 0;
	lil_u16 width = static_cast<lil_u16>(image->data.width), height = static_cast<lil_u16>(image->data.height);
	// misc   bits 7-6 unused   5-4 img origin   3-0 alpha/attr bits per pixel
	// misc   bit 5 is top if set, bottom if reset   4 is right if set, left if reset
	lil_u8 misc = static_cast<lil_u8>((outFormat.flipped ? 0 : 32) | GetAlphaBits(&outFormat));

	//------------------------
	// 2) GET IMAGE, FORMAT, EXTENDED INFO
	if (!MatchFormat(&outFormat, ranges, N_RANGES, 0, 0, &err))
		return err;

	if (detectComp && bpp <= 8 && outFormat.numChannels == 1)
		compression = LC_TGA_RLE;

	if (compression == LC_TGA_RLE) {
		if (outFormat.numChannels > 1)
			imageType = IMAGETYPE_RLERGB;
		else if (colorLen)
			imageType = IMAGETYPE_RLECOLORMAP;
		else
			imageType = IMAGETYPE_RLEMONO;
	}
	else {
		if (outFormat.numChannels > 1)
			imageType = IMAGETYPE_RGB;
		else if (colorLen)
			imageType = IMAGETYPE_COLORMAP;
		else
			imageType = IMAGETYPE_MONO;
	}

	// Palette info
	if (colorLen) {
		// Check format is ok, convert if need be, return unsupported if unsupported
		//if (!palUseType)
		//	return LE_UNSUPPORTEDPALETTEFORMAT;
	}

	//------------------------
	// 3a) WRITE FILE FORMAT HEADER
	// 3b) WRITE IMAGE INFO
	// 3c) WRITE PALETTE INFO
	io->seekFunc(io->handle, 0, LS_SET);
	WriteFle(io, "111 221 2222 11", descLen, colorMap, imageType, colorStart, colorLen, colorBpp,
		left, bottom, width, height, bpp, misc);
	// @@ Write desc len if it exists
	WriteN(io, colorLen * (colorBpp / 8), image->palette.lut);

	//------------------------
	// 4) SAVE IMAGE DATA
	std::auto_ptr<LilCodecRleTga> codec;
	const LilIoDesc *pIo = io;
	if (compression == LC_TGA_RLE) {
		codec = std::auto_ptr<LilCodecRleTga>(new LilCodecRleTga(io, image, &outFormat));
		pIo = codec->GetIoPtr();
	}

	return WriteImageData(pIo, image, &outFormat, true);
}

}  // namespace lil

#endif // LIL_SUPPORTS_TGA
