#include "Lil.h"
#include "LilUtilFormat.h"
#include "LilIo.h"
#include "LilMem.h"
#include "LilConvert.h"
#include "LilBitManip.h"

namespace lil {

//-----------------------------------------------------------------------------------------------------------
// SUB IO HANDLE
//-----------------------------------------------------------------------------------------------------------

namespace {
	struct SubIoHandle {
		SubIoHandle(const LilIoDesc *ioInput, lil_u64 start, lil_u64 size) : ioInput(ioInput), start(start), size(size) { }
		const LilIoDesc *ioInput;
		lil_u64 start;
		lil_u64 size;
	};

	//-----------------------------------------------------------------------------------------------------------
	LilErrorType LilIoSubRead(lil_io_handle handle, size_t byteCount, void *outMemPtr) {
		SubIoHandle *sub = reinterpret_cast<SubIoHandle *>(handle);
		return sub->ioInput->readFunc(sub->ioInput->handle, byteCount, outMemPtr);
	}

	//-----------------------------------------------------------------------------------------------------------
	LilErrorType LilIoSubWrite(lil_io_handle handle, size_t byteCount, LIL_CONST void *inMemPtr) {
		SubIoHandle *sub = reinterpret_cast<SubIoHandle *>(handle);
		return sub->ioInput->writeFunc(sub->ioInput->handle, byteCount, inMemPtr);
	}

	//-----------------------------------------------------------------------------------------------------------
	LilErrorType LilIoSubSeek(lil_io_handle handle, lil_i64 offset, LilSeekType mode) {
		SubIoHandle *sub = reinterpret_cast<SubIoHandle *>(handle);
		switch (mode) {
		case LS_SET:
			if (offset > static_cast<lil_i64>(sub->size) || offset < 0)  // Allow to set at EOF
				return LE_IOOUTOFRANGE;
			offset += sub->start;
			break;
		case LS_RELATIVE:
			break;
		case LS_END:
			if (offset > static_cast<lil_i64>(sub->size) || offset < 0)  // Allow to set at EOF
				return LE_IOOUTOFRANGE;
			offset = sub->start + sub->size - offset;
			break;
		}
		return sub->ioInput->seekFunc(sub->ioInput->handle, offset, mode);
	}

	//-----------------------------------------------------------------------------------------------------------
	lil_u64 LilIoSubTell(lil_io_handle handle) {
		SubIoHandle *sub = reinterpret_cast<SubIoHandle *>(handle);
		return sub->ioInput->tellFunc(sub->ioInput->handle) - sub->start;
	}
}  // namespace

//-----------------------------------------------------------------------------------------------------------
bool IoOpenSub(LilIoDesc *pIo, const LilIoDesc *pInput, lil_u64 start, lil_u64 size) {
	pIo->handle = reinterpret_cast<lil_io_handle>(new SubIoHandle(pInput, start, size));
	pIo->readFunc = &LilIoSubRead;
	pIo->writeFunc = &LilIoSubWrite;
	pIo->seekFunc = &LilIoSubSeek;
	pIo->tellFunc = &LilIoSubTell;
	return true;
}

//-----------------------------------------------------------------------------------------------------------
void IoCloseSub(LilIoDesc *pIo) {
	if (pIo->handle) {
		delete reinterpret_cast<SubIoHandle *>(pIo->handle);
		pIo->handle = 0;
		pIo->readFunc = 0;
		pIo->writeFunc = 0;
		pIo->seekFunc = 0;
		pIo->tellFunc = 0;
	}
}


//-----------------------------------------------------------------------------------------------------------
// MEM IO HANDLE
//-----------------------------------------------------------------------------------------------------------

namespace {
	struct MemIoHandle {
		MemIoHandle(lil_u8 *ptr, size_t size) : ptr(ptr), size(size), pos(0) { }
		lil_u8 *ptr;
		size_t size;
		size_t pos;
	};


	//-----------------------------------------------------------------------------------------------------------
	LilErrorType LilIoMemRead(lil_io_handle handle, size_t byteCount, void *outMemPtr) {
		MemIoHandle *mem = reinterpret_cast<MemIoHandle *>(handle);
		if (mem->pos + byteCount > mem->size)
			return LE_IOOUTOFRANGE;
		memcpy(outMemPtr, mem->ptr + mem->pos, byteCount);
		mem->pos += byteCount;
		return LE_NONE;
	}

	//-----------------------------------------------------------------------------------------------------------
	LilErrorType LilIoMemWrite(lil_io_handle handle, size_t byteCount, LIL_CONST void *inMemPtr) {
		MemIoHandle *mem = reinterpret_cast<MemIoHandle *>(handle);
		if (mem->pos + byteCount > mem->size)
			return LE_IOOUTOFRANGE;
		memcpy(mem->ptr + mem->pos, inMemPtr, byteCount);
		mem->pos += byteCount;
		return LE_NONE;
	}

	//-----------------------------------------------------------------------------------------------------------
	LilErrorType LilIoMemSeek(lil_io_handle handle, lil_i64 offset, LilSeekType mode) {
		MemIoHandle *mem = reinterpret_cast<MemIoHandle *>(handle);
		switch (mode) {
		case LS_SET:
			if (static_cast<lil_u64>(offset) > mem->size || offset < 0)  // Allow to set at EOF
				return LE_IOOUTOFRANGE;
			mem->pos = offset;
			break;
		case LS_RELATIVE:
			if (static_cast<lil_u64>(mem->pos + offset) > mem->size)  // Should work for both + and - offsets...
				return LE_IOOUTOFRANGE;
			mem->pos += offset;
			break;
		case LS_END:
			if (static_cast<lil_u64>(offset) > mem->size || offset < 0)  // Allow to set at EOF
				return LE_IOOUTOFRANGE;
			mem->pos = mem->size - offset;
			break;
		}
		return LE_NONE;
	}

	//-----------------------------------------------------------------------------------------------------------
	lil_u64 LilIoMemTell(lil_io_handle handle) {
		MemIoHandle *mem = reinterpret_cast<MemIoHandle *>(handle);
		return mem->pos;
	}
}  // namespace

//-----------------------------------------------------------------------------------------------------------
bool IoOpenMem(LilIoDesc *pIo, lil_u8 *ptr, size_t size) {
	pIo->handle = reinterpret_cast<lil_io_handle>(new MemIoHandle(ptr, size));
	pIo->readFunc = &LilIoMemRead;
	pIo->writeFunc = &LilIoMemWrite;
	pIo->seekFunc = &LilIoMemSeek;
	pIo->tellFunc = &LilIoMemTell;
	return true;
}

//-----------------------------------------------------------------------------------------------------------
void IoCloseMem(LilIoDesc *pIo) {
	if (pIo->handle) {
		delete reinterpret_cast<MemIoHandle *>(pIo->handle);
		pIo->handle = 0;
		pIo->readFunc = 0;
		pIo->writeFunc = 0;
		pIo->seekFunc = 0;
		pIo->tellFunc = 0;
	}
}


//-----------------------------------------------------------------------------------------------------------
// INVERT IO HANDLE
//-----------------------------------------------------------------------------------------------------------

namespace {
	struct InvertIoHandle {
		InvertIoHandle(const LilIoDesc *ioInput) : ioInput(ioInput) { }
		const LilIoDesc *ioInput;
	};

	//-----------------------------------------------------------------------------------------------------------
	LilErrorType LilIoInvertRead(lil_io_handle handle, size_t byteCount, void *outInvertPtr) {
		InvertIoHandle *Invert = reinterpret_cast<InvertIoHandle *>(handle);
		lil_u8 *out8 = reinterpret_cast<lil_u8 *>(outInvertPtr);
		for (size_t i = 0; i < byteCount; i++) {
			lil_u8 data = Read8(Invert->ioInput);
			*out8++ = ~data;
		}
		return LE_NONE;
	}

	//-----------------------------------------------------------------------------------------------------------
	LilErrorType LilIoInvertWrite(lil_io_handle handle, size_t byteCount, LIL_CONST void *inInvertPtr) {
		InvertIoHandle *Invert = reinterpret_cast<InvertIoHandle *>(handle);
		LIL_CONST lil_u8 *in8 = reinterpret_cast<LIL_CONST lil_u8 *>(inInvertPtr);
		for (size_t i = 0; i < byteCount; i++) {
			Write8(Invert->ioInput, ~(*in8));
			in8++;
		}
		return LE_NONE;
	}

	//-----------------------------------------------------------------------------------------------------------
	LilErrorType LilIoInvertSeek(lil_io_handle handle, lil_i64 offset, LilSeekType mode) {
		InvertIoHandle *Invert = reinterpret_cast<InvertIoHandle *>(handle);
		return Invert->ioInput->seekFunc(Invert->ioInput->handle, offset, mode);
	}

	//-----------------------------------------------------------------------------------------------------------
	lil_u64 LilIoInvertTell(lil_io_handle handle) {
		InvertIoHandle *Invert = reinterpret_cast<InvertIoHandle *>(handle);
		return Invert->ioInput->tellFunc(Invert->ioInput->handle);
	}
}  // namespace

//-----------------------------------------------------------------------------------------------------------
bool IoOpenInvert(LilIoDesc *pIo, const LilIoDesc *pInput) {
	pIo->handle = reinterpret_cast<lil_io_handle>(new InvertIoHandle(pInput));
	pIo->readFunc = &LilIoInvertRead;
	pIo->writeFunc = &LilIoInvertWrite;
	pIo->seekFunc = &LilIoInvertSeek;
	pIo->tellFunc = &LilIoInvertTell;
	return true;
}

//-----------------------------------------------------------------------------------------------------------
void IoCloseInvert(LilIoDesc *pIo) {
	if (pIo->handle) {
		delete reinterpret_cast<InvertIoHandle *>(pIo->handle);
		pIo->handle = 0;
		pIo->readFunc = 0;
		pIo->writeFunc = 0;
		pIo->seekFunc = 0;
		pIo->tellFunc = 0;
	}
}


//-----------------------------------------------------------------------------------------------------------
// IO UTILITY FUNCTIONS
//-----------------------------------------------------------------------------------------------------------
void ReadFVle(const LilIoDesc *io, const char *format, va_list v) {
	while (*format) {
		switch (*format++) {
			case ' ': break;
			case '1': { lil_u8 *x = va_arg(v, lil_u8 *); *x = Read8(io); break; }
			case '2': { lil_i16 *x = va_arg(v, lil_i16 *); *x = Read16le(io); break; }
			case '4': { lil_i32 *x = va_arg(v, lil_i32 *); *x = Read32le(io); break; }
			default:
				LIL_ASSERT(0);
				va_end(v);
				return;
		}
	}
}

//-----------------------------------------------------------------------------------------------------------
void ReadFle(const LilIoDesc *io, const char *format, ...) {
   va_list v;
   va_start(v, format);
   ReadFVle(io, format, v);
   va_end(v);
}

//-----------------------------------------------------------------------------------------------------------
void ReadFVbe(const LilIoDesc *io, const char *format, va_list v) {
	while (*format) {
		switch (*format++) {
			case ' ': break;
			case '1': { lil_u8 *x = va_arg(v, lil_u8 *); *x = Read8(io); break; }
			case '2': { lil_i16 *x = va_arg(v, lil_i16 *); *x = Read16be(io); break; }
			case '4': { lil_i32 *x = va_arg(v, lil_i32 *); *x = Read32be(io); break; }
			default:
				LIL_ASSERT(0);
				va_end(v);
				return;
		}
	}
}

//-----------------------------------------------------------------------------------------------------------
void ReadFbe(const LilIoDesc *io, const char *format, ...) {
   va_list v;
   va_start(v, format);
   ReadFVbe(io, format, v);
   va_end(v);
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType WriteFVle(const LilIoDesc *io, const char *format, va_list v) {
	LilErrorType err = LE_NONE;
	while (*format) {
		switch (*format++) {
			case ' ': break;
			case '1': { lil_u8 x = static_cast<lil_u8>(va_arg(v, lil_u32)); err = Write8(io, x); break; }
			case '2': { lil_i16 x = static_cast<lil_i16>(va_arg(v, lil_i32)); err = Write16le(io, x); break; }
			case '4': { lil_i32 x = va_arg(v, lil_i32); err = Write32le(io, x); break; }
			default:
				LIL_ASSERT(0);
				va_end(v);
				return LE_INTERNALERROR;
		}
	}
	return err;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType WriteFle(const LilIoDesc *io, const char *format, ...) {
   va_list v;
   va_start(v, format);
   LilErrorType err = WriteFVle(io, format, v);
   va_end(v);
   return err;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType WriteFVbe(const LilIoDesc *io, const char *format, va_list v) {
	LilErrorType err = LE_NONE;
	while (*format) {
		switch (*format++) {
			case ' ': break;
			case '1': { lil_u8 x = static_cast<lil_u8>(va_arg(v, lil_u32)); err = Write8(io, x); break; }
			case '2': { lil_i16 x = static_cast<lil_i16>(va_arg(v, lil_i32)); err = Write16be(io, x); break; }
			case '4': { lil_i32 x = va_arg(v, lil_i32); err = Write32be(io, x); break; }
			default:
				LIL_ASSERT(0);
				va_end(v);
				return LE_INTERNALERROR;
		}
	}
	return err;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType WriteFbe(const LilIoDesc *io, const char *format, ...) {
   va_list v;
   va_start(v, format);
   LilErrorType err = WriteFVbe(io, format, v);
   va_end(v);
   return err;
}

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE int ToUpper(int c) { return (c - 32 * (c >= 'a' && c <= 'z')); }

// Indifferent to '.' and indifferent to cAsE.
bool CheckExtension(const char *fileName, const char *ext) {
	int needed = 0;
	if (ext[0] == '.')
		ext++;
	while (ext[needed])
		needed++;
	// Double while-loop ensures filenames with multiple .'s still work correctly
	while (*fileName) {
		while (*fileName && *fileName != '.')
			fileName++;
		int match = 0;
		while (*fileName) {
			fileName++;
			if (ToUpper(*fileName) == ToUpper(ext[match])) {
				match++;
				if (match == needed)
					return true;
			}
			else break;
		}
	}
	return false;
}

//-----------------------------------------------------------------------------------------------------------
lil_u64 CalcReadImageData(LilImageDesc *image, const LilFormatDesc *imageFormat, bool forceAlpha) {
	// NOTE if any allocations change, it must match ReadImageData
	lil_u64 size = 0;
	bool formatDifferent = IsFormatDifferent(&image->format, imageFormat);
	if (formatDifferent || (forceAlpha && HasAlpha(imageFormat ? imageFormat : &image->format)))
	{
		size += CalcByteSize(&image->format, image->data.width, image->data.height);
		if (formatDifferent)
			size += CalcByteSize(imageFormat, image->data.width, image->data.height);
	}

	// 1 MB, unless image is smaller, or if a single scan line is bigger, size of a scanline
	const lil_u32 DEF_SIZE = 1024 * 1024;
	size += GetMin<size_t>(LilImageGetByteSize(image), GetMax<size_t>(DEF_SIZE, LilImageGetBpl(image)));
	return size;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType ReadImageData(const LilIoDesc *io, LilImageDesc *image, const LilFormatDesc *imageFormat,
						   bool isLittleEndian, bool forceAlpha)
{
	// NOTE if any allocations change, it must match CalcReadImageData
	LilErrorType err = LE_NONE;
	if (!ImageAlloc(image, imageFormat, &err))
		return err;

	lil_u64 inTotalBytes = CalcByteSize(&image->format, image->data.width, image->data.height);
	lil_u64 outLineBytes = CalcBpl(imageFormat ? imageFormat : &image->format, image->data.width);
	lil_u64 outTotalBytes = CalcByteSize(imageFormat ? imageFormat : &image->format, image->data.width, image->data.height);

	const LilIoDesc *pIo = io;
	LilBuffer buffNew;
	LilBuffer buffOld;

	bool formatDifferent = IsFormatDifferent(&image->format, imageFormat);

	if (formatDifferent || (forceAlpha && HasAlpha(imageFormat ? imageFormat : &image->format))) {
		LilBuffer *initialBuf = formatDifferent ? &buffOld : &buffNew;

		if (!initialBuf->Realloc(&image->memFuncs, inTotalBytes))
			return LE_NOMEMBUFFER;
		if (formatDifferent && !buffNew.Realloc(&image->memFuncs, outTotalBytes))
			return LE_NOMEMBUFFER;

		// Decoding streaming tests
		//for (lil_u32 i = 0; i < inTotalBytes; i += 4)
		//	buffOld.GetPtr<lil_u32>()[i >> 2] = Read32le(io);
		//for (lil_u32 i = 0; i < inTotalBytes; i++)
		//	buffOld.GetPtr<lil_u8>()[i] = Read8(io);
		err = ReadN(io, inTotalBytes, initialBuf->GetPtr<lil_u8>());
		if (err != LE_NONE)
			return err;

		// Convert data needs data in big endian
		if (isLittleEndian)
			MakeBigEndian(initialBuf->GetPtr<void>(), image->data.width, image->data.height, &image->format);

		if (forceAlpha) {
			err = FillAlpha(initialBuf->GetPtr<lil_u8>(), image);
			if (err != LE_NONE)
				return err;
		}

		if (formatDifferent) {
			err = ConvertData(&image->memFuncs, buffOld.GetPtr<lil_u8>(), &image->format, &image->palette,
				buffNew.GetPtr<lil_u8>(), imageFormat, &image->palette,
				image->data.width, image->data.height, isLittleEndian);
			if (err != LE_NONE)
				return err;

			// If all went well, then the format SHOULD be right.
			image->format = *imageFormat;
		}

		// It came out big endian... this will flip it back if that's an issue...
		// NOTE image->format will be the correct new format at this point...
#ifdef LIL_ENDIAN_LITTLE
		MakeBigEndian(buffNew.GetPtr<void>(), image->data.width, image->data.height, &image->format);
#endif

		pIo = buffNew.GetIoDescPtr();
	}

	// 1 MB, unless image is smaller, or if a single scan line is bigger, size of a scanline
	const lil_u32 DEF_SIZE = 1024 * 1024;
	LilBuffer scanBuf;
	if (!scanBuf.Realloc(&image->memFuncs, GetMin<size_t>(LilImageGetByteSize(image),
		GetMax<size_t>(DEF_SIZE, LilImageGetBpl(image)))))
	{
		return LE_NOTENOUGHMEMORY;
	}
	lil_u8 *pData = scanBuf.GetPtr<lil_u8>();

	// Not ready to handle this yet!!!
	lil_u32 planes = image->format.planar ? image->format.numChannels : 1;
	for (lil_u32 iPlane = 0; iPlane < planes; iPlane++) {
		lil_u32 scanChunk = scanBuf.GetSize() / outLineBytes;
		lil_u32 scanChunkExtra = image->data.height % scanChunk;
		lil_u64 scanBytes = scanChunk * outLineBytes;
		lil_u32 numChunks = image->data.height / scanChunk;
		for (lil_u32 iScan = 0; iScan < numChunks; iScan++) {
			err = ReadN(pIo, scanBytes, pData);
			if (err != LE_NONE)
				return err;
#ifdef LIL_LSB_TO_MSB
			if (LilImageGetBpp(image) == 1)
				BitReverseBuf8(pData, scanBytes, 1, scanBytes);
#endif
			err = image->memFuncs.imageWriteFunc(image->data.data, pData, 0, scanChunk * iScan,
				image->data.width, scanChunk, iPlane, outLineBytes);
			if (err != LE_NONE)
				return err;
		}
		if (scanChunkExtra) {
			err = ReadN(pIo, scanChunkExtra * outLineBytes, pData);
			if (err != LE_NONE)
				return err;
#ifdef LIL_LSB_TO_MSB
			if (LilImageGetBpp(image) == 1)
				BitReverseBuf8(pData, scanBytes, 1, scanBytes);
#endif
			err = image->memFuncs.imageWriteFunc(image->data.data, pData, 0, scanChunk * numChunks,
				image->data.width, scanChunkExtra, iPlane, outLineBytes);
			if (err != LE_NONE)
				return err;
		}
	}

	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType WriteImageData(const LilIoDesc *io, const LilImageDesc *image, const LilFormatDesc *imageFormat,
							bool isLittleEndian)
{
	LilErrorType err = LE_NONE;

	lil_u64 inLineBytes = CalcBpl(&image->format, image->data.width);
	lil_u64 inTotalBytes = CalcByteSize(&image->format, image->data.width, image->data.height);
	//lil_u32 outLineBytes = CalcBpl(imageFormat ? imageFormat : &image->format, image->data.width);
	lil_u64 outTotalBytes = CalcByteSize(imageFormat ? imageFormat : &image->format, image->data.width, image->data.height);

	const LilIoDesc *pIo = io;
	LilBuffer buffNew;
	LilBuffer buffOld;

	const lil_u32 DEF_SIZE = 1024 * 1024;
	LilBuffer scanBuf;
	if (!scanBuf.Realloc(&image->memFuncs, GetMin<lil_u64>(LilImageGetByteSize(image),
		GetMax<lil_u64>(DEF_SIZE, LilImageGetBpl(image)))))
	{
		return LE_NOTENOUGHMEMORY;
	}
	lil_u8 *pData = scanBuf.GetPtr<lil_u8>();

	// --- STEP 1: If formats are different, redirect writing to a to-be-converted buffer
	if (IsFormatDifferent(&image->format, imageFormat)) {
		if (!buffOld.Realloc(&image->memFuncs, inTotalBytes) || !buffNew.Realloc(&image->memFuncs, outTotalBytes))
			return LE_NOMEMBUFFER;
		pIo = buffOld.GetIoDescPtr();
	}

	// --- STEP 2: Write data in the current format
	// Not ready to handle this yet!!!
	lil_u32 planes = image->format.planar ? image->format.numChannels : 1;
	for (lil_u32 iPlane = 0; iPlane < planes; iPlane++) {
		lil_u32 scanChunk = scanBuf.GetSize() / inLineBytes;
		lil_u32 scanChunkExtra = image->data.height % scanChunk;
		lil_u64 scanBytes = scanChunk * inLineBytes;
		lil_u32 numChunks = image->data.height / scanChunk;
		for (lil_u32 iScan = 0; iScan < numChunks; iScan++) {
			err = image->memFuncs.imageReadFunc(image->data.data, pData, 0, scanChunk * iScan,
				image->data.width, scanChunk, iPlane, inLineBytes);
			if (err != LE_NONE)
				return err;
#ifdef LIL_LSB_TO_MSB
			if (LilImageGetBpp(image) == 1)
				BitReverseBuf8(pData, scanBytes, 1, scanBytes);
#endif
			err = WriteN(pIo, scanBytes, pData);
			if (err != LE_NONE)
				return err;
		}
		if (scanChunkExtra) {
			err = image->memFuncs.imageReadFunc(image->data.data, pData, 0, scanChunk * numChunks,
				image->data.width, scanChunkExtra, iPlane, inLineBytes);
			if (err != LE_NONE)
				return err;
#ifdef LIL_LSB_TO_MSB
			if (LilImageGetBpp(image) == 1)
				BitReverseBuf8(pData, scanBytes, 1, scanBytes);
#endif
			// If formats were NOT different, this is where we encode, otherwise this
			// is just getting the raw data.
			err = WriteN(pIo, scanChunkExtra * inLineBytes, pData);
			if (err != LE_NONE)
				return err;
		}
	}

	// --- STEP 3: If formats are different, convert to proper format...
	if (IsFormatDifferent(&image->format, imageFormat)) {
		// Convert data needs data in big endian
#ifdef LIL_ENDIAN_LITTLE
		MakeBigEndian(buffOld.GetPtr<void>(), image->data.width, image->data.height, &image->format);
#endif

		err = ConvertData(&image->memFuncs, buffOld.GetPtr<lil_u8>(), &image->format, &image->palette,
			buffNew.GetPtr<lil_u8>(), imageFormat, 0,
			image->data.width, image->data.height, isLittleEndian);
		if (err != LE_NONE)
			return err;

		// It came out big endian... need to see if file format requires little endian
		if (isLittleEndian)
			MakeBigEndian(buffNew.GetPtr<void>(), image->data.width, image->data.height, imageFormat);

		// Encoding streaming tests
		//for (lil_u32 i = 0; i < outTotalBytes; i += 4)
		//	buffNew.GetPtr<lil_u32>()[i >> 2] = Write32le(io);
		//for (lil_u32 i = 0; i < outTotalBytes; i++)
		//	buffNew.GetPtr<lil_u8>()[i] = Write8(io);

		// If we get here, than THIS is where we encode.
		err = WriteN(io, outTotalBytes, buffNew.GetPtr<lil_u8>());
		if (err != LE_NONE)
			return err;
	}

	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
void ReadHandle(const LilMemDesc *pMemFuncs, const LilIoDesc *io, lil_u64 size, lil_mem_handle *outHandle, lil_u64 *outSize) {
	if (outSize)
		*outSize = size;
	pMemFuncs->memReallocFunc(outHandle, size);
	ReadN(io, size, *outHandle);
}

//-----------------------------------------------------------------------------------------------------------
bool ReadPaletteWin(const LilIoDesc *io, LilImageDesc *image, lil_u32 numColors, lil_u8 numChannels, LilErrorType *pErr) {
	if (pErr)
		*pErr = LE_NONE;
	image->palette.numEntries = numColors;
	lil_u32 paletteBytes = numColors * numChannels;
	paletteBytes = RoundUpToNearest(paletteBytes, image->format.blockLineByteAlign);
	if (paletteBytes == 0)
		return true;

	// Read the palette in its BGR form
	ReadHandle(&image->memFuncs, io, paletteBytes, &image->palette.lut, 0);

	return SetFormat(&image->palette.format, LDF_SIMPLE, 8, numChannels, image->format.blockLineByteAlign,
		false, numChannels == 3 ? LCO_BGR : LCO_BGRA, false, pErr);
}

//-----------------------------------------------------------------------------------------------------------
lil_u64 GetIoSize(const LilIoDesc *io) {
	lil_u64 pos = io->tellFunc(io->handle);
	io->seekFunc(io->handle, 0, LS_END);
	lil_u64 size = io->tellFunc(io->handle);
	io->seekFunc(io->handle, pos, LS_SET);
	return size;
}

}  // namespace lil
