#ifndef LILIO_H_INCLUDED
#define LILIO_H_INCLUDED

#include <stdarg.h>
#include <string.h>
#include "Lil.h"
#include "LilPlatform.h"

namespace lil {

class BitReadLsb;
class BitWriteLsb;
class BitReadMsb;

bool CheckExtension(const char *fileName, const char *ext);
LIL_INLINE bool IsSpace(char c) { return c == ' ' || c == '\t' || c == '\n' || c == '\r'; }
LIL_INLINE bool IsNum(char c) { return c >= '0' && c <= '9'; }
LilErrorType ReadImageData(const LilIoDesc *io, LilImageDesc *image, const LilFormatDesc *imageFormat,
						   bool isLittleEndian, bool forceAlpha);
LilErrorType WriteImageData(const LilIoDesc *io, const LilImageDesc *image, const LilFormatDesc *imageFormat,
							bool isLittleEndian);
bool ReadPaletteWin(const LilIoDesc *io, LilImageDesc *image, lil_u32 numColors, lil_u8 numChannels, LilErrorType *pErr);
LIL_INLINE lil_u8 Read8(const LilIoDesc *io);
LIL_INLINE lil_u16 Read16be(const LilIoDesc *io);
LIL_INLINE lil_u16 Read16le(const LilIoDesc *io);
LIL_INLINE lil_u32 Read32be(const LilIoDesc *io);
LIL_INLINE lil_u32 Read32le(const LilIoDesc *io);
LIL_INLINE LilErrorType ReadN(const LilIoDesc *io, size_t size, void *pOut);
void ReadFVle(const LilIoDesc *io, const char *format, va_list v);
void ReadFle(const LilIoDesc *io, const char *format, ...);
void ReadFVbe(const LilIoDesc *io, const char *format, va_list v);
void ReadFbe(const LilIoDesc *io, const char *format, ...);
void ReadHandle(const LilMemDesc *pMemFuncs, const LilIoDesc *io, lil_u64 size, lil_mem_handle *outHandle, lil_u64 *outSize);
LIL_INLINE LilErrorType Write8(const LilIoDesc *io, lil_u8 val);
LIL_INLINE LilErrorType Write16be(const LilIoDesc *io, lil_u16 val);
LIL_INLINE LilErrorType Write16le(const LilIoDesc *io, lil_u16 val);
LIL_INLINE LilErrorType Write32be(const LilIoDesc *io, lil_u32 val);
LIL_INLINE LilErrorType Write32le(const LilIoDesc *io, lil_u32 val);
LIL_INLINE LilErrorType WriteN(const LilIoDesc *io, size_t size, const void *pIn);
LilErrorType WriteFVle(const LilIoDesc *io, const char *format, va_list v);
LilErrorType WriteFle(const LilIoDesc *io, const char *format, ...);
LilErrorType WriteFVbe(const LilIoDesc *io, const char *format, va_list v);
LilErrorType WriteFbe(const LilIoDesc *io, const char *format, ...);
lil_u64 GetIoSize(const LilIoDesc *io);

//-----------------------------------------------------------------------------------------------------------
class BitReadLsb {
public:
	LIL_INLINE BitReadLsb(const LilIoDesc *io);
	LIL_INLINE void AlignToByte();
	LIL_INLINE void ReadBits(lil_u32 &output, lil_u8 numBits);
	LIL_INLINE void UndoBits(lil_u32 input, lil_u8 numBits);

private:
	const LilIoDesc *io;
	lil_u8 hasBits;
	lil_u32 currBits;
};

//-----------------------------------------------------------------------------------------------------------
class BitWriteLsb {
public:
	LIL_INLINE BitWriteLsb(const LilIoDesc *io);
	LIL_INLINE ~BitWriteLsb();
	LIL_INLINE void FlushByte();
	LIL_INLINE void WriteBits(lil_u32 input, lil_u8 numBits);

private:
	const LilIoDesc *io;
	lil_u8 hasBits;
	lil_u32 currBits;
};

//-----------------------------------------------------------------------------------------------------------
class BitReadMsb {
public:
	LIL_INLINE BitReadMsb(const LilIoDesc *io);
	LIL_INLINE void AlignToByte();
	LIL_INLINE void ReadBits(lil_u32 &output, lil_u8 numBits);
	LIL_INLINE void UndoBits(lil_u32 input, lil_u8 numBits);

private:
	const LilIoDesc *io;
	lil_u8 hasBits;
	lil_u32 currBits;
};

//-----------------------------------------------------------------------------------------------------------
struct RaiiIo : public LilIoDesc {
	typedef void (* CloseFunc)(LilIoDesc *);
	RaiiIo() : closeFunc(0) { handle = 0; tellFunc = 0; readFunc = 0; writeFunc = 0; seekFunc = 0; }
	RaiiIo(CloseFunc closeFunc) : closeFunc(closeFunc) { handle = 0; tellFunc = 0; readFunc = 0; writeFunc = 0; seekFunc = 0; }
	~RaiiIo() { if (closeFunc) closeFunc(this); }
	CloseFunc closeFunc;
};

//-----------------------------------------------------------------------------------------------------------
bool IoOpenSub(LilIoDesc *pIo, const LilIoDesc *pInput, lil_u64 start, lil_u64 size);
void IoCloseSub(LilIoDesc *pIo);
bool IoOpenMem(LilIoDesc *pIo, lil_u8 *ptr, size_t size);
void IoCloseMem(LilIoDesc *pIo);
bool IoOpenInvert(LilIoDesc *pIo, const LilIoDesc *pInput);
void IoCloseInvert(LilIoDesc *pIo);


//-----------------------------------------------------------------------------------------------------------
// DEFINITIONS
//-----------------------------------------------------------------------------------------------------------
// Inspired by stb_image!  Awesome stuff for file formats we know.
LIL_INLINE lil_u8 Read8(const LilIoDesc *io) {
	lil_u8 ret;
	io->readFunc(io->handle, 1, &ret);
	return ret;
}

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE lil_u16 Read16be(const LilIoDesc *io) {
	lil_u16 high = Read8(io) << 8;
	return high | Read8(io);
}

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE lil_u16 Read16le(const LilIoDesc *io) {
	lil_u16 low = Read8(io);
	return low | (Read8(io) << 8);
}

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE lil_u32 Read32be(const LilIoDesc *io) {
	lil_u32 high = Read16be(io) << 16;  // Don't want 16 bit value to be sign-extended
	return high | Read16be(io);
}

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE lil_u32 Read32le(const LilIoDesc *io) {
	lil_u32 low = Read16le(io);  // Don't want 16 bit value to be sign-extended
	return low | (Read16le(io) << 16);
}

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE LilErrorType ReadN(const LilIoDesc *io, size_t size, void *pOut) {
	return io->readFunc(io->handle, size, pOut);
}

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE LilErrorType Write8(const LilIoDesc *io, lil_u8 val) {
	return io->writeFunc(io->handle, 1, &val);
}

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE LilErrorType Write16be(const LilIoDesc *io, lil_u16 val) {
	lil_u32 err = Write8(io, val >> 8);
	err |= Write8(io, val & 0xff);
	return static_cast<LilErrorType>(err);
}

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE LilErrorType Write16le(const LilIoDesc *io, lil_u16 val) {
	lil_u32 err = Write8(io, val & 0xff);
	err |= Write8(io, val >> 8);
	return static_cast<LilErrorType>(err);
}

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE LilErrorType Write32be(const LilIoDesc *io, lil_u32 val) {
	lil_u32 err = Write16be(io, val >> 16);
	err |= Write16be(io, val & 0xffff);
	return static_cast<LilErrorType>(err);
}

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE LilErrorType Write32le(const LilIoDesc *io, lil_u32 val) {
	lil_u32 err = Write16le(io, val & 0xffff);
	err |= Write16le(io, val >> 16);
	return static_cast<LilErrorType>(err);
}

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE LilErrorType WriteN(const LilIoDesc *io, size_t size, const void *pIn) {
	if (size == 0)
		return LE_NONE;
	return io->writeFunc(io->handle, size, pIn);
}



//-----------------------------------------------------------------------------------------------------------
// BITREADLSB
//-----------------------------------------------------------------------------------------------------------
LIL_INLINE BitReadLsb::BitReadLsb(const LilIoDesc *io) : io(io), hasBits(0), currBits(0) { }

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE void BitReadLsb::AlignToByte() {
	hasBits = 0;  // effectively goes to the next byte in logic
	currBits = 0; // reader algorithm needs this to be correct to avoid extra bit logic
}

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE void BitReadLsb::ReadBits(lil_u32 &output, lil_u8 numBits) {
	while (hasBits < numBits) {
		currBits |= Read8(io) << hasBits;
		hasBits += 8;
	}
	output = currBits & ((1 << numBits) - 1);
	currBits >>= numBits;
	hasBits -= numBits;
}

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE void BitReadLsb::UndoBits(lil_u32 input, lil_u8 numBits) {
	currBits <<= numBits;
	currBits |= input & ((1 << numBits) - 1);
	hasBits += numBits;
}



//-----------------------------------------------------------------------------------------------------------
// BITWRITELSB
//-----------------------------------------------------------------------------------------------------------
LIL_INLINE BitWriteLsb::BitWriteLsb(const LilIoDesc *io) : io(io), hasBits(0), currBits(0) { }

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE BitWriteLsb::~BitWriteLsb() {
    FlushByte();
}

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE void BitWriteLsb::FlushByte() {
    if (hasBits) {
        Write8(io, static_cast<lil_u8>(currBits));
        hasBits = 0;
        currBits = 0;
    }
}

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE void BitWriteLsb::WriteBits(lil_u32 input, lil_u8 numBits) {
	currBits |= input << hasBits;
	hasBits += numBits;
	while (hasBits >= 8) {
        Write8(io, currBits & 0xff);
        currBits >>= 8;
        hasBits -= 8;
	}
}



//-----------------------------------------------------------------------------------------------------------
// BITREADMSB
//-----------------------------------------------------------------------------------------------------------
LIL_INLINE BitReadMsb::BitReadMsb(const LilIoDesc *io) : io(io), hasBits(0), currBits(0) { }

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE void BitReadMsb::AlignToByte() {
	// Keep the lowest multiple of 8 bits, since they are the most recent whole bytes
	hasBits -= hasBits & 7;  // effectively goes to the next byte in logic
	currBits &= (1 << hasBits) - 1;
}

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE void BitReadMsb::ReadBits(lil_u32 &output, lil_u8 numBits) {
	while (hasBits < numBits) {
		currBits = (currBits << 8) | Read8(io);
		hasBits += 8;
	}
	hasBits -= numBits;
	output = (currBits >> hasBits);
	currBits &= ((1 << hasBits) - 1);
}

//-----------------------------------------------------------------------------------------------------------
// Note that input is considered to be LSB-MSB aligned, even though this reader is
// for MSB-LSB aligned input.
LIL_INLINE void BitReadMsb::UndoBits(lil_u32 input, lil_u8 numBits) {
	currBits |= (input & ((1 << numBits) - 1)) << hasBits;
	hasBits += numBits;
}

}  // namespace lil

#endif // LILIO_H_INCLUDED
