#include <memory>
#include <stdlib.h>
#include "Lil.h"
#include "LilPlatform.h"
#include "LilMem.h"
#include "LilUtilFormat.h"

namespace lil {

//-----------------------------------------------------------------------------------------------------------
LilConfigDesc globalConfig;



//-----------------------------------------------------------------------------------------------------------
// MEM UTILITY FUNCTIONS
//-----------------------------------------------------------------------------------------------------------
LilErrorType MemPoolCreate(LilMemDesc *memFuncs, lil_u32 size) {
	// If 2 Gb or more for 32 bit, or really really big for 64 bit, fail.
	if (size > (1ull << ((sizeof(size_t) * 8) - 1)) - 1)
		return LE_NOTENOUGHMEMORY;

	// Format in memory
	// 16 bytes - 0x19842012 0xssssssss 0xoooooooo 0xaaaaaaaa
	// 0x19842012 = marker to id this as ours
	// ss = size of buffer
	// oo = offset for next available byte
	// aa = allocated size (may be more than just size)
	lil_u32 *pBuf = reinterpret_cast<lil_u32 *>(memFuncs->mem);
	if (pBuf) {
		if (pBuf[0] == 0x19842012) {
			// Already exists... let's see if we can use it as is
			if (pBuf[3] >= size) {
				pBuf[1] = size;
				pBuf[2] = 0;
				return LE_NONE;
			}
		}
		else
			return LE_NEEDINITIMAGE;
	}
	// Not enough mem allocated to use it
	LilErrorType err;
	if (!SetError(memFuncs->memReallocFunc(&memFuncs->mem, size + 16), &err)) {
		memFuncs->mem = 0;
		return err;
	}
	pBuf = reinterpret_cast<lil_u32 *>(memFuncs->mem);
	pBuf[0] = 0x19842012;
	pBuf[1] = size;
	pBuf[2] = 0;
	pBuf[3] = size;
	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType MemPoolFree(LilMemDesc *memFuncs) {
	lil_u32 *pBuf = reinterpret_cast<lil_u32 *>(memFuncs->mem);
	if (pBuf) {
		if (pBuf[0] != 0x19842012)
			return LE_NEEDINITIMAGE;
		memFuncs->memReallocFunc(&memFuncs->mem, 0);
		memFuncs->mem = 0;
	}
	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
void *MemAlloc(const LilMemDesc *memFuncs, lil_u32 size) {
	lil_u32 *pBuf = reinterpret_cast<lil_u32 *>(memFuncs->mem);
	LIL_ASSERT(pBuf[0] == 0x19842012);
	lil_u32 offset = pBuf[2];
	if (offset + size > pBuf[1])
		return 0;
	pBuf[2] += size;
	lil_u8 *pBuf8 = reinterpret_cast<lil_u8 *>(memFuncs->mem);
	return pBuf8 + 16 + offset;
}

//-----------------------------------------------------------------------------------------------------------
lil_u32 MemGetSize(const LilMemDesc *memFuncs) {
	lil_u32 *pBuf = reinterpret_cast<lil_u32 *>(memFuncs->mem);
	LIL_ASSERT(pBuf[0] == 0x19842012);
	return pBuf[1];
}

//-----------------------------------------------------------------------------------------------------------
bool ImageAlloc(LilImageDesc *image, const LilFormatDesc *imageFormat, LilErrorType *pErr) {
	// Free what's there first... don't trust Image library... it might not free on create/allocation!
	LilErrorType err = LE_NONE;
	err = image->memFuncs.imageFreeFunc(image->data.data);
	if (err != LE_NONE)
		return SetError(err, pErr);

	// Now allocate the appropriate image...
	LilFormatDesc fDesc = image->format;
	if (imageFormat)
		image->format = *imageFormat;
	err = image->memFuncs.imageAllocFunc(image->data.data, image);
	if (imageFormat)
		image->format = fDesc;
	return SetError(err, pErr);
}

//-----------------------------------------------------------------------------------------------------------
lil_u32 RoundUpToNearest(lil_u32 input, lil_u32 multiple)
{
	if (multiple <= 1)
		return input;
	lil_u32 remainder = input % multiple;
	if (remainder)
		return input + (multiple - remainder);
	return input;
}

//-----------------------------------------------------------------------------------------------------------
lil_u32 RoundDownToNearest(lil_u32 input, lil_u32 multiple)
{
	if (multiple <= 1)
		return input;
	lil_u32 remainder = input % multiple;
	return input - remainder;
}

//-----------------------------------------------------------------------------------------------------------
lil_u32 CalcRemainder(lil_u32 input, lil_u32 multiple) {
	if (input == 0)
        return multiple;
    return RoundUpToNearest(input, multiple) - input;
}

//-----------------------------------------------------------------------------------------------------------
lil_u32 CalcBpp(const LilFormatDesc *format)
{
	switch (format->dataFormat) {
		case LDF_SIMPLE:
			return format->bitsPerChannel * format->numChannels;
		case LDF_SIMPLEBITS:
		{
			lil_u32 total = 0;
			for (int i = 0; i < format->numChannels; i++)
				total += format->bitCounts[i];
			return total;
		}
		default:
			// @@ If this happens, it means we forgot to implement this dataFormat!
			LIL_BREAK();
			return 0;
	}
}

//-----------------------------------------------------------------------------------------------------------
lil_u32 CalcMinBytesNeeded(const LilFormatDesc *format, lil_u32 numPixels)
{
	return (CalcMinBitsNeeded(format, numPixels) + 7) >> 3;
}

//-----------------------------------------------------------------------------------------------------------
lil_u32 CalcMinBitsNeeded(const LilFormatDesc *format, lil_u32 numPixels)
{
	lil_u32 bpu;  // bits per unit
	lil_u32 retVal;
	switch (format->dataFormat) {
		case LDF_SIMPLE:
			bpu = format->bitsPerChannel;
			if (!format->planar)
				bpu *= format->numChannels;
			retVal = numPixels * bpu;
			break;
		case LDF_SIMPLEBITS:
		{
			lil_u32 total = 0;
			for (int i = 0; i < format->numChannels; i++)
				total += format->bitCounts[i];
			retVal = numPixels * total;
			break;
		}
		default:
			// @@ If this happens, it means we forgot to implement this dataFormat!
			LIL_BREAK();
			return 0;
	}
	return retVal;
}

//-----------------------------------------------------------------------------------------------------------
lil_u32 CalcBpl(const LilFormatDesc *format, lil_u32 width)
{
	lil_u32 bytes = CalcMinBytesNeeded(format, width);
	bytes = RoundUpToNearest(bytes, format->blockLineByteAlign);
	return bytes;
}

//-----------------------------------------------------------------------------------------------------------
lil_u32 CalcBitsPerLine(const LilFormatDesc *format, lil_u32 width)
{
	if (format->blockLineByteAlign > 0)
		return CalcBpl(format, width) * 8;
	return CalcMinBitsNeeded(format, width);
}

//-----------------------------------------------------------------------------------------------------------
lil_u64 CalcByteSize(const LilFormatDesc *format, lil_u32 width, lil_u32 height)
{
	lil_u64 BPL = static_cast<lil_u64>(CalcBpl(format, width));
	return BPL * height * (format->planar ? format->numChannels : 1);
}



//-----------------------------------------------------------------------------------------------------------
// LILBUFFER
//-----------------------------------------------------------------------------------------------------------
LilBuffer::LilBuffer() : mHandle(0), mSize(0), ReallocFunc(0) {
	memset(&mIoMem, 0, sizeof(mIoMem));
}

//-----------------------------------------------------------------------------------------------------------
LilBuffer::~LilBuffer() {
	if (mHandle)
		ReallocFunc(&mHandle, 0);
	lil::IoCloseMem(&mIoMem);
}

//-----------------------------------------------------------------------------------------------------------
bool LilBuffer::Realloc(const LilMemDesc *pMemFuncs, size_t size) {
//	LIL_ASSERT(mSize == 0);

	LilErrorType err = pMemFuncs->memReallocFunc(&mHandle, size);
	if (!mHandle && size)
		return false;
	if (err != LE_NONE)
		return false;

	mSize = size;
	ReallocFunc = pMemFuncs->memReallocFunc;

	IoOpenMem(&mIoMem, GetPtr<lil_u8>(), mSize);
	return true;
}

}  // namespace lil
