#include "LilUtilFormat.h"
#include "LilMem.h"
#include "LilBitManip.h"
#include "LilConvert.h"

namespace lil {

//-----------------------------------------------------------------------------------------------------------
bool SetError(LilErrorType err, LilErrorType *pErr) {
    if (pErr)
        *pErr = err;
    return err == LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
bool IsFormatValid(const LilFormatDesc *pFormat, LilErrorType *pErr)
{
	if (pErr)
		*pErr = LE_INVALIDFORMAT;
	if (pFormat->channelOrder != LCO_UNKNOWN && LIL_NUM_CHANNELS[pFormat->channelOrder] != pFormat->numChannels)
		return false;

	switch (pFormat->dataFormat) {
    case LDF_SIMPLE: break;
	case LDF_SIMPLEBITS:
		if (pFormat->planar)
			return false;
		if (pFormat->numChannels < 2)
			return false;
		if (pFormat->bitsPerChannel != 0)
			return false;
		if (pFormat->channelOrder == LCO_SINGLE)
			return false;
		break;
	}

	if (pErr)
		*pErr = LE_NONE;
	return true;
}

//-----------------------------------------------------------------------------------------------------------
bool SimplifyFormat(LilFormatDesc *pFormat) {
	if (pFormat->dataFormat == LDF_SIMPLEBITS) {
		bool diff = false;
		for (lil_u32 i = 1; i < pFormat->numChannels; i++)
			if (pFormat->bitCounts[i] != pFormat->bitCounts[0])
				diff = true;
		if (!diff) {
			pFormat->dataFormat = LDF_SIMPLE;
			pFormat->bitsPerChannel = pFormat->bitCounts[0];
			return true;
		}
	}
	return false;
}

//-----------------------------------------------------------------------------------------------------------
bool SetFormat(LilFormatDesc *pFormat, LilDataFormatType dataFormat, lil_u8 bitsPerChannel,
			   lil_u8 numChannels, lil_u8 blockLineByteAlign, bool planar, LilChannelOrderType channelOrder,
			   bool flipped, LilErrorType *pErr)
{
	pFormat->dataFormat = dataFormat;
	pFormat->bitsPerChannel = bitsPerChannel;
	pFormat->numChannels = numChannels;
	pFormat->blockLineByteAlign = blockLineByteAlign;
	pFormat->planar = planar;
	pFormat->channelOrder = channelOrder;
	pFormat->flipped = flipped;
	return IsFormatValid(pFormat, pErr);
}

//-----------------------------------------------------------------------------------------------------------
bool SetImageSize(LilDataDesc *pDataDesc, lil_u32 width, lil_u32 height, LilErrorType *pErr) {
	pDataDesc->width = width;
	pDataDesc->height = height;
	if (width == 0 || height == 0)
		return SetError(LE_BADIMAGESIZE, pErr);
	return SetError(LE_NONE, pErr);
}

//-----------------------------------------------------------------------------------------------------------
bool SetExtendedInfo(LilExtendedDesc *pExt, lil_u32 numPages, lil_u32 compression,
	LilDpuType dpu, lil_u32 dpuX, lil_u32 dpuY, LilErrorType *pErr)
{
	// PAGES
	pExt->numPages = numPages;
	// pageIndex should have been set externally.
	if (pExt->pageIndex >= pExt->numPages)
		return SetError(LE_INVALIDPAGE, pErr);
	// COMPRESSION
	pExt->compression = static_cast<LilCompressionType>(compression);
	// DPU
	pExt->dotsPerWhat = dpu;
	pExt->dpuX = dpuX;
	pExt->dpuY = dpuY;

	return SetError(LE_NONE, pErr);
}

//-----------------------------------------------------------------------------------------------------------
bool SetPalette(LilPaletteDesc *pPalette, const LilImageDesc *pImage, const LilFormatDesc *pFormat,
    lil_u32 numEntries, LilBuffer *pTempBuf, LilErrorType *pErr)
{
    LilErrorType err = LE_NONE;
    size_t size = static_cast<size_t>(CalcByteSize(pFormat, numEntries, 1));

    pPalette->format = *pFormat;
    pPalette->numEntries = 0;
    // pPalette->lut is taken care of properly inside ConvertPalette!
    err = ConvertPalette(pImage, pPalette, pTempBuf);
	if (err != LE_NONE)
		return SetError(err, pErr);

    if (numEntries != pPalette->numEntries) {
        if (pTempBuf) {
            if (!pTempBuf->Realloc(&pImage->memFuncs, size))
				return SetError(LE_NOTENOUGHMEMORY, pErr);
            pPalette->lut = pTempBuf->GetPtr<void>();
        }
        else {
            err = pImage->memFuncs.memReallocFunc(&pPalette->lut, size);
			if (err != LE_NONE)
				return SetError(err, pErr);
		}

        // @@ Default grayscale palette on extra parts... create in rgb, convert to
        // palette format with ConvertPalette, copy only entries that are needed...

        pPalette->numEntries = numEntries;
    }

    return SetError(LE_NONE, pErr);
}

//-----------------------------------------------------------------------------------------------------------
bool IsFormatDifferent(const LilFormatDesc *pSrcFormat, const LilFormatDesc *pDestFormat, lil_u32 flags)
{
	LIL_ASSERT(pSrcFormat || pDestFormat);
	if ((pSrcFormat && pDestFormat == 0) || (pDestFormat && pSrcFormat == 0))
		return false;
	// @@ Add ability to count LDF_SIMPLEBITS with all like bit sizes as the same as LDF_SIMPLE with a bit size for channel depth
	if (pSrcFormat->bitsPerChannel != pDestFormat->bitsPerChannel)
		return true;
	if (!(flags & IFD_IGNORE_STRIDE) && pSrcFormat->blockLineByteAlign != pDestFormat->blockLineByteAlign)
		return true;
	if (pSrcFormat->dataFormat != pDestFormat->dataFormat)
		return true;
	if (pSrcFormat->numChannels != pDestFormat->numChannels)
		return true;
	if (pSrcFormat->channelOrder != pDestFormat->channelOrder)
		return true;
	if (!(flags & IFD_IGNORE_FLIPPED) && pSrcFormat->flipped != pDestFormat->flipped)
		return true;
	if (!(flags & IFD_IGNORE_PLANAR) && pSrcFormat->planar != pDestFormat->planar)
		return true;

	bool diff = false;
	if (pSrcFormat->dataFormat == LDF_SIMPLEBITS) {
		for (lil_u32 i = 0; i < pSrcFormat->numChannels; i++)
			if (pSrcFormat->bitCounts[i] != pDestFormat->bitCounts[i])
				diff = true;
	}
	if (!diff)
		return false;

	return true;
}

//-----------------------------------------------------------------------------------------------------------
bool IsColorSpaceDifferent(LilChannelOrderType lco1, LilChannelOrderType lco2) {
	LilChannelOrderType lcos[2] = { lco1, lco2 };
	lil_u32 colorSpace[2] = { 0xffffffff, 0xffffffff };
	for (lil_u32 k = 0; k < 2; k++)
		for (lil_u32 i = 0; i < COLOR_SPACE_COUNT; i++)
			for (lil_u32 j = 0; j < COLOR_SPACES[i].count; j++)
				if (COLOR_SPACES[i].colorSpace[j] == lcos[k]) {
					colorSpace[k] = i;
					j = COLOR_SPACES[i].count;
					i = COLOR_SPACE_COUNT;
				}
	LIL_ASSERT(colorSpace[0] != 0xffffffff && colorSpace[1] != 0xffffffff);
	return colorSpace[0] != colorSpace[1];
}

//-----------------------------------------------------------------------------------------------------------
LilChannelOrderType GetDefaultColorspace(LilChannelOrderType lco)
{
	for (lil_u32 i = 0; i < COLOR_SPACE_COUNT; i++)
		for (lil_u32 j = 0; j < COLOR_SPACES[i].count; j++)
			if (COLOR_SPACES[i].colorSpace[j] == lco) {
				if (LIL_ALPHA_CHANNEL[lco] != LIL_CHANNEL_NONE)
					return COLOR_SPACE_DEFAULTS[i].yesAlpha;
				return COLOR_SPACE_DEFAULTS[i].noAlpha;
			}
	LIL_ASSERT(0);  // it should have found a color space!!
	return LCO_UNKNOWN;
}

//-----------------------------------------------------------------------------------------------------------
bool HasAlpha(const LilFormatDesc *pFormat) {
	return LIL_ALPHA_CHANNEL[pFormat->channelOrder] != LIL_CHANNEL_NONE;
}

//-----------------------------------------------------------------------------------------------------------
bool SetFormatFromBppWin(LilFormatDesc *format, lil_u8 bpp, bool flipped, LilErrorType *pErr)
{
	switch (bpp) {
	case 1: case 4: case 8:
		return SetFormat(format, LDF_SIMPLE, bpp, 1, 4, false, LCO_SINGLE, flipped, pErr);
	case 15: case 16:
		format->bitCounts[0] = 1;
		format->bitCounts[1] = 5;
		format->bitCounts[2] = 5;
		format->bitCounts[3] = 5;
		return SetFormat(format, LDF_SIMPLEBITS, 0, 4, 4, false, LCO_ARGB, flipped, pErr);
	case 24: case 32:
		return SetFormat(format, LDF_SIMPLE, 8, bpp / 8, 4, false,
			bpp == 24 ? LCO_BGR : LCO_BGRA, flipped, pErr);
	}
	if (pErr)
		*pErr = LE_BADBPP;
	return false;
}

//-----------------------------------------------------------------------------------------------------------
bool SetFormatFromBppTga(LilFormatDesc *format, lil_u8 attrBits, lil_u8 bpp, bool alignBottom, LilErrorType *pErr) {
	if ((attrBits && attrBits != bpp / 4) || bpp == 16) {
		lil_u8 sub = attrBits;
		if (sub == 0)
			sub = 1;  // bpp == 16
		format->bitCounts[0] = sub;
		format->bitCounts[1] = format->bitCounts[2] = format->bitCounts[3] = (bpp - sub) / 3;
		return SetFormat(format, LDF_SIMPLEBITS, 0, 4, 0, false, LCO_ARGB, alignBottom, pErr);
	}

	lil_u8 chan = 1;
	LilChannelOrderType co = LCO_SINGLE;

	if (bpp > 8) {
		chan = 3;
		co = LCO_BGR;
	}
	if (attrBits || bpp == 32) {
		chan++;
		co = LCO_BGRA;
	}
	return SetFormat(format, LDF_SIMPLE, bpp / chan, chan, 0, false, co, alignBottom, pErr);
}

//-----------------------------------------------------------------------------------------------------------
lil_u32 GetAlphaBits(const LilFormatDesc *pFormat) {
	return GetChannelBits(pFormat, LIL_ALPHA_CHANNEL[pFormat->channelOrder]);
}

//-----------------------------------------------------------------------------------------------------------
lil_u32 GetChannelBits(const LilFormatDesc *pFormat, lil_u8 channel) {
	if (channel != LIL_CHANNEL_NONE) {
		if (pFormat->dataFormat == LDF_SIMPLE)
			return pFormat->bitsPerChannel;
		if (pFormat->dataFormat == LDF_SIMPLEBITS)
			return pFormat->bitCounts[channel];
		LIL_ASSERT(0);  // oops... need support for something we forgot to add apparently
	}
	return 0;
}

//-----------------------------------------------------------------------------------------------------------
bool CheckValue(lil_i32 value, const lil_i32 *valids, lil_u8 numValids, const lil_i32 *invalids, lil_u8 numInvalids,
	LilErrorType invalidErr, LilErrorType noneErr, LilErrorType *outErr)
{
	for (lil_u8 i = 0; i < numValids; i++)
		if (value == valids[i]) {
			*outErr = LE_NONE;
			return true;
		}
	*outErr = noneErr;
	for (lil_u8 i = 0; i < numInvalids; i++)
		if (value == invalids[i])
			*outErr = invalidErr;
	return false;
}

//-----------------------------------------------------------------------------------------------------------
bool MatchFormat(LilFormatDesc *pFormat, const LilFormatRange *pRanges, lil_u32 rangeCount,
				 const LilSimpleBitsRange *pBitRanges, lil_u32 bitRangeCount, LilErrorType *pErr)
{
	// We can deal with planar/flipped/color order as long as the other stuff matches.
	// Stride changes are practically irrelevant.
	SimplifyFormat(pFormat);
	lil_u32 match = 0xffffffff;  // Which pRange did we match?
	lil_u32 sub = 0xffffffff;  // Which index in pRange's bit descriptions did we match?

	//-------------------------------
	// Part 1: Try the regular ranges
	//-------------------------------
	for (lil_u32 i = 0; i < rangeCount; i++) {
		if (pFormat->dataFormat != pRanges[i].dataFormat) continue;
		if (pFormat->numChannels != pRanges[i].numChannels) continue;
		bool found = false;
		lil_u32 maybeSub = 0;

		switch (pFormat->dataFormat) {
		case LDF_SIMPLE:
			for (lil_u32 j = 0; j < pRanges[i].bitArrSize; j++)
				if (pFormat->bitsPerChannel == pRanges[i].bitCounts[j][0]) {
					maybeSub = j;
					found = true;
					break; // out of for loop
				}
			if (!found) continue;
			// If the rest is the same, we have a definite match!
			if (pFormat->channelOrder == pRanges[i].lco) {
				if (((pFormat->flipped == false && (pRanges[i].flippedFlags & FLAG_NO)) ||
					(pFormat->flipped == true && (pRanges[i].flippedFlags & FLAG_YES))) &&
					((pFormat->planar == false && (pRanges[i].planarFlags & FLAG_NO)) ||
					(pFormat->planar == true && (pRanges[i].planarFlags & FLAG_YES))))
				{
					match = i;
					sub = maybeSub;
					i = rangeCount;  // DEFINITE MATCH
					continue;  // This will break out of entire for loop and go straight to end of function
				}
				// Once we have a potential, don't worry about finding more - just look for direct matches
				else if (match != 0xffffffff)
					continue;
				else {
					match = i;
					sub = maybeSub;
					break;  // from switch statement
				}
			}
			// Once we have a potential, don't worry about finding more - just look for direct matches
			if (match != 0xffffffff)
				continue;
			// Otherwise, check for potentials along the way
			for (lil_u32 j = 0; j < CHANNEL_SWAPS[pRanges[i].lco].count; j++) {
				// If we can get a close-match for channel order, flipping and planar
				// will fall into the potential category no matter what.
				if (pFormat->channelOrder == CHANNEL_SWAPS[pRanges[i].lco].channels[j]) {
					match = i;
					sub = maybeSub;
					break;
				}
			}
			break;
		case LDF_SIMPLEBITS:
			for (lil_u32 j = 0; j < pRanges[i].bitArrSize; j++) {
				bool def = true;
				for (lil_u32 k = 0; k < pRanges[i].numChannels; k++)
					if (pFormat->bitCounts[k] != pRanges[i].bitCounts[j][k])
						def = false;
				if (def) {
					found = true;
					maybeSub = j;
					break; // out of for loop
				}
			}
			if (!found) continue;
			// If the rest is the same, we have a definite match!
			if (pFormat->channelOrder == pRanges[i].lco) {
				if (((pFormat->flipped == false && (pRanges[i].flippedFlags & FLAG_NO)) ||
					(pFormat->flipped == true && (pRanges[i].flippedFlags & FLAG_YES))) &&
					((pFormat->planar == false && (pRanges[i].planarFlags & FLAG_NO)) ||
					(pFormat->planar == true && (pRanges[i].planarFlags & FLAG_YES))))
				{
					match = i;
					sub = maybeSub;
					i = rangeCount;  // DEFINITE MATCH
					continue;  // This will break out of entire for loop and go straight to end of function
				}
				// Once we have a potential, don't worry about finding more - just look for direct matches
				else if (match != 0xffffffff)
					continue;
				else {
					match = i;
					sub = maybeSub;
					break;  // from switch statement
				}
			}
			// Once we have a potential, don't worry about finding more - just look for direct matches
			if (match != 0xffffffff)
				continue;
			// Otherwise, check for potentials along the way


			// @@ LATER!
			// We will need to test not only that all bits match, but that the supported
			// swizzling of these bits match, depending on the channel order vs supported channel
			// orders!
			// @@ At least check for a direct match for now
			break;
		}
	}

	if (match != 0xffffffff) {
		LIL_ASSERT(sub != 0xffffffff);
		switch (pFormat->dataFormat) {
		case LDF_SIMPLE:
			{
				bool finalPlanar = pFormat->planar;
				if (pRanges[match].planarFlags != FLAG_YESNO)
					finalPlanar = (pRanges[match].planarFlags & FLAG_YES ? true : false);
				bool finalFlipped = pFormat->flipped;
				if (pRanges[match].flippedFlags != FLAG_YESNO)
					finalFlipped = (pRanges[match].flippedFlags & FLAG_YES ? true : false);
				return SetFormat(pFormat, pRanges[match].dataFormat, pRanges[match].bitCounts[sub][0],
					pRanges[match].numChannels, pRanges[match].blockLineByteAlign,
					finalPlanar, pRanges[match].lco, finalFlipped, pErr);
			}
		case LDF_SIMPLEBITS:
			{
				bool finalPlanar = pFormat->planar;
				if (pRanges[match].planarFlags != FLAG_YESNO)
					finalPlanar = (pRanges[match].planarFlags & FLAG_YES ? true : false);
				bool finalFlipped = pFormat->flipped;
				if (pRanges[match].flippedFlags != FLAG_YESNO)
					finalFlipped = (pRanges[match].flippedFlags & FLAG_YES ? true : false);
				for (lil_u32 i = 0; i < pRanges[match].numChannels; i++)
					pFormat->bitCounts[i] = pRanges[match].bitCounts[sub][i];
				return SetFormat(pFormat, pRanges[match].dataFormat, 0,
					pRanges[match].numChannels, pRanges[match].blockLineByteAlign,
					finalPlanar, pRanges[match].lco, finalFlipped, pErr);
			}
		}
	}

	//-------------------------------
	// Part 2: Try the bit ranges
	//-------------------------------
	for (lil_u32 i = 0; i < bitRangeCount; i++) {
		if (pFormat->numChannels != pBitRanges[i].numChannels) continue;
		if (CalcBpp(pFormat) != pBitRanges[i].totalBits) continue;
		lil_u32 maybeSub = 0;
		if (pFormat->channelOrder == pBitRanges[i].lco) {
			if (((pFormat->flipped == false && (pBitRanges[i].flippedFlags & FLAG_NO)) ||
				(pFormat->flipped == true && (pBitRanges[i].flippedFlags & FLAG_YES))) &&
				((pFormat->planar == false && (pBitRanges[i].planarFlags & FLAG_NO)) ||
				(pFormat->planar == true && (pBitRanges[i].planarFlags & FLAG_YES))))
			{
				match = i;
				sub = maybeSub;
				i = bitRangeCount;  // DEFINITE MATCH
				continue;  // This will break out of entire for loop and go straight to end of function
			}
			// Once we have a potential, don't worry about finding more - just look for direct matches
			else if (match != 0xffffffff)
				continue;
			else {
				match = i;
				sub = maybeSub;
				break;  // from switch statement
			}
		}
		// Once we have a potential, don't worry about finding more - just look for direct matches
		if (match != 0xffffffff)
			continue;
		// Otherwise, check for potentials along the way
		for (lil_u32 j = 0; j < CHANNEL_SWAPS[pBitRanges[i].lco].count; j++) {
			// If we can get a close-match for channel order, flipping and planar
			// will fall into the potential category no matter what.
			if (pFormat->channelOrder == CHANNEL_SWAPS[pBitRanges[i].lco].channels[j]) {
				match = i;
				sub = maybeSub;
				break;
			}
		}
	}

	if (match != 0xffffffff) {
		LIL_ASSERT(sub != 0xffffffff);
		switch (pFormat->dataFormat) {
			// @@ What about... like... 4 channels x 4 bits per channel for bmp files???
/*		case LDF_SIMPLE:
			{
				bool finalPlanar = pFormat->planar;
				if (pBitRanges[match].planarFlags != FLAG_YESNO)
					finalPlanar = (pBitRanges[match].planarFlags & FLAG_YES ? true : false);
				bool finalFlipped = pFormat->flipped;
				if (pBitRanges[match].flippedFlags != FLAG_YESNO)
					finalFlipped = (pBitRanges[match].flippedFlags & FLAG_YES ? true : false);
				return SetFormat(pFormat, pBitRanges[match].dataFormat, pBitRanges[match].bitCounts[sub][0],
					pBitRanges[match].numChannels, pBitRanges[match].blockLineByteAlign,
					finalPlanar, pBitRanges[match].lco, finalFlipped, pErr);
			}*/
		case LDF_SIMPLEBITS:
			{
				bool finalPlanar = pFormat->planar;
				if (pBitRanges[match].planarFlags != FLAG_YESNO)
					finalPlanar = (pBitRanges[match].planarFlags & FLAG_YES ? true : false);
				bool finalFlipped = pFormat->flipped;
				if (pBitRanges[match].flippedFlags != FLAG_YESNO)
					finalFlipped = (pBitRanges[match].flippedFlags & FLAG_YES ? true : false);
				return SetFormat(pFormat, LDF_SIMPLEBITS, 0,
					pBitRanges[match].numChannels, pBitRanges[match].blockLineByteAlign,
					finalPlanar, pBitRanges[match].lco, finalFlipped, pErr);
			}
		}
	}
	return false;
}

}  // namespace lil
